/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Bank_DRAM.h"

#include "Bank_ROM.h"			// ROMBank::kMemoryStart
#include "CPU_REG.h"			// Software::BusError, Screen::fgScreenLow, fgScreenHigh
#include "DebugMgr.h"			// Debug::CheckStepSpy
#include "MetaMemory.h"			// MetaMemory
#include "RAM_ROM.h"			// gRAMBank_Size, gRAM_Memory, gMemoryAccess
#include "TrapPatches.h"		// Patches::fgPreventLowMemAccess


// ===========================================================================
//		 DRAM Bank Accessors
// ===========================================================================
// These functions provide fetch and store access to the emulator's random
// access memory.

AddressBank	DRAMBank::fgAddressBank =
{
	DRAMBank::GetLong,
	DRAMBank::GetWord,
	DRAMBank::GetByte,
	DRAMBank::SetLong,
	DRAMBank::SetWord,
	DRAMBank::SetByte,
	DRAMBank::GetRealAddress,
	DRAMBank::ValidAddress
};

const uae_u32	DRAMBank::kMemoryStart = 0x00000000;
const uaecptr	kGlobalStart = offsetof (LowMemHdrType, globals);

/*
// !!! TBD Use these to determine heap size(s):

	UInt		MemNumCards(void)
	UInt		MemNumHeaps(UInt cardNo)
	UInt		MemHeapID(UInt cardNo, UInt heapNo)
	Boolean		MemHeapDynamic(UInt heapID)
	VoidPtr		MemHeapPtr(UInt heapID)
	ULong		MemHeapSize(UInt heapID)
*/

#define META_CHECK(op, size, forRead)							\
	if (ROMBank::IsPCInRAM ())									\
	{															\
		if (!MetaMemory::CanApp##op (iAddress))					\
		{														\
			ProbableCause (iAddress, sizeof (size), forRead);	\
		}														\
	}															\
	else if (Patches::IsPCInMemMgr ())							\
	{															\
		if (!MetaMemory::CanMemMgr##op (iAddress))				\
		{														\
			ProbableCause (iAddress, sizeof (size), forRead);	\
		}														\
	}															\
	else														\
	{															\
		if (!MetaMemory::CanSystem##op (iAddress))				\
		{														\
			ProbableCause (iAddress, sizeof (size), forRead);	\
		}														\
	}


static inline void ScreenCheck (uaecptr iAddress, uae_u32 size)
{
#if defined (macintosh)

	if (size == 1 && !MetaMemory::IsScreenBuffer8 (iAddress))
		return;

	if (size == 2 && !MetaMemory::IsScreenBuffer16 (iAddress))
		return;

	if (size == 4 && !MetaMemory::IsScreenBuffer32 (iAddress))
		return;
#else

	if (MetaMemory::IsScreenBuffer (iAddress, size))

#endif

	{
		// SUBTLE BUG NOTE:  Both of these tests need to be performed.
		// Originally, I had this as an if (...) ... else if (...) ...
		// After all, how could a given address be above the high-
		// water mark if it was below the low-water mark?  "Obviously",
		// it couldn't, so I didn't see the need to test against the
		// high-water mark if we were below the low-water mark.
		//
		// Well, duh, that assumption was false.  After the LCD buffer
		// is copied to the screen, we set the low-water mark to the
		// end of the buffer and the high-water mark to the beginning
		// of the buffer.  If the screen is modified in such a way
		// that each pixel affected appears lower in memory than any
		// previous pixel (as happens when we scroll a document down),
		// then we always entered the low-water mark of the code below;
		// we never entered the high-water mark of the code.  Thus, the
		// high-water mark stayed set to the beginning of the buffer,
		// giving us a NULL update range.

		if (iAddress < Screen::fgScreenLow)
			Screen::fgScreenLow = iAddress;

		if (iAddress + size > Screen::fgScreenHigh)
			Screen::fgScreenHigh = iAddress + size;
	}
}


// ---------------------------------------------------------------------------
//		 DRAMBank::Initialize
// ---------------------------------------------------------------------------

void DRAMBank::Initialize (void)
{
	// Memory banks 0 and 1 are managed by the functions in DRAMBank.

	Memory::InitializeBanks (	fgAddressBank,
								bankindex (kMemoryStart), 2);
}


// ---------------------------------------------------------------------------
//		 DRAMBank::Reset
// ---------------------------------------------------------------------------

void DRAMBank::Reset (void)
{
}


// ---------------------------------------------------------------------------
//		 DRAMBank::GetLong
// ---------------------------------------------------------------------------

uae_u32 DRAMBank::GetLong (uaecptr iAddress)
{
#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMLongRead]++;
	if (iAddress & 2)
		gMemoryAccess[kDRAMLongRead2]++;
#endif

	if (CHECK_FOR_ADDRESS_ERROR && (iAddress & 1) != 0)
	{
		Software::AddressError ();
	}

	META_CHECK (GetLong, long, true)

	if (VALIDATE_DRAM_GET &&
		gMemAccessFlags.fValidate_DRAMGet &&
		!ValidAddress (iAddress, sizeof (long)))
	{
		InvalidAccess (iAddress, sizeof (long), true);
	}

#if (HAS_PROFILING)
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementRead (2, WAITSTATES_DRAM);
	}
#endif

	return do_get_mem_long (gRAM_Memory + iAddress);
}


// ---------------------------------------------------------------------------
//		 DRAMBank::GetWord
// ---------------------------------------------------------------------------

uae_u32 DRAMBank::GetWord (uaecptr iAddress)
{
#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMWordRead]++;
#endif

	if (CHECK_FOR_ADDRESS_ERROR && (iAddress & 1) != 0)
	{
		Software::AddressError ();
	}

	META_CHECK (GetWord, short, true)

	if (VALIDATE_DRAM_GET &&
		gMemAccessFlags.fValidate_DRAMGet &&
		!ValidAddress (iAddress, sizeof (short)))
	{
		InvalidAccess (iAddress, sizeof (short), true);
	}

#if (HAS_PROFILING)
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementRead (1, WAITSTATES_DRAM);
	}
#endif

	return do_get_mem_word (gRAM_Memory + iAddress);
}


// ---------------------------------------------------------------------------
//		 DRAMBank::GetByte
// ---------------------------------------------------------------------------

uae_u32 DRAMBank::GetByte (uaecptr iAddress)
{
#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMByteRead]++;
#endif

	META_CHECK (GetByte, char, true)

	if (VALIDATE_DRAM_GET &&
		gMemAccessFlags.fValidate_DRAMGet &&
		!ValidAddress (iAddress, sizeof (char)))
	{
		InvalidAccess (iAddress, sizeof (char), true);
	}

#if (HAS_PROFILING)
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementRead (1, WAITSTATES_DRAM);
	}
#endif

	return do_get_mem_byte (gRAM_Memory + iAddress);
}


// ---------------------------------------------------------------------------
//		 DRAMBank::SetLong
// ---------------------------------------------------------------------------

void DRAMBank::SetLong (uaecptr iAddress, uae_u32 iLongValue)
{
#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMLongWrite]++;
	if (iAddress & 2)
		gMemoryAccess[kDRAMLongWrite2]++;
#endif

	if (CHECK_FOR_ADDRESS_ERROR && (iAddress & 1) != 0)
	{
		Software::AddressError ();
	}

	META_CHECK (SetLong, long, false)

	if (VALIDATE_DRAM_SET &&
		gMemAccessFlags.fValidate_DRAMSet &&
		!ValidAddress (iAddress, sizeof (long)))
	{
		InvalidAccess (iAddress, sizeof (long), false);
	}

	ScreenCheck (iAddress, sizeof (long));

#if (HAS_PROFILING)
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementWrite (2, WAITSTATES_DRAM);
	}
#endif

	do_put_mem_long (gRAM_Memory + iAddress, iLongValue);

	// Mark that this memory location can now be read from.

	MetaMemory::MarkLongInitialized (iAddress);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 4);
}


// ---------------------------------------------------------------------------
//		 DRAMBank::SetWord
// ---------------------------------------------------------------------------

void DRAMBank::SetWord (uaecptr iAddress, uae_u32 iWordValue)
{
#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMWordWrite]++;
#endif

	if (CHECK_FOR_ADDRESS_ERROR && (iAddress & 1) != 0)
	{
		Software::AddressError ();
	}

	META_CHECK (SetWord, short, false)

	if (VALIDATE_DRAM_SET &&
		gMemAccessFlags.fValidate_DRAMSet &&
		!ValidAddress (iAddress, sizeof (short)))
	{
		InvalidAccess (iAddress, sizeof (short), false);
	}

	ScreenCheck (iAddress, sizeof (short));

#if (HAS_PROFILING)
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementWrite (1, WAITSTATES_DRAM);
	}
#endif

	do_put_mem_word (gRAM_Memory + iAddress, iWordValue);

	// Mark that this memory location can now be read from.

	MetaMemory::MarkWordInitialized (iAddress);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 2);
}


// ---------------------------------------------------------------------------
//		 DRAMBank::SetByte
// ---------------------------------------------------------------------------

void DRAMBank::SetByte (uaecptr iAddress, uae_u32 iByteValue)
{
#if (PROFILE_MEMORY)
	gMemoryAccess[kDRAMByteWrite]++;
#endif

	META_CHECK (SetByte, char, false)

	if (VALIDATE_DRAM_SET &&
		gMemAccessFlags.fValidate_DRAMSet &&
		!ValidAddress (iAddress, sizeof (char)))
	{
		InvalidAccess (iAddress, sizeof (char), false);
	}

	ScreenCheck (iAddress, sizeof (char));

#if (HAS_PROFILING)
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementWrite (1, WAITSTATES_DRAM);
	}
#endif

	do_put_mem_byte (gRAM_Memory + iAddress, iByteValue);

	// Mark that this memory location can now be read from.

	MetaMemory::MarkByteInitialized (iAddress);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 1);
}


// ---------------------------------------------------------------------------
//		 DRAMBank::ValidAddress
// ---------------------------------------------------------------------------

int DRAMBank::ValidAddress (uaecptr iAddress, uae_u32 iSize)
{
	int	result = (iAddress + iSize) <= gRAMBank_Size;

	assert (result);

	return result;
}


// ---------------------------------------------------------------------------
//		 DRAMBank::GetRealAddress
// ---------------------------------------------------------------------------

uae_u8* DRAMBank::GetRealAddress (uaecptr iAddress)
{
	return (uae_u8*) &(gRAM_Memory[iAddress]);
}


// ---------------------------------------------------------------------------
//		 DRAMBank::InvalidAccess
// ---------------------------------------------------------------------------

void DRAMBank::InvalidAccess (uaecptr iAddress, long size, Bool forRead)
{
	Software::BusError ();
}


// ---------------------------------------------------------------------------
//		 DRAMBank::ProbableCause
// ---------------------------------------------------------------------------

void DRAMBank::ProbableCause (uaecptr iAddress, long size, Bool forRead)
{
	Errors::EAccessType	whatHappened = MetaMemory::GetWhatHappened (iAddress, size, forRead);

	if (whatHappened != Errors::kOKAccess)
	{
		assert (whatHappened != Errors::kUnknownAccess);

		Memory::PreventedAccess (iAddress, size, forRead, whatHappened);
	}
}
