/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Bank_ROM.h"

#include "CPU_REG.h"			// Software::BusError
#include "RAM_ROM.h"			// gROMBank_Size, gROM_Memory, gMemoryAccess
#include "UAE_Utils.h"			// uae_memset


// ===========================================================================
//		 ROM Bank Accessors
// ===========================================================================
// These functions provide fetch and store access to the emulator's read only
// memory.

AddressBank	ROMBank::fgAddressBank =
{
	ROMBank::GetLong,
	ROMBank::GetWord,
	ROMBank::GetByte,
	ROMBank::SetLong,
	ROMBank::SetWord,
	ROMBank::SetByte,
	ROMBank::GetRealAddress,
	ROMBank::ValidAddress
};

const uae_u32	ROMBank::kMemoryStart = 0x10C00000;


// ---------------------------------------------------------------------------
//		 ROMBank::Initialize
// ---------------------------------------------------------------------------

void ROMBank::Initialize (void)
{
	// Memory banks 0x10C0 to <mumble> are managed by the functions
	// in ROMBank.  <mumble> is based on the amount of ROM being emulated.

	Memory::InitializeBanks (	fgAddressBank,
								bankindex (kMemoryStart),
								bankindex (kMemoryStart + gROMBank_Size - 1) -
									bankindex (kMemoryStart) + 1);
}


// ---------------------------------------------------------------------------
//		 ROMBank::Reset
// ---------------------------------------------------------------------------

void ROMBank::Reset (void)
{
}


// ---------------------------------------------------------------------------
//		 ROMBank::GetLong
// ---------------------------------------------------------------------------

uae_u32 ROMBank::GetLong (uaecptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kROMLongRead]++;
	if (iAddress & 2)
		gMemoryAccess[kROMLongRead2]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (VALIDATE_ROM_GET)
	if (gMemAccessFlags.fValidate_ROMGet && !ValidAddress (iAddress, sizeof (long)))
	{
		InvalidAccess (iAddress, sizeof (long), true);
	}
#endif

#if (PREVENT_USER_ROM_GET || PREVENT_SYSTEM_ROM_GET)
	if (IsPCInRAM ())
	{
		if (PREVENT_USER_ROM_GET && gMemAccessFlags.fProtect_ROMGet)
		{
			InvalidAccess (iAddress, sizeof (long), true);
		}
	}
	else
	{
		if (PREVENT_SYSTEM_ROM_GET && gMemAccessFlags.fProtect_SysROMGet)
		{
			InvalidAccess (iAddress, sizeof (long), true);
		}
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementRead (2, WAITSTATES_ROM);
	}
#endif

	iAddress &= gROMBank_Mask;

	return do_get_mem_long (gROM_Memory + iAddress);
}


// ---------------------------------------------------------------------------
//		 ROMBank::GetWord
// ---------------------------------------------------------------------------

uae_u32 ROMBank::GetWord (uaecptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kROMWordRead]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (VALIDATE_ROM_GET)
	if (gMemAccessFlags.fValidate_ROMGet && !ValidAddress (iAddress, sizeof (short)))
	{
		InvalidAccess (iAddress, sizeof (short), true);
	}
#endif

#if (PREVENT_USER_ROM_GET || PREVENT_SYSTEM_ROM_GET)
	if (IsPCInRAM ())
	{
		if (PREVENT_USER_ROM_GET && gMemAccessFlags.fProtect_ROMGet)
		{
			InvalidAccess (iAddress, sizeof (short), true);
		}
	}
	else
	{
		if (PREVENT_SYSTEM_ROM_GET && gMemAccessFlags.fProtect_SysROMGet)
		{
			InvalidAccess (iAddress, sizeof (short), true);
		}
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementRead (1, WAITSTATES_ROM);
	}
#endif

	iAddress &= gROMBank_Mask;

	return do_get_mem_word (gROM_Memory + iAddress);
}


// ---------------------------------------------------------------------------
//		 ROMBank::GetByte
// ---------------------------------------------------------------------------

uae_u32 ROMBank::GetByte (uaecptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kROMByteRead]++;
#endif

#if (VALIDATE_ROM_GET)
	if (gMemAccessFlags.fValidate_ROMGet && !ValidAddress (iAddress, sizeof (char)))
	{
		InvalidAccess (iAddress, sizeof (char), true);
	}
#endif

#if (PREVENT_USER_ROM_GET || PREVENT_SYSTEM_ROM_GET)
	if (IsPCInRAM ())
	{
		if (PREVENT_USER_ROM_GET && gMemAccessFlags.fProtect_ROMGet)
		{
			InvalidAccess (iAddress, sizeof (char), true);
		}
	}
	else
	{
		if (PREVENT_SYSTEM_ROM_GET && gMemAccessFlags.fProtect_SysROMGet)
		{
			InvalidAccess (iAddress, sizeof (char), true);
		}
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementRead (1, WAITSTATES_ROM);
	}
#endif

	iAddress &= gROMBank_Mask;

	return do_get_mem_byte (gROM_Memory + iAddress);
}


// ---------------------------------------------------------------------------
//		 ROMBank::SetLong
// ---------------------------------------------------------------------------

void ROMBank::SetLong (uaecptr iAddress, uae_u32 iLongValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kROMLongWrite]++;
	if (iAddress & 2)
		gMemoryAccess[kROMLongWrite2]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

	assert (ValidAddress (iAddress, sizeof (long)));

#if (PREVENT_USER_ROM_SET || PREVENT_SYSTEM_ROM_SET)
	if (IsPCInRAM ())
	{
		if (PREVENT_USER_ROM_SET && gMemAccessFlags.fProtect_ROMSet)
		{
			InvalidAccess (iAddress, sizeof (long), false);
		}
	}
	else
	{
		if (PREVENT_SYSTEM_ROM_SET && gMemAccessFlags.fProtect_SysROMSet)
		{
			InvalidAccess (iAddress, sizeof (long), false);
		}
	}
#endif

#if HAS_PROFILING
	// writing to ROM?? don't know what to make of this
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementWrite (2, WAITSTATES_ROM);
	}
#endif

	iAddress &= gROMBank_Mask;

	do_put_mem_long (gROM_Memory + iAddress, iLongValue);
}


// ---------------------------------------------------------------------------
//		 ROMBank::SetWord
// ---------------------------------------------------------------------------

void ROMBank::SetWord (uaecptr iAddress, uae_u32 iWordValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kROMWordWrite]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

	assert (ValidAddress (iAddress, sizeof (short)));

#if (PREVENT_USER_ROM_SET || PREVENT_SYSTEM_ROM_SET)
	if (IsPCInRAM ())
	{
		if (PREVENT_USER_ROM_SET && gMemAccessFlags.fProtect_ROMSet)
		{
			InvalidAccess (iAddress, sizeof (short), false);
		}
	}
	else
	{
		if (PREVENT_SYSTEM_ROM_SET && gMemAccessFlags.fProtect_SysROMSet)
		{
			InvalidAccess (iAddress, sizeof (short), false);
		}
	}
#endif

#if HAS_PROFILING
	// writing to ROM?? don't know what to make of this
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementWrite (1, WAITSTATES_ROM);
	}
#endif

	iAddress &= gROMBank_Mask;

	do_put_mem_word (gROM_Memory + iAddress, iWordValue);
}


// ---------------------------------------------------------------------------
//		 ROMBank::SetByte
// ---------------------------------------------------------------------------

void ROMBank::SetByte (uaecptr iAddress, uae_u32 iByteValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kROMByteWrite]++;
#endif

	assert (ValidAddress (iAddress, sizeof (char)));
	
#if (PREVENT_USER_ROM_SET || PREVENT_SYSTEM_ROM_SET)
	if (IsPCInRAM ())
	{
		if (PREVENT_USER_ROM_SET && gMemAccessFlags.fProtect_ROMSet)
		{
			InvalidAccess (iAddress, sizeof (char), false);
		}
	}
	else
	{
		if (PREVENT_SYSTEM_ROM_SET && gMemAccessFlags.fProtect_SysROMSet)
		{
			InvalidAccess (iAddress, sizeof (char), false);
		}
	}
#endif

#if HAS_PROFILING
	// writing to ROM?? don't know what to make of this
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementWrite (1, WAITSTATES_ROM);
	}
#endif

	iAddress &= gROMBank_Mask;

	do_put_mem_byte (gROM_Memory + iAddress, iByteValue);
}


// ---------------------------------------------------------------------------
//		 ROMBank::ValidAddress
// ---------------------------------------------------------------------------

int ROMBank::ValidAddress (uaecptr iAddress, uae_u32 iSize)
{
	// Make the address a zero offset into ROM memory.

	iAddress -= kMemoryStart & gROMBank_Mask;

	// Strip the uppermost bit of the address.

	iAddress &= gROMBank_Mask;

	int	result = (iAddress + iSize) <= gROMBank_Size;

	assert (result);

	return result;
}


// ---------------------------------------------------------------------------
//		 ROMBank::GetRealAddress
// ---------------------------------------------------------------------------

uae_u8* ROMBank::GetRealAddress (uaecptr iAddress)
{
	// Make the address a zero offset into ROM memory.

	iAddress -= kMemoryStart & gROMBank_Mask;

	// Strip the uppermost bit of the address.

	iAddress &= gROMBank_Mask;

	return (uae_u8*) &gROM_Memory[iAddress];
}


// ---------------------------------------------------------------------------
//		 ROMBank::InvalidAccess
// ---------------------------------------------------------------------------

void ROMBank::InvalidAccess (uaecptr iAddress, long size, Bool forRead)
{
	Software::BusError ();
}




// ===========================================================================
//		 Flash Bank Accessors
// ===========================================================================
// These functions provide fetch and store access to the emulator's read only
// memory.

/*
	There are five types of flash that our flash manager routines identify:
	Mitsubishi, Hitachi, Intel, AMD, and Fujitsu.  Of these, only the last
	two are really supported right now.  Unfortunately, they are the hardest
	to emulate.  :-(

	To identify the kind of flash being used, our ROM routines:

		- Read a word from FLASHBASE
		- Write 0x00FF to FLASHBASE
		- Read a word from FLASHBASE
		- Write 0x0090 to FLASHBASE
		- Read a manufacturer ID (word) from FLASHBASE
		- Read a device ID (word) from FLASHBASE + 2
		- Write 0x00FF to FLASHBASE

	Mitsubishi:	manufacturer == 0x001C, device == 0x005E
	Hitachi:	manufacturer == 0x0007, device == 0x0086
	Intel:		manufacturer == 0x0089, device == 0x0091

	If the flash is not one of those, our ROM routines:

		- Read a word from FLASHBASE
		- Write 0x00F0 to FLASHBASE
		- Write 0x00AA to FLASHBASE + 0xAAAA
		- Write 0x0055 to FLASHBASE + 0x5554
		- Write 0x0090 to FLASHBASE + 0xAAAA
		- Read a manufacturer ID (word) from FLASHBASE
		- Read a device ID (word) from FLASHBASE + 2
		- Write 0x00F0 to FLASHBASE

	AMD:		manufacturer == 0x0001, device == 0x0049
	Fujitsu:	manufacturer == 0x0004, device == 0x0049


	To erase a word of flash, our ROM routines:

		AMD, Fujitsu:

		- Read a word from FLASHBASE
		- Write 0x00F0 to FLASHBASE
		- Write 0x00AA to FLASHBASE + 0xAAAA
		- Write 0x0055 to FLASHBASE + 0x5554
		- Write 0x0080 to FLASHBASE + 0xAAAA
		- Write 0x00AA to FLASHBASE + 0xAAAA
		- Write 0x0055 to FLASHBASE + 0x5554
		- Write 0x0030 to location to be erased
		- Check erase location for 0x0080
		- Read from FLASHBASE
		- Write 0x00F0 to FLASHBASE

		Mitsubishi, Hitachi:

		- Read a word from FLASHBASE
		- Write 0x00FF to FLASHBASE
		- Write 0x0020 to FLASHBASE
		- Write 0x00D0 to location to be erased
		- Check erase location for 0x0080
			-- If 0x0020 is also set, an error occured
		- Read from FLASHBASE
		- Write 0x00FF to FLASHBASE

		Intel

		- Not supported

	To program a block of flash:

		AMD, Fujitsu:

		- Read a word from FLASHBASE
		- Write 0x00F0 to FLASHBASE
		- For each word to write
			- If the word is already there, continue
			- Write 0x00AA to FLASHBASE + 0xAAAA
			- Write 0x0055 to FLASHBASE + 0x5554
			- Write 0x00A0 to FLASHBASE + 0xAAAA
			- Write the word to the dest location
			- Check write location for 0x0080
		- Read from FLASHBASE
		- Write 0x00F0 to FLASHBASE

		Mitsubishi, Hitachi, Intel:

		- Not supported
*/


enum
{
	kAMDState_Normal,		// Read-only mode. Acts like a normal ROM.
							// Almost any write of 0x00F0 to any address will get us here.
							// Also enter here automatically after Erase or Program operation.
							// A write of 0x00AA to 0xAAAA will put us in kAMDState_Unlocked1.

	kAMDState_Unlocked1,	// After first unlock cycle (0x00AA written to 0xAAAA)
							// Looking for the second unlock cycle (0x0055 written to 0x5554);
							// If we find one, go to kAMDState_Unlocked2.
							// ??? What happens on other operations?

	kAMDState_Unlocked2,	// After second unlock cycle (0x0055 written to 0x5554)
							// Now looking for a command to be written to 0xAAAA
							// If we find 0x0090, go to kAMDState_Autoselect.
							// If we find 0x0080, set fgEraseIsSetup and go to kAMDState_Normal. ??? When should fgEraseIsSetup get cleared?
							// If we find 0x0030, if fgEraseIsSetup erase the sector and go to kAMDState_EraseDone.
							// If we fine 0x00A0, go to kAMDState_Program.
							// ??? What happens on other operations?

	kAMDState_Autoselect,	// After a write of 0x0090 to 0x5555.
							// A read of 0x0000 returns 0x0001 (manufacturer ID).
							// A read of 0x0002 returns 0x0049 (device ID).
							// A write of 0x00F0 to any address returns us to kAMDState_Normal.
							// ??? What happens on other operations?

	kAMDState_Program,		// After a program sequence was entered.
							// Accept any write operation and go to kAMDState_ProgramDone.
							// ??? What happens on other operations?

	kAMDState_EraseDone,	// After a Erase operation.
							// On the next read, return successful operation and return to kAMDState_Normal
							// ??? What happens on other operations?

	kAMDState_ProgramDone,	// After a Program operation.
							// On the next read, return successful operation and return to kAMDState_Normal
							// ??? What happens on other operations?
	kAMDState_Dummy
};


AddressBank	FlashBank::fgAddressBank =
{
	FlashBank::GetLong,
	FlashBank::GetWord,
	FlashBank::GetByte,
	FlashBank::SetLong,
	FlashBank::SetWord,
	FlashBank::SetByte,
	FlashBank::GetRealAddress,
	FlashBank::ValidAddress
};

const uae_u32	FlashBank::kMemoryStart = 0x10C00000;
#define FLASHBASE	FlashBank::kMemoryStart

int		FlashBank::fgState = kAMDState_Normal;
uaecptr	FlashBank::fgAddressOfInterest;
Bool	FlashBank::fgEraseIsSetup;


// ---------------------------------------------------------------------------
//		 FlashBank::Initialize
// ---------------------------------------------------------------------------

void FlashBank::Initialize (void)
{
	// Memory banks 0x10C0 to <mumble> are managed by the functions
	// in FlashBank.  <mumble> is based on the amount of ROM being emulated.

	Memory::InitializeBanks (	fgAddressBank,
								bankindex (kMemoryStart),
								1);
}


// ---------------------------------------------------------------------------
//		 FlashBank::Reset
// ---------------------------------------------------------------------------

void FlashBank::Reset (void)
{
	fgState = kAMDState_Normal;
}


// ---------------------------------------------------------------------------
//		 FlashBank::GetLong
// ---------------------------------------------------------------------------

uae_u32 FlashBank::GetLong (uaecptr iAddress)
{
	return ROMBank::GetLong (iAddress);
}


// ---------------------------------------------------------------------------
//		 FlashBank::GetWord
// ---------------------------------------------------------------------------

uae_u32 FlashBank::GetWord (uaecptr iAddress)
{
	switch (fgState)
	{
		case kAMDState_Normal:
			// Read-only mode. Acts like a normal ROM.
			// Almost any write of 0x00F0 to any address will get us here.
			// Also enter here automatically after Erase or Program operation.
			// A write of 0x00AA to 0xAAAA will put us in kAMDState_Unlocked1.

			break;

		case kAMDState_Unlocked1:
			// After first unlock cycle (0x00AA written to 0xAAAA)
			// Looking for the second unlock cycle (0x0055 written to 0x5554);
			// If we find one, go to kAMDState_Unlocked2.
			// ??? What happens on other operations?

			break;

		case kAMDState_Unlocked2:
			// After second unlock cycle (0x0055 written to 0x5554)
			// Now looking for a command to be written to 0xAAAA
			// If we find 0x0090, go to kAMDState_Autoselect.
			// If we find 0x0080, set fgEraseIsSetup and go to kAMDState_Normal. ??? When should fgEraseIsSetup get cleared?
			// If we find 0x0030, if fgEraseIsSetup erase the sector and go to kAMDState_ProgramDone.
			// If we fine 0x00A0, go to kAMDState_Program.
			// ??? What happens on other operations?

			break;

		case kAMDState_Autoselect:
			// After a write of 0x0090 to 0x5555.
			// A read of 0x0000 returns 0x0001 (manufacturer ID).
			// A read of 0x0002 returns 0x0049 (device ID).
			// A write of 0x00F0 to any address returns us to kAMDState_Normal.
			// ??? What happens on other operations?

			if (iAddress == FLASHBASE)
			{
				return 0x0001;
			}
			else if (iAddress == FLASHBASE + 2)
			{
				return 0x0049;
			}
			break;

		case kAMDState_Program:
			// After a program sequence was entered.
			// Accept any write operation and go to kAMDState_ProgramDone.
			// ??? What happens on other operations?

			break;

		case kAMDState_EraseDone:
			// After a Program or Erase operation.
			// On the next read, return successful operation and return to kAMDState_Normal
			// ??? What happens on other operations?

			// !!! We should really check that iAddress is the same as the
			// one specified in the Erase command.

			fgState = kAMDState_Normal;
			return 0x0080;

		case kAMDState_ProgramDone:
			// After a Program or Erase operation.
			// On the next read, return successful operation and return to kAMDState_Normal
			// ??? What happens on other operations?

			// !!! We should really check that iAddress is the same as the
			// one specified in the Program command.

			fgState = kAMDState_Normal;
			return 0x0080 & ROMBank::GetWord (iAddress);
	}

	return ROMBank::GetWord (iAddress);
}


// ---------------------------------------------------------------------------
//		 FlashBank::GetByte
// ---------------------------------------------------------------------------

uae_u32 FlashBank::GetByte (uaecptr iAddress)
{
	return ROMBank::GetByte (iAddress);
}


// ---------------------------------------------------------------------------
//		 FlashBank::SetLong
// ---------------------------------------------------------------------------

void FlashBank::SetLong (uaecptr iAddress, uae_u32 iLongValue)
{
	ROMBank::SetLong (iAddress, iLongValue);
}


// ---------------------------------------------------------------------------
//		 FlashBank::SetWord
// ---------------------------------------------------------------------------

void FlashBank::SetWord (uaecptr iAddress, uae_u32 iWordValue)
{
	switch (fgState)
	{
		case kAMDState_Normal:
			// Read-only mode. Acts like a normal ROM.
			// Almost any write of 0x00F0 to any address will get us here.
			// Also enter here automatically after Erase or Program operation.
			// A write of 0x00AA to 0xAAAA will put us in kAMDState_Unlocked1.

			if (iAddress == FLASHBASE + 0xAAAA && iWordValue == 0x00AA)
			{
				fgState = kAMDState_Unlocked1;
				return;
			}
			break;

		case kAMDState_Unlocked1:
			// After first unlock cycle (0x00AA written to 0xAAAA)
			// Looking for the second unlock cycle (0x0055 written to 0x5554);
			// If we find one, go to kAMDState_Unlocked2.
			// ??? What happens on other operations?

			if (iWordValue == 0x00F0)
			{
				fgState = kAMDState_Normal;
				return;
			}
			else if (iAddress == FLASHBASE + 0x5554 && iWordValue == 0x0055)
			{
				fgState = kAMDState_Unlocked2;
				return;
			}
			break;

		case kAMDState_Unlocked2:
			// After second unlock cycle (0x0055 written to 0x5554)
			// Now looking for a command to be written to 0xAAAA
			// If we find 0x0090, go to kAMDState_Autoselect.
			// If we find 0x0080, set fgEraseIsSetup and go to kAMDState_Normal. ??? When should fgEraseIsSetup get cleared?
			// If we find 0x0030, if fgEraseIsSetup erase the sector and go to kAMDState_EraseDone.
			// If we fine 0x00A0, go to kAMDState_Program.
			// ??? What happens on other operations?

			if (iWordValue == 0x00F0)
			{
				fgState = kAMDState_Normal;
				return;
			}
			else if (iWordValue == 0x0030 && fgEraseIsSetup)
			{
				const int	kEraseValue			= 0xFF;
#if 0
				DWord	readWriteParmsOffset	= get_long (0x10C00000 + offsetof (CardHeaderType, readWriteParmsOffset));
				DWord	readWriteParmsSize		= get_long (0x10C00000 + offsetof (CardHeaderType, readWriteParmsSize));
				DWord	readOnlyParmsOffset		= get_long (0x10C00000 + offsetof (CardHeaderType, readOnlyParmsOffset));
				DWord	readOnlyParmsSize		= 0x00002000;
				DWord	readWriteWorkingOffset	= get_long (0x10C00000 + offsetof (CardHeaderType, readWriteWorkingOffset));
				DWord	readWriteWorkingSize	= get_long (0x10C00000 + offsetof (CardHeaderType, readWriteWorkingSize));

				if (iAddress >= 0x10000000 + readWriteParmsOffset &&
					iAddress < 0x10000000 + readWriteParmsOffset + readWriteParmsSize)
				{
					uae_memset (0x10000000 + readWriteParmsOffset, kEraseValue, readWriteParmsSize);
				}
				else if (iAddress >= 0x10000000 + readOnlyParmsOffset &&
					iAddress < 0x10000000 + readOnlyParmsOffset + readOnlyParmsSize)
				{
					uae_memset (0x10000000 + readOnlyParmsOffset, kEraseValue, readOnlyParmsSize);
				}
				else if (iAddress >= 0x10000000 + readWriteWorkingOffset &&
					iAddress < 0x10000000 + readWriteWorkingOffset + readWriteWorkingSize)
				{
					uae_memset (0x10000000 + readWriteWorkingOffset, kEraseValue, readWriteWorkingSize);
				}
#endif

				const unsigned long	kSector1Start	= 0x10C00000;
				const unsigned long	kSector1Size	= 0x00004000;
				const unsigned long	kSector2Start	= 0x10C04000;
				const unsigned long	kSector2Size	= 0x00002000;
				const unsigned long	kSector3Start	= 0x10C06000;
				const unsigned long	kSector3Size	= 0x00002000;
				const unsigned long	kSector4Start	= 0x10C08000;
				const unsigned long	kSector4Size	= 0x00008000;

				if (iAddress == kSector1Start)
				{
					uae_memset (kSector1Start, kEraseValue, kSector1Size);
				}
				else if (iAddress == kSector2Start)
				{
					uae_memset (kSector2Start, kEraseValue, kSector2Size);
				}
				else if (iAddress == kSector3Start)
				{
					uae_memset (kSector3Start, kEraseValue, kSector3Size);
				}
				else if (iAddress == kSector4Start)
				{
					uae_memset (kSector4Start, kEraseValue, kSector4Size);
				}

				fgState = kAMDState_EraseDone;
				return;
			}
			else if (iAddress == FLASHBASE + 0xAAAA)
			{
				if (iWordValue == 0x0090)
				{
					fgState = kAMDState_Autoselect;
					return;
				}
				else if (iWordValue == 0x0080)
				{
					fgEraseIsSetup = true;
					fgState = kAMDState_Normal;
					return;
				}
				else if (iWordValue == 0x00A0)
				{
					fgState = kAMDState_Program;
					return;
				}
			}
			break;

		case kAMDState_Autoselect:
			// After a write of 0x0090 to 0x5555.
			// A read of 0x0000 returns 0x0001 (manufacturer ID).
			// A read of 0x0002 returns 0x0049 (device ID).
			// A write of 0x00F0 to any address returns us to kAMDState_Normal.
			// ??? What happens on other operations?

			if (iWordValue == 0x00F0)
			{
				fgState = kAMDState_Normal;
				return;
			}
			break;

		case kAMDState_Program:
			// After a program sequence was entered.
			// Accept any write operation and go to kAMDState_ProgramDone.
			// ??? What happens on other operations?

			iAddress &= gROMBank_Mask;
			do_put_mem_word (gROM_Memory + iAddress, iWordValue);

			fgState = kAMDState_ProgramDone;
			break;

		case kAMDState_EraseDone:
			// After a Program or Erase operation.
			// On the next read, return successful operation and return to kAMDState_Normal
			// ??? What happens on other operations?

			break;

		case kAMDState_ProgramDone:
			// After a Program or Erase operation.
			// On the next read, return successful operation and return to kAMDState_Normal
			// ??? What happens on other operations?

			break;
	}

	ROMBank::SetWord (iAddress, iWordValue);
}


// ---------------------------------------------------------------------------
//		 FlashBank::SetByte
// ---------------------------------------------------------------------------

void FlashBank::SetByte (uaecptr iAddress, uae_u32 iByteValue)
{
	ROMBank::SetByte (iAddress, iByteValue);
}


// ---------------------------------------------------------------------------
//		 FlashBank::ValidAddress
// ---------------------------------------------------------------------------

int FlashBank::ValidAddress (uaecptr iAddress, uae_u32 iSize)
{
	return ROMBank::ValidAddress (iAddress, iSize);
}


// ---------------------------------------------------------------------------
//		 FlashBank::GetRealAddress
// ---------------------------------------------------------------------------

uae_u8* FlashBank::GetRealAddress (uaecptr iAddress)
{
	return ROMBank::GetRealAddress (iAddress);
}
