/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Bank_SRAM.h"

#include "CPU_REG.h"			// Software::BusError
#include "DebugMgr.h"			// Debug::CheckStepSpy
#include "RAM_ROM.h"			// gRAMBank_Size, gRAM_Memory, gMemoryAccess


// ===========================================================================
//		 SRAM Bank Accessors
// ===========================================================================
// These functions provide fetch and store access to the emulator's random
// access memory.

AddressBank	SRAMBank::fgAddressBank =
{
	SRAMBank::GetLong,
	SRAMBank::GetWord,
	SRAMBank::GetByte,
	SRAMBank::SetLong,
	SRAMBank::SetWord,
	SRAMBank::SetByte,
	SRAMBank::GetRealAddress,
	SRAMBank::ValidAddress
};

	// Note: I'd've used hwrCardBase0 here, except that that
	// changes on different hardware. It's 0x10000000 in some
	// cases, 0x00000000 in others.

const uae_u32	SRAMBank::kMemoryStart = 0x10000000;
const uae_u32	SRAMBank::kMemoryStartEZ = 0x00000000;
uae_u32			SRAMBank::fgMemoryStart = SRAMBank::kMemoryStart;


// ---------------------------------------------------------------------------
//		 SRAMBank::Initialize
// ---------------------------------------------------------------------------

void SRAMBank::Initialize (void)
{
	int	device = Emulator::GetHardwareDevice ();
	if (Emulator::EZMode ())
		fgMemoryStart = kMemoryStartEZ;
	else
		fgMemoryStart = kMemoryStart;

	// Memory banks 0x1000 to <mumble> are managed byt the functions
	// in SRAMBank.  <mumble> is based on the amount of RAM being emulated.

	Memory::InitializeBanks (	fgAddressBank,
								bankindex (fgMemoryStart),
								bankindex (fgMemoryStart + gRAMBank_Size - 1) -
									bankindex (fgMemoryStart) + 1);
}


// ---------------------------------------------------------------------------
//		 SRAMBank::Reset
// ---------------------------------------------------------------------------

void SRAMBank::Reset (void)
{
}


// ---------------------------------------------------------------------------
//		 SRAMBank::GetLong
// ---------------------------------------------------------------------------

uae_u32 SRAMBank::GetLong (uaecptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSRAMLongRead]++;
	if (iAddress & 2)
		gMemoryAccess[kSRAMLongRead2]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_SRAM_GET)
	if (gMemAccessFlags.fProtect_SRAMGet)
	{
		ProtectedAccess (iAddress, sizeof (long), true);
	}
#endif

#if (VALIDATE_SRAM_GET)
	if (gMemAccessFlags.fValidate_SRAMGet && !ValidAddress (iAddress, sizeof (long)))
	{
		InvalidAccess (iAddress, sizeof (long), true);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementRead (2, WAITSTATES_SRAM);
	}
#endif

	iAddress &= gRAMBank_Mask;

	return do_get_mem_long (gRAM_Memory + iAddress);
}


// ---------------------------------------------------------------------------
//		 SRAMBank::GetWord
// ---------------------------------------------------------------------------

uae_u32 SRAMBank::GetWord (uaecptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSRAMWordRead]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_SRAM_GET)
	if (gMemAccessFlags.fProtect_SRAMGet)
	{
		ProtectedAccess (iAddress, sizeof (short), true);
	}
#endif

#if (VALIDATE_SRAM_GET)
	if (gMemAccessFlags.fValidate_SRAMGet && !ValidAddress (iAddress, sizeof (short)))
	{
		InvalidAccess (iAddress, sizeof (short), true);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementRead (1, WAITSTATES_SRAM);
	}
#endif

	iAddress &= gRAMBank_Mask;

	return do_get_mem_word (gRAM_Memory + iAddress);
}


// ---------------------------------------------------------------------------
//		 SRAMBank::GetByte
// ---------------------------------------------------------------------------

uae_u32 SRAMBank::GetByte (uaecptr iAddress)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSRAMByteRead]++;
#endif

#if (PREVENT_USER_SRAM_GET)
	if (gMemAccessFlags.fProtect_SRAMGet)
	{
		ProtectedAccess (iAddress, sizeof (char), true);
	}
#endif

#if (VALIDATE_SRAM_GET)
	if (gMemAccessFlags.fValidate_SRAMGet && !ValidAddress (iAddress, sizeof (char)))
	{
		InvalidAccess (iAddress, sizeof (char), true);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementRead (1, WAITSTATES_SRAM);
	}
#endif

	iAddress &= gRAMBank_Mask;

	return do_get_mem_byte (gRAM_Memory + iAddress);
}


// ---------------------------------------------------------------------------
//		 SRAMBank::SetLong
// ---------------------------------------------------------------------------

void SRAMBank::SetLong (uaecptr iAddress, uae_u32 iLongValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSRAMLongWrite]++;
	if (iAddress & 2)
		gMemoryAccess[kSRAMLongWrite2]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_SRAM_SET)
	if (gMemAccessFlags.fProtect_SRAMSet)
	{
		ProtectedAccess (iAddress, sizeof (long), false);
	}
#endif

#if (VALIDATE_SRAM_SET)
	if (gMemAccessFlags.fValidate_SRAMSet && !ValidAddress (iAddress, sizeof (long)))
	{
		InvalidAccess (iAddress, sizeof (long), false);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementWrite (2, WAITSTATES_SRAM);
	}
#endif

	uaecptr	phyAddress = iAddress;
	phyAddress &= gRAMBank_Mask;

	do_put_mem_long (gRAM_Memory + phyAddress, iLongValue);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 4);
}


// ---------------------------------------------------------------------------
//		 SRAMBank::SetWord
// ---------------------------------------------------------------------------

void SRAMBank::SetWord (uaecptr iAddress, uae_u32 iWordValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSRAMWordWrite]++;
#endif

#if (CHECK_FOR_ADDRESS_ERROR)
	if ((iAddress & 1) != 0)
	{
		Software::AddressError ();
	}
#endif

#if (PREVENT_USER_SRAM_SET)
	if (gMemAccessFlags.fProtect_SRAMSet)
	{
		ProtectedAccess (iAddress, sizeof (short), false);
	}
#endif

#if (VALIDATE_SRAM_SET)
	if (gMemAccessFlags.fValidate_SRAMSet && !ValidAddress (iAddress, sizeof (short)))
	{
		InvalidAccess (iAddress, sizeof (short), false);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementWrite (1, WAITSTATES_SRAM);
	}
#endif

	uaecptr	phyAddress = iAddress;
	phyAddress &= gRAMBank_Mask;

	do_put_mem_word (gRAM_Memory + phyAddress, iWordValue);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 2);
}


// ---------------------------------------------------------------------------
//		 SRAMBank::SetByte
// ---------------------------------------------------------------------------

void SRAMBank::SetByte (uaecptr iAddress, uae_u32 iByteValue)
{
#if PROFILE_MEMORY
	gMemoryAccess[kSRAMByteWrite]++;
#endif

#if (PREVENT_USER_SRAM_SET)
	if (gMemAccessFlags.fProtect_SRAMSet)
	{
		ProtectedAccess (iAddress, sizeof (char), false);
	}
#endif

#if (VALIDATE_SRAM_SET)
	if (gMemAccessFlags.fValidate_SRAMSet && !ValidAddress (iAddress, sizeof (char)))
	{
		InvalidAccess (iAddress, sizeof (char), false);
	}
#endif

#if HAS_PROFILING
	if (gProfilingEnabled && gProfilingCounted)
	{
		ProfileIncrementWrite (1, WAITSTATES_SRAM);
	}
#endif

	uaecptr	phyAddress = iAddress;
	phyAddress &= gRAMBank_Mask;

	do_put_mem_byte (gRAM_Memory + phyAddress, iByteValue);

	// See if any interesting memory locations have changed.  If so,
	// CheckStepSpy will change gBreakReason, which we'll check later.

	Debug::CheckStepSpy (iAddress, 1);
}


// ---------------------------------------------------------------------------
//		 SRAMBank::ValidAddress
// ---------------------------------------------------------------------------

int SRAMBank::ValidAddress (uaecptr iAddress, uae_u32 iSize)
{
	int	result = false;

	// Check for a valid RAM access.

	// Strip the uppermost bit of the address.

	iAddress &= gRAMBank_Mask;

	result = (iAddress + iSize) <= gRAMBank_Size;

	assert (result);

	return result;
}


// ---------------------------------------------------------------------------
//		 SRAMBank::GetRealAddress
// ---------------------------------------------------------------------------

uae_u8* SRAMBank::GetRealAddress (uaecptr iAddress)
{
	// Strip the uppermost bit of the address.

	iAddress &= gRAMBank_Mask;

	return (uae_u8*) &(gRAM_Memory[iAddress]);
}


// ---------------------------------------------------------------------------
//		 SRAMBank::InvalidAccess
// ---------------------------------------------------------------------------

void SRAMBank::InvalidAccess (uaecptr iAddress, long size, Bool forRead)
{
	Software::BusError ();
}


// ---------------------------------------------------------------------------
//		 SRAMBank::ProtectedAccess
// ---------------------------------------------------------------------------

void SRAMBank::ProtectedAccess (uaecptr iAddress, long size, Bool forRead)
{
	Software::BusError ();
}

