/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "MetaMemory.h"

#include "CPU_REG.h"			// LowMem_GetGlobal, HWRegisters, Software::ForgetStackChunk
#include "DebugMgr.h"			// Debug::GetRoutineName
#include "RAM_ROM.h"			// CEnableFullAccess
#include "TrapPatches.h"		// IsInSysBinarySearch, OSMajorMinorVersion


#pragma mark -


#define MASK(x) (((uaecptr) (x)) & 0x0FFFFFFF)

// Some handy macros lifted from MemoryPrvNew.h

#define	memHandleProtect(h)											\
			((VoidHand)((DWord)h | 0x80000000))
#define	memHandleUnProtect(h)										\
			((VoidHand)((DWord)h & 0x7FFFFFFF))


// ===========================================================================
//		 MetaMemory
// ===========================================================================

uae_u8*			MetaMemory::fgMetaMemory;
uae_u32			MetaMemory::fgMetaMemorySize;
Bool			MetaMemory::fgHeapsAreInitialized;


class FunctionRange
{
	public:
					FunctionRange	(const char* functionName);

		Bool		InRange			(uaecptr = m68k_getpc ());
		void		Reset			(void);

	private:
		void		GetRange		(uaecptr addr);

		const char*	fName;
		uaecptr		fBegin;
		uaecptr		fEnd;
};


	// Note: can't use the full name of PrvMisAlignedBackwardInnerBitBlt, as
	// Debug::GetRoutineName returns only the first 31 characters.

static FunctionRange	gcj_kptkdelete						("cj_kptkdelete");
static FunctionRange	gBackspaceChar						("BackspaceChar");
static FunctionRange	gCrc16CalcBlock						("Crc16CalcBlock");
static FunctionRange	gDmWrite							("DmWrite");
static FunctionRange	gFindShowResults					("FindShowResults");
static FunctionRange	gFindSaveFindStr					("FindSaveFindStr");
static FunctionRange	gFldDelete							("FldDelete");
static FunctionRange	gFntDefineFont						("FntDefineFont");
static FunctionRange	gGrfProcessStroke					("GrfProcessStroke");
static FunctionRange	gMenuHandleEvent					("MenuHandleEvent");
static FunctionRange	gNetPrvSettingSet					("NetPrvSettingSet");
static FunctionRange	gNetPrvTaskMain						("NetPrvTaskMain");
static FunctionRange	gPrvCompressedInnerBitBlt			("PrvCompressedInnerBitBlt");
static FunctionRange	gPrvFindMemoryLeaks					("PrvFindMemoryLeaks");
static FunctionRange	gPrvMisAlignedForwardInnerBitBlt	("PrvMisAlignedForwardInnerBitBlt");
static FunctionRange	gPrvMisAlignedBackwardInnerBitBlt	("PrvMisAlignedBackwardInnerBitBl");
static FunctionRange	gPrvSystemTimerProc					("PrvSystemTimerProc");
static FunctionRange	gSecPrvRandomSeed					("SecPrvRandomSeed");
static FunctionRange	gSysAppExit							("SysAppExit");


// ---------------------------------------------------------------------------
//		 MetaMemory::Initialize
// ---------------------------------------------------------------------------

void MetaMemory::Initialize (uae_u32 ramSize)
{
	fgMetaMemorySize = ramSize;
	fgMetaMemory = (uae_u8*) Platform::AllocateMemory (ramSize, false);

	gcj_kptkdelete						.Reset ();
	gBackspaceChar						.Reset ();
	gCrc16CalcBlock						.Reset ();
	gDmWrite							.Reset ();
	gFindShowResults					.Reset ();
	gFindSaveFindStr					.Reset ();
	gFldDelete							.Reset ();
	gFntDefineFont						.Reset ();
	gGrfProcessStroke					.Reset ();
	gMenuHandleEvent					.Reset ();
	gNetPrvSettingSet					.Reset ();
	gNetPrvTaskMain						.Reset ();
	gPrvCompressedInnerBitBlt			.Reset ();
	gPrvFindMemoryLeaks					.Reset ();
	gPrvMisAlignedForwardInnerBitBlt	.Reset ();
	gPrvMisAlignedBackwardInnerBitBlt	.Reset ();
	gPrvSystemTimerProc					.Reset ();
	gSecPrvRandomSeed					.Reset ();
	gSysAppExit							.Reset ();
}


// ---------------------------------------------------------------------------
//		 MetaMemory::Reset
// ---------------------------------------------------------------------------

void MetaMemory::Reset (void)
{
	// Mark everything as free-for-all.

	memset (fgMetaMemory, kFreeAccessBits, fgMetaMemorySize);

	fgHeapsAreInitialized = false;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::Dispose
// ---------------------------------------------------------------------------

void MetaMemory::Dispose (void)
{
	Platform::DisposeMemory (fgMetaMemory);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MarkLowMemory
// ---------------------------------------------------------------------------
//	Mark the "low memory" range of memory (the first 256 bytes of
//	memory that hold exception vectors).

void MetaMemory::MarkLowMemory (uaecptr begin, uaecptr end)
{
	MarkRange (begin, end, kLowMemoryBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::UnmarkLowMemory
// ---------------------------------------------------------------------------

void MetaMemory::UnmarkLowMemory (uaecptr begin, uaecptr end)
{
	UnmarkRange (begin, end, kLowMemoryBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MarkSystemGlobals
// ---------------------------------------------------------------------------
//	Mark the "system globals" range of memory.  This range holds the
//	global variables used by the Palm OS, as well as the jump table.

void MetaMemory::MarkSystemGlobals (uaecptr begin, uaecptr end)
{
	MarkRange (begin, end, kGlobalsBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::UnmarkSystemGlobals
// ---------------------------------------------------------------------------

void MetaMemory::UnmarkSystemGlobals (uaecptr begin, uaecptr end)
{
	UnmarkRange (begin, end, kGlobalsBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MarkHeapHeader
// ---------------------------------------------------------------------------

void MetaMemory::MarkHeapHeader (uaecptr begin, uaecptr end)
{
	MarkRange (begin, end, kMemStructBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::UnmarkHeapHeader
// ---------------------------------------------------------------------------

void MetaMemory::UnmarkHeapHeader (uaecptr begin, uaecptr end)
{
	UnmarkRange (begin, end, kMemStructBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MarkLowStack
// ---------------------------------------------------------------------------

void MetaMemory::MarkLowStack (uaecptr begin, uaecptr end)
{
	MarkRange (begin, end, kLowStackBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::UnmarkLowStack
// ---------------------------------------------------------------------------

void MetaMemory::UnmarkLowStack (uaecptr begin, uaecptr end)
{
	UnmarkRange (begin, end, kLowStackBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MarkScreen
// ---------------------------------------------------------------------------

void MetaMemory::MarkScreen (uaecptr begin, uaecptr end)
{
	MarkRange (begin, end, kScreenBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::UnmarkScreen
// ---------------------------------------------------------------------------

void MetaMemory::UnmarkScreen (uaecptr begin, uaecptr end)
{
	UnmarkRange (begin, end, kScreenBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MarkUnlockedChunk
// ---------------------------------------------------------------------------

void MetaMemory::MarkUnlockedChunk (uaecptr begin, uaecptr end)
{
	MarkRange (begin, end, kUnlockedChunkBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::UnmarkUnlockedChunk
// ---------------------------------------------------------------------------

void MetaMemory::UnmarkUnlockedChunk (uaecptr begin, uaecptr end)
{
	UnmarkRange (begin, end, kUnlockedChunkBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MarkFreeChunk
// ---------------------------------------------------------------------------

void MetaMemory::MarkFreeChunk (uaecptr begin, uaecptr end)
{
	MarkRange (begin, end, kFreeChunkBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::UnmarkFreeChunk
// ---------------------------------------------------------------------------

void MetaMemory::UnmarkFreeChunk (uaecptr begin, uaecptr end)
{
	UnmarkRange (begin, end, kFreeChunkBits);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MarkUninitialized
// ---------------------------------------------------------------------------

void MetaMemory::MarkUninitialized (uaecptr begin, uaecptr end)
{
#if FOR_LATER
	MarkRange (begin, end, kUninitialized);
#endif
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MoveUninitialized
// ---------------------------------------------------------------------------
//	Transfer the initialized/uninitialized state of a range of bytes to a
//	new location.  Called during MemMove, so that the state of the bytes
//	being moved can be preserved.

void MetaMemory::MoveUninitialized (uaecptr source, uaecptr dest, uae_u32 size)
{
#if FOR_LATER
	source = MASK (source);
	dest = MASK (dest);

	if (source > fgMetaMemorySize)
		return;

	if (dest > fgMetaMemorySize)
		return;

	uae_u8*	p0 = fgMetaMemory + source;
	uae_u8*	p1 = fgMetaMemory + dest;

	while (size--)
	{
		uae_u8	dValue = *p0;
		uae_u8	sValue = *p1;

		dValue = (dValue & ~kUninitialized) | (sValue & kUninitialized);

		*p0 = dValue;

		++p0;
		++p1;
	}
#endif
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MarkInitialized
// ---------------------------------------------------------------------------

void MetaMemory::MarkInitialized (uaecptr begin, uaecptr end)
{
#if FOR_LATER
	UnmarkRange (begin, end, kUninitialized);
#endif
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemChunkNew
// ---------------------------------------------------------------------------
//	Mark all the newly allocated bytes as uninitialized.

void MetaMemory::MemChunkNew (VoidPtr p, UInt attributes)
{
	// Sync up with new free space and relocated handles.

	SyncWithHeap (0, false);

	// Mark the newly created block.
	// We don't need to do this if the block is relocatable, as
	// those blocks are marked in SyncWithHeap.

	if ((attributes & memNewChunkFlagNonMovable) && ValidChunk (p))
	{
		DWord	size = GetChunkSize (p);
		uae_u8*	ptr = fgMetaMemory + MASK(p);

		memset (ptr, kFreeAccessBits, size);
	}

#if FOR_LATER
	if (attributes & memNewChunkFlagNonMovable)
	{
		if (ValidChunk (p))
		{
			MarkUninitialized (((uaecptr) p), ((uaecptr) p) + GetChunkSize (p));
		}
	}
	else
	{
		if (ValidHandle ((VoidHand) p))
		{
			p = PalmHeap::GetPointer ((VoidHand) p);
			if (p)
			{
				MarkUninitialized (((uaecptr) p), ((uaecptr) p) + GetChunkSize (p));
			}
		}
	}
#endif
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemChunkFree
// ---------------------------------------------------------------------------
//	Mark all the newly deallocated bytes as being in a free chunk.

void MetaMemory::MemChunkFree (VoidPtr p)
{
	if (ValidChunk (p))
	{
		MarkFreeChunk ((uaecptr) p, ((uaecptr) p) + GetChunkSize (p));

		Software::ForgetStackChunk ((uaecptr) p);

		// In case any following free blocks were coalesced with this one,
		// remove any "memory manager only" access bits corresponding to
		// chunk headers that no longer exist.

//		UnmarkRange ((uaecptr) p, ((uaecptr) p) + GetChunkSize (p), kMemStructBits);
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemPtrNew
// ---------------------------------------------------------------------------
//	Mark all the newly allocated bytes as uninitialized.

void MetaMemory::MemPtrNew (VoidPtr p)
{
	// Sync up with new free space and relocated handles.

	SyncWithHeap (0, false);

	// Mark the newly created block.

	if (ValidChunk (p))
	{
		DWord	size = GetChunkSize (p);
		uae_u8*	ptr = fgMetaMemory + MASK(p);

		memset (ptr, kFreeAccessBits, size);
	}

#if FOR_LATER
	if (ValidChunk (p))
	{
		MarkUninitialized (((uaecptr) p), ((uaecptr) p) + GetChunkSize (p));
	}
#endif
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemPtrResize
// ---------------------------------------------------------------------------
//	Mark all the newly allocated bytes as uninitialized.

void MetaMemory::MemPtrResize (VoidPtr p, UInt oldSize)
{
	// Sync up with new free space and relocated handles.

	SyncWithHeap (0, false);

	// Mark the newly resized block.

	if (ValidChunk (p))
	{
		DWord	newSize = GetChunkSize (p);

		if (newSize > oldSize)
		{
			uae_u8*	ptr = fgMetaMemory + MASK(p);
			memset (ptr + oldSize, kFreeAccessBits, newSize - oldSize);
		}
	}

#if FOR_LATER
	if (ValidChunk (p))
	{
		UInt	newSize = GetChunkSize (p);
		if (newSize > oldSize)
		{
			MarkUninitialized (((uaecptr) p), ((uaecptr) p) + newSize);
		}
	}
#endif
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemHandleNew
// ---------------------------------------------------------------------------
//	Mark all the newly allocated bytes as uninitialized.

void MetaMemory::MemHandleNew (VoidHand h)
{
	// Sync up with new free space and relocated handles.

	SyncWithHeap (0, false);

#if 0
	// Mark the newly created block.
	// (Actually, since this is a relocatable block, it was
	//  marked in SyncWithHeap).

	if (ValidHandle (h))
	{
		VoidPtr	p = PalmHeap::GetPointer (h);
		if (p)
		{
			DWord	size = GetChunkSize (p);
			uae_u8*	ptr = fgMetaMemory + MASK(p);

			memset (ptr, kFreeAccessBits, size);
		}
	}
#endif

#if FOR_LATER
	if (ValidHandle (h))
	{
		VoidPtr	p = PalmHeap::GetPointer (h);
		if (p)
		{
			MarkUninitialized (((uaecptr) p), ((uaecptr) p) + GetChunkSize (p));
		}
	}
#endif
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemHandleResize
// ---------------------------------------------------------------------------
//	Mark all the newly allocated bytes as uninitialized.

void MetaMemory::MemHandleResize (VoidHand h, UInt oldSize)
{
	// Sync up with new free space and relocated handles.

	SyncWithHeap (0, false);

	// Mark the newly resized block.
	// If the block was unlocked, it will have been marked in
	// SyncWithHeap, so we don't need to deal with those here.

	if (ValidHandle (h))
	{
		VoidPtr	p = PalmHeap::GetPointer (h);
		if (p && GetChunkLockCount (p) > 0)
		{
			DWord	newSize = GetChunkSize (p);

			if (newSize > oldSize)
			{
				uae_u8*	ptr = fgMetaMemory + MASK(p);
				memset (ptr, kFreeAccessBits, newSize);
			}
		}
	}

#if FOR_LATER
	if (ValidHandle (h))
	{
		VoidPtr	p = PalmHeap::GetPointer (h);
		if (p && GetChunkLockCount (p) > 0)
		{
			DWord	newSize = GetChunkSize (p);
			if (newSize > oldSize)
			{
				MarkUninitialized (((uaecptr) p) + oldSize, ((uaecptr) p) + newSize);
			}
		}
	}
#endif
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemHandleFree
// ---------------------------------------------------------------------------
//	Mark all the newly deallocated bytes as being in a free chunk.

void MetaMemory::MemHandleFree (VoidHand h)
{
	if (ValidHandle (h))
	{
		VoidPtr	p = PalmHeap::GetPointer (h);

		if (p)
		{
			MarkFreeChunk ((uaecptr) p, ((uaecptr) p) + GetChunkSize (p));

			// In case any following free blocks were coalesced with this one,
			// remove any "memory manager only" access bits corresponding to
			// chunk headers that no longer exist.

//			UnmarkRange ((uaecptr) p, ((uaecptr) p) + GetChunkSize (p), kMemStructBits);
		}
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemLocalIDToLockedPtr
// ---------------------------------------------------------------------------
//	Removed the "unlocked" access bit from the bytes in the chunk.

void MetaMemory::MemLocalIDToLockedPtr (VoidPtr p)
{
	if (ValidChunk (p))
	{
		UnmarkUnlockedChunk ((uaecptr) p, ((uaecptr) p) + GetChunkSize (p));
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemHandleLock
// ---------------------------------------------------------------------------
//	Removed the "unlocked" access bit from the bytes in the chunk.

void MetaMemory::MemHandleLock (VoidHand h)
{
	if (ValidHandle (h))
	{
		VoidPtr	p = PalmHeap::GetPointer (h);
		if (p)
		{
			UnmarkUnlockedChunk ((uaecptr) p, ((uaecptr) p) + GetChunkSize (p));
		}
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemHandleUnlock
// ---------------------------------------------------------------------------
//	If the lockCount is now zero, mark all the bytes in this chunk as being
//	in an unlocked chunk.

void MetaMemory::MemHandleUnlock (VoidHand h)
{
	if (ValidHandle (h))
	{
		VoidPtr	p = PalmHeap::GetPointer (h);
		if (p && GetChunkLockCount (p) == 0)
		{
			MarkUnlockedChunk ((uaecptr) p, ((uaecptr) p) + GetChunkSize (p));
		}
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemHandleResetLock
// ---------------------------------------------------------------------------
//	Mark all the bytes in this chunk as being in an unlocked chunk.

void MetaMemory::MemHandleResetLock (VoidHand h)
{
	if (ValidHandle (h))
	{
		VoidPtr	p = PalmHeap::GetPointer (h);
		if (p)
		{
			MarkUnlockedChunk ((uaecptr) p, ((uaecptr) p) + GetChunkSize (p));
		}
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemPtrResetLock
// ---------------------------------------------------------------------------
//	Mark all the bytes in this chunk as being in an unlocked chunk.

void MetaMemory::MemPtrResetLock (VoidPtr p)
{
	if (ValidChunk (p))
	{
		MarkUnlockedChunk ((uaecptr) p, ((uaecptr) p) + GetChunkSize (p));
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MemPtrUnlock
// ---------------------------------------------------------------------------
//	If the lockCount is now zero, mark all the bytes in this chunk as being
//	in an unlocked chunk.

void MetaMemory::MemPtrUnlock (VoidPtr p)
{
	if (ValidChunk (p))
	{
		if (GetChunkLockCount (p) == 0)
		{
			MarkUnlockedChunk ((uaecptr) p, ((uaecptr) p) + GetChunkSize (p));
		}
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::SyncWithHeap
// ---------------------------------------------------------------------------
//	Find chunk headers and mark them as memory manager structures.  Mark
//	unlocked relocatable chunks as "unlocked" and "initialized" (as we don't
//	have any idea any more what their state was before they were moved).
//	Mark free blocks as "uninitialized".

void MetaMemory::SyncWithHeap (UInt heapID, Bool fullSync)
{
	if (heapID == 0)
	{
		// Gather information about all the allocated, free, and
		// unlocked blocks.

		PalmHeap::WalkHeap (0, SyncOneBlock, &fullSync);

		// Mark the heap header itself, all the parts up to the master pointer table array.

		PalmHeap::HeapInfo	heapInfo;
		PalmHeap::GetHeapInfo (heapID, &heapInfo);

		PalmHeap::MPTInfo		mptInfo;
		PalmHeap::GetMPTInfo (heapInfo.heapHdrStart + heapInfo.heapHdrSize,
								heapInfo.version, &mptInfo);

		uae_u8*	ptr = fgMetaMemory + MASK (heapInfo.heapHdrStart);
		memset (ptr, kMemStructBits, heapInfo.heapHdrSize + mptInfo.mptHdrSize);

		// Mark the master pointer blocks as OK for the system.
		// A lot of the ROM uses a MemMgr macro to deref handles.

		while (1)
		{
			uaecptr	mptArray = mptInfo.mptHdrStart + mptInfo.mptHdrSize;
			uae_u8*	ptr = fgMetaMemory + MASK(mptArray);

			memset (ptr, kMPTBits, mptInfo.numEntries * sizeof (Ptr));

			if (mptInfo.nextTblOffset == 0)
				break;

			PalmHeap::GetMPTInfo (heapInfo.heapHdrStart + mptInfo.nextTblOffset,
									heapInfo.version, &mptInfo);
		}
	}

	if (fullSync)
	{
		fgHeapsAreInitialized = true;
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::GetWhatHappened
// ---------------------------------------------------------------------------
//	Some memory access violation was detected.  Nail down more closely what
//	it was.  These checks can be expensive, as they only occur when we're
//	pretty sure we're about to report an error in a dialog.

Errors::EAccessType MetaMemory::GetWhatHappened (uaecptr iAddress, long size, Bool forRead)
{
	// If  we've whisked away all memory access checking, return now.

	if (CEnableFullAccess::AccessOK ())
		return Errors::kOKAccess;


	CEnableFullAccess	munge;	// Remove blocks on memory access.

	Errors::EAccessType	whatHappened = Errors::kUnknownAccess;

	// Now figure out what just happened.

	if (MetaMemory::IsLowMemory (iAddress, size))
	{
		whatHappened = Errors::kLowMemAccess;
	}

	else if (MetaMemory::IsSystemGlobal (iAddress, size))
	{
		whatHappened = Errors::kGlobalVarAccess;
	}

	else if (MetaMemory::IsScreenBuffer (iAddress, size))
	{
		whatHappened = Errors::kScreenAccess;
	}

	else if (MetaMemory::IsLowStack (iAddress, size))
	{
		whatHappened = Errors::kLowStackAccess;
	}

	else
	{
		// See if we touched a memory manager structure.
		// Let's start with the headers.

		UInt	numHeaps = PalmHeap::GetNumHeaps (0);
		for (UInt heapID = 0; heapID < numHeaps; ++heapID)
		{
			// Is in the heap header?

			PalmHeap::HeapInfo	heapInfo;
			PalmHeap::GetHeapInfo (heapID, &heapInfo);

			if (iAddress >= heapInfo.heapHdrStart && iAddress < heapInfo.firstChunk)
			{
				whatHappened = Errors::kMemMgrAccess;
				goto Exit;
			}

			// Is it in any of the master pointer tables?

			PalmHeap::MPTInfo		mptInfo;
			PalmHeap::GetMPTInfo (heapInfo.heapHdrStart + heapInfo.heapHdrSize, heapInfo.version, &mptInfo);

			while (1)
			{
				uaecptr	mptHdrStart = mptInfo.mptHdrStart;
				uaecptr	mptEnd = mptInfo.mptHdrStart + mptInfo.mptHdrSize + mptInfo.numEntries * sizeof (Ptr);

				if (iAddress >= mptHdrStart && iAddress < mptEnd)
				{
					whatHappened = Errors::kMemMgrAccess;
					goto Exit;
				}

				if (mptInfo.nextTblOffset == 0)
					break;

				PalmHeap::GetMPTInfo (heapInfo.heapHdrStart + mptInfo.nextTblOffset, heapInfo.version, &mptInfo);
			}
		}

		// Check against all of the memory manager chunks.

		WhatHappenedData	info;
		info.result			= Errors::kUnknownAccess;
		info.address		= iAddress;
		info.size			= size;
		info.forRead		= forRead;

		PalmHeap::WalkHeap (0, WhatHappenedCallback, &info);

		if (info.result != Errors::kUnknownAccess)
		{
			whatHappened = info.result;
		}
	}

#if 0
	else if (MetaMemory::IsUninitialized (iAddress, size))
	{
		if (MetaMemory::IsStack (iAddress, size))
		{
			whatHappened = Errors::kUninitializedStack;
		}

		else if (MetaMemory::IsAllocatedChunk (iAddress, size))
		{
			whatHappened = Errors::kUninitializedChunk;
		}
	}
#endif

Exit:
	whatHappened = AllowForBugs (iAddress, size, forRead, whatHappened);

	return whatHappened;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::AllowForBugs
// ---------------------------------------------------------------------------

Errors::EAccessType MetaMemory::AllowForBugs (uaecptr iAddress, long size, Bool forRead, Errors::EAccessType whatHappened)
{
	if (whatHappened != Errors::kOKAccess && forRead)
	{
		// Let PrvFindMemoryLeaks have the run of the show.

		if (gPrvFindMemoryLeaks.InRange ())
		{
			return Errors::kOKAccess;
		}

		// SecPrvRandomSeed calls Crc16CalcBlock (NULL, 0x1000, 0).

		uaecptr	a6 = m68k_areg (regs, 6);
		uaecptr	rtnAddr = get_long (a6 + 4);

		if (iAddress < 0x1000 &&
			gCrc16CalcBlock.InRange () &&
			gSecPrvRandomSeed.InRange (rtnAddr))
		{
			return Errors::kOKAccess;
		}

		if (whatHappened == Errors::kLowMemAccess)
		{
			// See if the the low-memory checksummer is at work.

			if (size == 4 &&
				gPrvSystemTimerProc.InRange ())
			{
				return Errors::kOKAccess;
			}

			// There's a bug in BackspaceChar (pre-3.2) that accesses the word at 0x0000.

			if (Patches::HasBackspaceCharBug () &&
				iAddress == 0x0000 && size == 2 &&
				gBackspaceChar.InRange ())
			{
				return Errors::kOKAccess;
			}

			// There's a bug in FldDelete (pre-3.2) that accesses the word at 0x0000.

			if (Patches::HasFldDeleteBug () &&
				iAddress == 0x0000 && size == 2 &&
				gFldDelete.InRange ())
			{
				return Errors::kOKAccess;
			}

			// There's a bug in GrfProcessStroke (pre-3.1) that accesses the word at 0x0002.

			if (Patches::HasGrfProcessStrokeBug () &&
				iAddress == 0x0002 && size == 2 &&
				gGrfProcessStroke.InRange ())
			{
				return Errors::kOKAccess;
			}

			// There's a bug in NetPrvTaskMain (pre-3.2) that accesses low-memory.

			if (Patches::HasNetPrvTaskMainBug () &&
				gNetPrvTaskMain.InRange ())
			{
				return Errors::kOKAccess;
			}
		}

		//	There's a bug in pre-3.0 versions of SysBinarySearch that cause it to
		//	call the user callback function with a pointer just past the array
		//	to search.

		if (Patches::HasSysBinarySearchBug () &&
			Patches::IsInSysBinarySearch ())
		{
			return Errors::kOKAccess;
		}
	}

	return whatHappened;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::GetLowMemoryBegin
// ---------------------------------------------------------------------------

uaecptr MetaMemory::GetLowMemoryBegin (void)
{
	return 0;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::GetLowMemoryEnd
// ---------------------------------------------------------------------------

uaecptr MetaMemory::GetLowMemoryEnd (void)
{
	return offsetof (LowMemHdrType, globals);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::GetSysGlobalsBegin
// ---------------------------------------------------------------------------

uaecptr MetaMemory::GetSysGlobalsBegin (void)
{
	return offsetof (LowMemHdrType, globals);
}


// ---------------------------------------------------------------------------
//		 MetaMemory::GetSysGlobalsEnd
// ---------------------------------------------------------------------------

uaecptr MetaMemory::GetSysGlobalsEnd (void)
{
	CEnableFullAccess	munge;	// Remove blocks on memory access.

	return LowMem_GetGlobal (sysDispatchTableP) + 
			LowMem_GetGlobal (sysDispatchTableSize) * (sizeof (void*));
}


// ---------------------------------------------------------------------------
//		 MetaMemory::GetScreenBegin
// ---------------------------------------------------------------------------

uaecptr MetaMemory::GetScreenBegin (void)
{
	CEnableFullAccess	munge;	// Remove blocks on memory access.

	return HWRegisters::GetLCDStartAddr ();
}


// ---------------------------------------------------------------------------
//		 MetaMemory::GetScreenEnd
// ---------------------------------------------------------------------------

uaecptr MetaMemory::GetScreenEnd (void)
{
	CEnableFullAccess	munge;	// Remove blocks on memory access.

	// Get the screen's current width, height, depth, and buffer size.

	uaecptr	screenStart = HWRegisters::GetLCDStartAddr ();
	int		rowBytes	= HWRegisters::GetLCDRowBytes ();
	int		height		= HWRegisters::GetLCDHeight ();
	long	screenSize	= height * rowBytes;

	return screenStart + screenSize;
}

// ---------------------------------------------------------------------------
//		 MetaMemory::GetHeapHdrBegin
// ---------------------------------------------------------------------------

uaecptr MetaMemory::GetHeapHdrBegin (UInt heapID)
{
	PalmHeap::HeapInfo	info;
	PalmHeap::GetHeapInfo (heapID, &info);

	return info.heapHdrStart;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::GetHeapHdrEnd
// ---------------------------------------------------------------------------

uaecptr MetaMemory::GetHeapHdrEnd (UInt heapID)
{
	PalmHeap::HeapInfo	info;
	PalmHeap::GetHeapInfo (heapID, &info);

	return info.firstChunk;
}


Bool MetaMemory::IsLowMemory (uaecptr testAddress, uae_u32 size)
{
	testAddress = MASK (testAddress);
	if (testAddress >= fgMetaMemorySize)
		return false;

	return testAddress >= GetLowMemoryBegin () && testAddress < GetLowMemoryEnd ();
}


Bool MetaMemory::IsSystemGlobal (uaecptr testAddress, uae_u32 size)
{
	testAddress = MASK (testAddress);
	if (testAddress >= fgMetaMemorySize)
		return false;

	return testAddress >= GetSysGlobalsBegin () && testAddress < GetSysGlobalsEnd ();
}


#if 0	// Now inline
Bool MetaMemory::IsScreenBuffer (uaecptr testAddress, uae_u32 size)
{
	testAddress = MASK (testAddress);
	if (testAddress >= fgMetaMemorySize)
		return false;

	return testAddress >= GetScreenBegin () && testAddress < GetScreenEnd ();
}
#endif


Bool MetaMemory::IsMemMgrData (uaecptr testAddress, uae_u32 size)
{
	testAddress = MASK (testAddress);
	if (testAddress >= fgMetaMemorySize)
		return false;

	if (testAddress >= GetHeapHdrBegin (0) && testAddress < GetHeapHdrEnd (0))
		return true;

	ChunkCheck	check;
	check.testAddress	= testAddress;
	check.result		= false;

	PalmHeap::WalkHeap (0, CheckForChunkHeader, &check);

	return check.result;
}


Bool MetaMemory::IsUnlockedChunk (uaecptr testAddress, uae_u32 size)
{
	testAddress = MASK (testAddress);
	if (testAddress >= fgMetaMemorySize)
		return false;

	ChunkCheck	check;
	check.testAddress	= testAddress;
	check.result		= false;

	PalmHeap::WalkHeap (0, CheckForUnlockedChunk, &check);

	return check.result;
}


Bool MetaMemory::IsFreeChunk (uaecptr testAddress, uae_u32 size)
{
	testAddress = MASK (testAddress);
	if (testAddress >= fgMetaMemorySize)
		return false;

	ChunkCheck	check;
	check.testAddress	= testAddress;
	check.result		= false;

	PalmHeap::WalkHeap (0, CheckForFreeChunk, &check);

	return check.result;
}


Bool MetaMemory::IsUninitialized (uaecptr testAddress, uae_u32 size)
{
	testAddress = MASK (testAddress);
	if (testAddress >= fgMetaMemorySize)
		return false;

#if FOR_LATER
	return (fgMetaMemory [MASK(testAddress)] & kUninitialized) != 0;
#else
	return false;
#endif
}


Bool MetaMemory::IsStack (uaecptr testAddress, uae_u32 size)
{
	testAddress = MASK (testAddress);
	if (testAddress >= fgMetaMemorySize)
		return false;

	return false;
}


Bool MetaMemory::IsLowStack (uaecptr testAddress, uae_u32 size)
{
	testAddress = MASK (testAddress);
	if (testAddress >= fgMetaMemorySize)
		return false;

	return false;
}


Bool MetaMemory::IsAllocatedChunk (uaecptr testAddress, uae_u32 size)
{
	testAddress = MASK (testAddress);
	if (testAddress >= fgMetaMemorySize)
		return false;

	ChunkCheck	check;
	check.testAddress	= testAddress;
	check.result		= false;

	PalmHeap::WalkHeap (0, CheckForAllocatedChunk, &check);

	return check.result;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MarkRange
// ---------------------------------------------------------------------------

void MetaMemory::MarkRange (uaecptr start, uaecptr end, uae_u8 v)
{
	start = MASK (start);
	end = MASK (end);

	if (start >= fgMetaMemorySize)
		return;

	if (end >= fgMetaMemorySize)
		end = fgMetaMemorySize;

	uae_u8*	p0 = fgMetaMemory + start;
	uae_u8*	p1 = fgMetaMemory + end;
	uae_u8*	p;

	for (p = p0; p < p1; ++p)
	{
		*p |= v;
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::UnmarkRange
// ---------------------------------------------------------------------------

void MetaMemory::UnmarkRange (uaecptr start, uaecptr end, uae_u8 v)
{
	start = MASK (start);
	end = MASK (end);

	if (start >= fgMetaMemorySize)
		return;

	if (end >= fgMetaMemorySize)
		end = fgMetaMemorySize;

	uae_u8*	p0 = fgMetaMemory + start;
	uae_u8*	p1 = fgMetaMemory + end;
	uae_u8*	p;

	v = ~v;

	for (p = p0; p < p1; ++p)
	{
		*p &= v;
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::MarkUnmarkRange
// ---------------------------------------------------------------------------

void MetaMemory::MarkUnmarkRange (uaecptr start, uaecptr end,
							uae_u8 andValue, uae_u8 orValue)
{
	start = MASK (start);
	end = MASK (end);

	if (start >= fgMetaMemorySize)
		return;

	if (end >= fgMetaMemorySize)
		end = fgMetaMemorySize;

	uae_u8*	p0 = fgMetaMemory + start;
	uae_u8*	p1 = fgMetaMemory + end;
	uae_u8*	p;

	if (andValue == 0xFF)
	{
		for (p = p0; p < p1; ++p)
		{
			*p |= orValue;
		}
	}
	else if (orValue == 0x00)
	{
		for (p = p0; p < p1; ++p)
		{
			*p &= andValue;
		}
	}
	else
	{
		for (p = p0; p < p1; ++p)
		{
			*p = (*p & andValue) | orValue;
		}
	}
}


// ---------------------------------------------------------------------------
//		 MetaMemory::GetChunkSize
// ---------------------------------------------------------------------------

DWord MetaMemory::GetChunkSize (VoidPtr p)
{
#if 1
	// !!! Too many calls to PalmHeap here.  Should make a single PalmHeap
	// function to do all this.

	// Get information on the heap for this pointer.

	PalmHeap::HeapInfo	heapInfo = PalmHeap::GetHeapContaining ((uaecptr) p);

	uaecptr				chunkHdr = PalmHeap::GetChunkHeader (p, heapInfo);

	PalmHeap::ChunkInfo	chunkInfo;
	PalmHeap::GetChunkInfo (chunkHdr, heapInfo.version, &chunkInfo);

	return chunkInfo.size - (chunkInfo.chunkHdrSize + chunkInfo.sizeAdj);
#else
	PalmHeap::ChunkInfo	chunkInfo = PalmHeap::GetChunkContaining ((uaecptr) p);

	return chunkInfo.size - (chunkInfo.chunkHdrSize + chunkInfo.sizeAdj);
#endif
}


// ---------------------------------------------------------------------------
//		 MetaMemory::GetChunkLockCount
// ---------------------------------------------------------------------------

DWord MetaMemory::GetChunkLockCount (VoidPtr p)
{
#if 1
	// !!! Too many calls to PalmHeap here.  Should make a single PalmHeap
	// function to do all this.

	// Get information on the heap for this pointer.

	PalmHeap::HeapInfo	heapInfo = PalmHeap::GetHeapContaining ((uaecptr) p);

	uaecptr				chunkHdr = PalmHeap::GetChunkHeader (p, heapInfo);

	PalmHeap::ChunkInfo	chunkInfo;
	PalmHeap::GetChunkInfo (chunkHdr, heapInfo.version, &chunkInfo);

	return chunkInfo.lockCount;
#else
	PalmHeap::ChunkInfo	chunkInfo = PalmHeap::GetChunkContaining ((uaecptr) p);

	return chunkInfo.lockCount;
#endif
}


// ---------------------------------------------------------------------------
//		 MetaMemory::SyncOneBlock
// ---------------------------------------------------------------------------
//	Mark the chunk header, free chunks, and unlocked chunks.

PalmHeap::EWalkerProcResult MetaMemory::SyncOneBlock (	const PalmHeap::HeapInfo& heapInfo,
														const PalmHeap::ChunkInfo& chunkInfo,
														void* userData)
{
	Bool	fullSync		= *(Bool*) userData;
	uae_u8*	ptr				= fgMetaMemory + MASK(chunkInfo.chunkHdrStart);
	DWord	chunkBodySize	= chunkInfo.size - (chunkInfo.chunkHdrSize + chunkInfo.sizeAdj);

	// Set the access for the chunk header.

	memset (ptr, kMemStructBits, chunkInfo.chunkHdrSize);
	ptr += chunkInfo.chunkHdrSize;

	// Set the access for the body of the chunk.

	// Check for free chunks.

	if (chunkInfo.free)
	{
		memset (ptr, kFreeChunkBits, chunkBodySize);

		// Hack for startup time.  When MemInit is called, it creates
		// a free block spanning the entire dynamic heap.  However,
		// that's where the current stack happens to be.  We still need
		// access to that, so free it up.

		uaecptr	SP = m68k_areg (regs, 7);
		if (SP > chunkInfo.chunkHdrStart && SP < chunkInfo.chunkHdrStart + chunkInfo.size)
		{
			uaecptr	stackLow = SP - 0x0400;
			uaecptr stackHigh = SP + 0x0100;

			memset (fgMetaMemory + stackLow, kFreeAccessBits, stackHigh - stackLow);
		}
	}

	// It's not a free chunk; see if it's unlocked.

	else if (chunkInfo.lockCount == 0)
	{
		memset (ptr, kUnlockedChunkBits, chunkBodySize);
	}

	// It's an allocated, locked chunk.  Unless we're doing a full
	// sync, there's no need to do anything with it.
	//
	// (Note: this optimization doesn't really get us anything.  Most of
	//  the heap is going be either free or full of relocatable chunks.)

	else if (fullSync)
	{
		memset (ptr, kFreeAccessBits, chunkBodySize);
	}

	// Set the access for any unallocated trailing bytes.

	if (chunkInfo.sizeAdj)
	{
		ptr += chunkBodySize;
		memset (ptr, kMemStructBits, chunkInfo.sizeAdj);
	}

	return PalmHeap::kKeepIterating;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::WhatHappenedCallback
// ---------------------------------------------------------------------------
//	Check to see if there was an access to memory manager data, an unlocked
//	block, or a free block.

PalmHeap::EWalkerProcResult MetaMemory::WhatHappenedCallback (	const PalmHeap::HeapInfo& heapInfo,
																const PalmHeap::ChunkInfo& chunkInfo,
																void* userData)
{
	WhatHappenedData*	info			= (WhatHappenedData*) userData;

	PalmHeap::EWalkerProcResult result	= PalmHeap::kKeepIterating;
	info->result						= Errors::kOKAccess;

	uaecptr	addrStart	= info->address;
	uaecptr	addrEnd		= info->address + info->size;

	// Sort out some ranges we'll be comparing against.

	uaecptr	hdrStart	= chunkInfo.chunkHdrStart;
	uaecptr	bodyStart	= chunkInfo.chunkHdrStart + chunkInfo.chunkHdrSize;
	uaecptr	trlStart	= chunkInfo.chunkHdrStart + chunkInfo.size - chunkInfo.sizeAdj;
	uaecptr	chunkEnd	= chunkInfo.chunkHdrStart + chunkInfo.size;

	/*
		A note on the comparison below (I had to draw this on the whiteboard
		before I could figure it out).  I need to see if one range of memory
		(the test address, extending for 1, 2, or 4 byte) overlaps another
		range of memory (the chunk header, the chunk body, etc.).  I started
		by drawing the second range as follows:

			+---------+---------+---------+
			|    A    |    B    |    C    |
			+---------+---------+---------+

		"B" is the range I'm interested in testing against, "A" is everything
		before it, and "C" is everything after it.

		The range of memory I want to test to see if it overlaps can be thought
		of as starting in region A, B, or C, and ending in A, B, or C.  We can
		identify all possible ranges by labelling each range "xy", where x is
		the starting section and y is the ending section.  Thus, all the possible
		ranges are AA, AB, AC, BB, BC, and CC.  From inspection, we can see that
		only segments AA and CC don't overlap B in any way.  Looking at it
		another way, any segment ending in A or starting in C will not overlap
		B.  All other ranges will.
	*/

	// See if we hit this chunk at all.  If not, leave.

	if (addrEnd > hdrStart && addrStart < chunkEnd)
	{
		// See if the access was to the body of the block.

		if (addrEnd > bodyStart && addrStart < trlStart)
		{
			// See if this is a free (unallocated) block.

			if (chunkInfo.free)
			{
				info->result = Errors::kFreeChunkAccess;
				result = PalmHeap::kStopIterating;
			}

			// See if this is an unlocked block.

			else if (chunkInfo.lockCount == 0)
			{
				info->result = Errors::kUnlockedChunkAccess;
				result = PalmHeap::kStopIterating;
			}
		}

		// See if the access was to the chunk header.

		else if (addrEnd > hdrStart && addrStart < bodyStart)
		{
			info->result = Errors::kMemMgrAccess;
			result = PalmHeap::kStopIterating;
		}

		// See if the access was to the chunk trailer.
		// Note: this case MUST be true since we've eliminated
		// all other parts of the chunk by now.

		else	// if (addrEnd > trlStart && addrStart < chunkEnd)
		{
			info->result = Errors::kMemMgrAccess;
			result = PalmHeap::kStopIterating;
		}


		// ============================== ALLOW FOR BUGS ==============================

		// Allow the screen driver some liberty.  Some BitBlt functions get a running
		// start or come to a skidding halt before or after an allocated chunk.  They
		// read these two bytes, but the logic is such that those bits get masked out
		// or shifted away.
		//
		// The test we make here is for a two-byte read just before or after a chunk
		// made by PrvMisAlignedForwardInnerBitBlt or PrvMisAlignedBackwardInnerBitBlt.
		// When testing the accessed memory location, we have to check three places:
		// two bytes before the chunk body, two bytes after the chunk body, and the
		// first two bytes of the chunk header.  The reason for making the last check
		// is because there may not be any bytes trailing the chunk body.  If that's
		// the case, then we'll never even make it down to here; the test above to
		// see if we've hit this chunk will already have skipped us to the next chunk.

		if (info->result == Errors::kMemMgrAccess &&
			info->forRead && info->size == sizeof (Word) &&
			(gPrvCompressedInnerBitBlt.InRange () ||
			 gPrvMisAlignedForwardInnerBitBlt.InRange () ||
			 gPrvMisAlignedBackwardInnerBitBlt.InRange ()))
		{
			goto HideBug;
		}


		// There's a bug in MenuHandleEvent (pre-3.1 only) that causes it to read a
		// random long (actually, the random location is the result of using a -1 to index
		// into a menu array).

		if (Patches::HasMenuHandleEventBug () &&
			info->forRead && info->size == sizeof (WinHandle) &&	// MenuPullDownType.menuWin
			gMenuHandleEvent.InRange ())
		{
			goto HideBug;
		}


		// There's a bug in NetPrvSettingSet that causes it to read 0x020C bytes
		// from the beginning of its globals buffer when that buffer is only 8 bytes long.

		if (Patches::HasNetPrvSettingSetBug () &&
			info->forRead && info->size == 4 &&
			gNetPrvSettingSet.InRange ())
		{
			goto HideBug;
		}


		// SysAppExit deletes the block of memory containing the current task's stack
		// and TCB before the task has been deleted.  When SysTaskDelete is called,
		// there are many accesses to both the stack (as functions are called) and
		// the TCB (as tasks are scheduled and as the task to be deleted is finally
		// deleted).
		//
		// We need to detect when any of these SysTaskDelete accesses occurs and
		// allow them.  One hueristic that gets most of them is to see if there
		// are any TCBs with a pointer into the deleted memory chunk.  That will
		// work up until the point the TCB is marked as deleted.  This marking
		// occurs in the AMX function cj_kptkdelete.  At that point, we can just
		// check to see if the access occured in that function and allow it.

		if (info->result == Errors::kFreeChunkAccess &&
			Patches::HasDeletedStackBug ())
		{
			// First see if there is an active TCB with a stack pointer
			// pointing into this deleted memory chunk.

			SysKernelInfoType	taskInfo;

			taskInfo.selector	= sysKernelInfoSelTaskInfo;
			taskInfo.id			= 0;	// Ask for the first task

			// Get the first task.  Remember the task IDs so that we can
			// detect when we've looped through them all.  Remembering *all*
			// the IDs (instead of just the first one) is necessary in case
			// we're called with the linked list of TCBs is inconsistant
			// (that is, it's in the process of being updated -- when that's
			// happening, we may find a loop that doesn't necessarily involve
			// the first TCB).

			vector<DWord>	taskIDList;
			Err				err = SysKernelInfo (&taskInfo);

			while (err == 0)
			{
				// See if we've seen this task ID before.  If so, leave.

				vector<DWord>::iterator	iter = taskIDList.begin ();
				while (iter != taskIDList.end ())
				{
					if (*iter++ == taskInfo.param.task.id)
					{
						goto PlanB;	// I'd really like a break(2) here... :-)
					}
				}

				// We haven't looked at this task before; check the stack.
				// Check against "stackStart" instead of stackP, as the
				// latter sometimes contains values outside of the current
				// stack range (such as the faux stacks we set up when
				// doing ATrap calls).

				if (taskInfo.param.task.stackStart >= bodyStart &&
					taskInfo.param.task.stackStart < trlStart)
				{
					goto HideBug;
				}

				// Save this task ID so we can see if we've visited this
				// record before.

				taskIDList.push_back (taskInfo.param.task.id);

				// Get the next TCB.

				taskInfo.selector	= sysKernelInfoSelTaskInfo;
				taskInfo.id			= taskInfo.param.task.nextID;
				err					= SysKernelInfo (&taskInfo);
			}

			// Plan B...
PlanB:
			// If there is no TCB pointing into the deleted chunk, see if the current
			// function is cj_kptkdelete.  If not, see if it looks like we're in a
			// function called by cj_kptkdelete.

			if (	m68k_areg (regs, 7) >= bodyStart &&
					m68k_areg (regs, 7) < trlStart)
			{
				// See if we're currently in cj_kptkdelete.

				if (gcj_kptkdelete.InRange ())
				{
					info->result = Errors::kOKAccess;
					result = PalmHeap::kStopIterating;
				}

				// If not, see if it's in the current call chain.  Normally (!)
				// I'd walk the a6 stack frame links, but there's a lot of
				// assembly stuff going on here, and not all of them set up
				// stack frames.  a6 *should* point to a valid stack frame
				// somewhere up the line, but it may be cj_kptkdelete's
				// stack frame, resulting in our skipping over it.
				//
				// It's important that gcj_kptkdelete.InRange() has returned
				// true at least once before entering this section of code.
				// That's because if it hasn't, the cached range of the
				// function will be empty, causing InRange to look for the
				// beginning and ending of the function containing the address
				// passed to it.  Since we're passing random contents of the
				// stack, it's very unlikely that there is a function range
				// corresponding to the values passed to InRange, sending it
				// on wild goose chases.  Fortunately, "Plan B" goes into effect
				// only after the TCB has been marked as deleted, which occurs
				// in cj_kptkdelete.  The very next thing cj_kptkdelete does is
				// push a parameter on the stack, causing us to get here while
				// the PC is in cj_kptkdelete.  This will cause the call to InRange
				// immediately above to return true, satisfying the precondition
				// we need here.

				else
				{
					// Get the current stack pointer and use it to iterate
					// over the contents of the stack.  We examine every
					// longword on every even boundary to see if it looks
					// like a return address into cj_kptkdelete.

					uaecptr	a7 = m68k_areg (regs, 7);

					while (a7 >= bodyStart && a7 < trlStart)
					{
						if (gcj_kptkdelete.InRange (get_long (a7)))
						{
							goto HideBug;
						}

						a7 += 2;
					}
				}	// end: checking the stack for cj_kptkdelete
			}	// end: A7 is in the deleted chunk
		}	// end: accessed a deleted chunk


		// There's a bug in FindShowResults (pre-3.0) that accesses
		// a byte in an unlocked block.

		if (info->result == Errors::kUnlockedChunkAccess &&
			Patches::HasFindShowResultsBug () &&
			info->forRead && info->size == sizeof (Boolean) &&
			addrStart == bodyStart + sizeof (Word) * 2 /*offsetof (FindParamsType, more)*/ &&
			gFindShowResults.InRange ())
		{
			goto HideBug;
		}


		// There's a bug in FindSaveFindStr that causes it to possibly read
		// off the end of the source handle.  The invalid access will be
		// generated in MemMove, which is called by DmWrite, which is called
		// by FindSaveFindStr.  So look up the stack a bit to see who's calling us.

		uaecptr	a6_0 = m68k_areg (regs, 6);	// MemMove's A6 (points to caller's A6 and return address to caller)
		uaecptr	a6_1 = get_long (a6_0);		// DmWrites's (points to caller's A6 and return address to caller)

		if (Patches::HasFindSaveFindStrBug () && info->forRead && info->size == 1 &&
#if !HAS_PROFILING
			// When we're not profiling, we have completely patched MemMove.
			// Therefore, A6 is not MemMove's but it's still DmWrite's.
			gFindSaveFindStr.InRange (get_long (a6_0 + 4))	// See if FindSaveFindStr is DmWrite's caller
#else
			gDmWrite.InRange (get_long (a6_0 + 4)) &&		// See if DmWrite is MemMove's caller
			gFindSaveFindStr.InRange (get_long (a6_1 + 4))	// See if FindSaveFindStr is DmWrite's caller
#endif
			)
		{
			goto HideBug;
		}
	}

	goto Exit;

HideBug:
	info->result = Errors::kOKAccess;
	result = PalmHeap::kStopIterating;

Exit:
	return result;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::CheckForChunkHeader
// ---------------------------------------------------------------------------

PalmHeap::EWalkerProcResult MetaMemory::CheckForChunkHeader (	const PalmHeap::HeapInfo& heapInfo,
																const PalmHeap::ChunkInfo& chunkInfo,
																void* userData)
{
	ChunkCheck*	data = (ChunkCheck*) userData;

	if (data->testAddress >= chunkInfo.chunkHdrStart &&
		data->testAddress < chunkInfo.chunkHdrStart + chunkInfo.chunkHdrSize)
	{
		data->result = true;	// it was found in the test range
		return PalmHeap::kStopIterating;
	}

	return PalmHeap::kKeepIterating;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::CheckForUnlockedChunk
// ---------------------------------------------------------------------------

PalmHeap::EWalkerProcResult MetaMemory::CheckForUnlockedChunk (	const PalmHeap::HeapInfo& heapInfo,
																const PalmHeap::ChunkInfo& chunkInfo,
																void* userData)
{
	ChunkCheck*	data = (ChunkCheck*) userData;

	if (chunkInfo.lockCount == 0
		&& data->testAddress >= chunkInfo.chunkHdrStart + chunkInfo.chunkHdrSize
		&& data->testAddress < chunkInfo.chunkHdrStart + chunkInfo.chunkHdrSize - chunkInfo.sizeAdj)
	{
		data->result = true;	// it was found in the test range
		return PalmHeap::kStopIterating;
	}

	return PalmHeap::kKeepIterating;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::CheckForFreeChunk
// ---------------------------------------------------------------------------

PalmHeap::EWalkerProcResult MetaMemory::CheckForFreeChunk (	const PalmHeap::HeapInfo& heapInfo,
															const PalmHeap::ChunkInfo& chunkInfo,
															void* userData)
{
	ChunkCheck*	data = (ChunkCheck*) userData;

	if (chunkInfo.free
		&& data->testAddress >= chunkInfo.chunkHdrStart + chunkInfo.chunkHdrSize
		&& data->testAddress < chunkInfo.chunkHdrStart + chunkInfo.chunkHdrSize - chunkInfo.sizeAdj)
	{
		data->result = true;	// it was found in the test range
		return PalmHeap::kStopIterating;
	}

	return PalmHeap::kKeepIterating;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::CheckForAllocatedChunk
// ---------------------------------------------------------------------------

PalmHeap::EWalkerProcResult MetaMemory::CheckForAllocatedChunk (	const PalmHeap::HeapInfo& heapInfo,
																	const PalmHeap::ChunkInfo& chunkInfo,
																	void* userData)
{
	ChunkCheck*	data = (ChunkCheck*) userData;

	if (!chunkInfo.free
		&& data->testAddress >= chunkInfo.chunkHdrStart + chunkInfo.chunkHdrSize
		&& data->testAddress < chunkInfo.chunkHdrStart + chunkInfo.chunkHdrSize - chunkInfo.sizeAdj)
	{
		data->result = true;	// it was found in the test range
		return PalmHeap::kStopIterating;
	}

	return PalmHeap::kKeepIterating;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::ValidChunk
// ---------------------------------------------------------------------------

Bool MetaMemory::ValidChunk (VoidPtr p)
{
	if (!p)
		return false;

	if (MASK (p) >= fgMetaMemorySize)
		return false;

	// Get information on the heap for this pointer.

	PalmHeap::HeapInfo	heapInfo = PalmHeap::GetHeapContaining ((uaecptr) p);

	if (!heapInfo.heapHdrStart)
		return false;

	PalmHeap::EPalmHeapResult	result = PalmHeap::ValidatePointer (p, heapInfo);

	return result == PalmHeap::kChunkOK;
}


// ---------------------------------------------------------------------------
//		 MetaMemory::ValidHandle
// ---------------------------------------------------------------------------

Bool MetaMemory::ValidHandle (VoidHand h)
{
	if (!h)
		return false;

	if (MASK (memHandleUnProtect(h)) >= fgMetaMemorySize)
		return false;

	// Get information on the heap for this pointer.

	PalmHeap::HeapInfo	heapInfo = PalmHeap::GetHeapContaining ((uaecptr) memHandleUnProtect(h));

	if (!heapInfo.heapHdrStart)
		return false;

	PalmHeap::EPalmHeapResult	result = PalmHeap::ValidateHandle (h, heapInfo);

	return result == PalmHeap::kChunkOK;
}


// ---------------------------------------------------------------------------
//		 FunctionRange::FunctionRange
// ---------------------------------------------------------------------------

FunctionRange::FunctionRange (const char* functionName) :
	fName (functionName),
	fBegin (UAE_NULL),
	fEnd (UAE_NULL)
{
}


// ---------------------------------------------------------------------------
//		 FunctionRange::InRange
// ---------------------------------------------------------------------------

Bool FunctionRange::InRange (uaecptr addr)
{
	if (fBegin == UAE_NULL)
	{
		GetRange (addr);
	}

	if ((fBegin != UAE_NULL) && (addr >= fBegin) && (addr < fEnd))
	{
		return true;
	}

	return false;
}


// ---------------------------------------------------------------------------
//		 FunctionRange::Reset
// ---------------------------------------------------------------------------

void FunctionRange::Reset (void)
{
	fBegin = UAE_NULL;
	fEnd = UAE_NULL;
}


// ---------------------------------------------------------------------------
//		 FunctionRange::GetRange
// ---------------------------------------------------------------------------

void FunctionRange::GetRange (uaecptr addr)
{
	char		name[80];
	uaecptr		startAddr;
	uaecptr		endAddr;

	::FindFunctionName (addr, name, &startAddr, &endAddr);

	if (strcmp (name, fName) == 0)
	{
		fBegin = startAddr;
		fEnd = endAddr;
	}
}


/*
Bool FindAllAppChunks (uaecptr chunkHdr, int ver, void* userData)
{
	MemChunkHeaderUnionType	chunk;
	chunk = *(MemChunkHeaderUnionType*) chunkP;
	Canonical (chunk);

	if (memUChunkOwner (&chunk, ver) == (long) userData)
		MemChunkFree ((BytePtr) chunkP + memUSizeOfChunkHeader(ver));

	return false;
}
*/
