/*********************************************************************
*
* Copyright (c) 1998
* 3Com/Palm Computing Division.  All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions
* are met:
*
* 1. Redistributions of source code must retain the above copyright
*	 notice, this list of conditions and the following disclaimer.
*
* 2. Redistributions in binary form must reproduce the above
*	 copyright notice, this list of conditions and the following
*	 disclaimer in the documentation and/or other materials provided
*	 with the distribution.
*
* 3. All advertising materials mentioning features or use of this
*	 software must display the following acknowledgement:
*
*		This product includes software developed by 3Com and its
*		contributors.
*
* 4. Neither 3Com nor the names of its contributors may be used to
*	 endorse or promote products derived from this software
*	 without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE 3COM AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
* TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 3COM OR
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
* USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
* OF SUCH DAMAGE.
*
*-------------------------------------------------------------------
* FileName:
*	SerLibSocketsWin.cpp
* 
* Description:
*
*	This source file implements the routines used by SerTransport.cpp
*	when the serial transport mode is set as tptSockets (as defined in
*	SerTransport.h). 
*
*	The SerTransport routines are used in place of the SerXXX routines
*	when running in a host application like PalmDebugger or PalmOSEmulator.
*	These routines allow us to generically replace the serial comm
*	with an alternative comm.  
*
*	The SerialLinkMgr.c source file contains an #ifdef near the top
*	that remaps all the SerXXX routines to TptXXX routines. Each of
*	the TptXXX routines contains run-time checks that dispatch to
*	the appropriate low level comm handler routine depending on the
*	current transport setting (stored in a global and setup by TptSetType()
*
* Notes:
*	An feeble attempt has been made to make this source file buildable on any
*	  platform that has Berkeley sockets API calls but for a number of
*	  reasons it's only Win32 for now. 
*
* History:
*	21-Nov-1998  - Created from SocketMessaging.cpp
*
****************************************************************/

#if defined (_WIN32)
#pragma warning( disable : 4127)	// conditional expression is constant
#include <windows.h>
#include <assert.h>

#define	  prvInvalidSocket	  INVALID_SOCKET
#define	  prvSocketErr		  SOCKET_ERROR
#endif


// Include error codes from SerTransport.h
// Define this so we only get the platform independent equates
//	 from SerTransport.h (i.e. the error code enums )
#define   SER_TRANSPORT_PLATFORM_IND_ONLY
#include "SerTransport.h"

// Prototypes for functions in this module 
#include "SerLibSockets.h"



// ---------------------------------------------------------------------------
//		 Stuff
// ---------------------------------------------------------------------------

enum {	kSocketState_Closed, kSocketState_NeedConnect, 
		kSocketState_Listening, kSocketState_Connected };

// gInitSocketState is shoved into gSocketState by SerSockStop()
static bool 			gConnectingSocket	= false;
static int				gSocketState		= kSocketState_Closed;
static SOCKET			gListeningSocket	= prvInvalidSocket;
static SOCKET			gDebuggerSocket 	= prvInvalidSocket;

static SOCKET			PrvNewSocket (void);
static struct sockaddr* PrvFillAddress (struct sockaddr* addr, bool forConnect);
static short			PrvGetTptError (void);
static int				PrvGetError (void);
static unsigned long	PrvGetTicks(void);



/***********************************************************************
*
* FUNCTION: 	SerSockStart
*
* DESCRIPTION:	Starts up the sockets mechanism for communicating with
*				the debugger.  It first attempts to connect to the
*				debugger, assuming that it's been started up and issued
*				a listen().  If that fails, it creates a socket to
*				listen on itself.
*
* PARAMETERS:	
*				connecting	  IN	if true, connect to other host. 
*									if false, creating listening socket
*
* RETURNED: 	0 if no err.
*
* CALLED BY:	TptOpen() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/
long SerSockStart (bool connecting)
{
  bool	alwaysFalse = false;
  
  assert (gSocketState == kSocketState_Closed);
  assert (gListeningSocket == prvInvalidSocket);
  assert (gDebuggerSocket == prvInvalidSocket);
  
  
  // Save socket type for use in ReceivePartial when we get EOF
  gConnectingSocket = connecting;
  
#if defined (_WIN32)
  sockaddr	addr;
  WORD		versionRequested = MAKEWORD (2, 0);
  WSADATA		wsaData;
  
  int			result = WSAStartup (versionRequested, &wsaData);
  assert (!result);
  
  if (LOBYTE (wsaData.wVersion) != 2 || HIBYTE (wsaData.wVersion) != 0)
	{
	  assert (alwaysFalse);
	  return tptErrInternal;
	}
#endif
  
  
  // If connecting to remote host, attempt to connect to that address 
  //  (in case anyone is listening). If we fail the connect, we try again
  //  at the start of every send and receive call. 
  if (connecting)
	{
	  gDebuggerSocket = PrvNewSocket ();
	  if (gDebuggerSocket == prvInvalidSocket) return tptErrInternal;
	  
	  gSocketState = kSocketState_NeedConnect;
	  result = connect (gDebuggerSocket, PrvFillAddress (&addr, true), sizeof(addr));
	  if (result == 0)
		{
		  gSocketState = kSocketState_Connected;
		  return 0;
		}
	  else
		return PrvGetTptError();
	}
  
  // Else, setup listening socket
  else
	{
	  gListeningSocket = PrvNewSocket ();
	  if (gListeningSocket == prvInvalidSocket) return tptErrInternal;
	  
	  result = bind (gListeningSocket, PrvFillAddress (&addr, false), sizeof(addr));
	  assert (!result);
	  if (result) return PrvGetTptError();
	  
	  // Start listening for a connection.
	  result = listen (gListeningSocket, 1);
	  assert (!result);
	  if (result) return PrvGetTptError();
	  
	  gSocketState = kSocketState_Listening;
	}

  return tptErrInternal;
}


/***********************************************************************
*
* FUNCTION: 	SerSockStop
*
* DESCRIPTION:	Stops the sockets session.	Closes all sockets, sets
*				our state to unconnected, and closes down the sockets
*				library (on Windows).
*
* PARAMETERS:	None.
*
* RETURNED:		0 if no err.
*
* CALLED BY:	TptClose() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/

long SerSockStop (void)
{

  if (gDebuggerSocket != prvInvalidSocket)
	{
	  closesocket (gDebuggerSocket);
	  gDebuggerSocket = prvInvalidSocket;
	}
  
  if (gListeningSocket != prvInvalidSocket)
	{
	  closesocket (gListeningSocket);
	  gListeningSocket = prvInvalidSocket;
	}
  
  gSocketState = kSocketState_Closed;
  
#if defined (_WIN32)
  WSACleanup ();
#endif

  return 0;
}


/***********************************************************************
*
* FUNCTION: SerSockConnectPending
*
* DESCRIPTION:	Returns whether or not there is a pending connect.	This
*				function can be called at any time.  It will only return
*				true if there is a non-blocking listen issued with a
*				pending connect that we haven't yet accepted.
*
* PARAMETERS:	None.
*
* RETURNED: 	True if there is a pending connect.  If there is, the
*				caller should next call SerSockAcceptConnection.
*
* CALLED BY:	PalmOSEmulator in it's idle loop
*
* NOTES:
*
***********************************************************************/

bool SerSockConnectPending (void)
{
  assert (!gConnectingSocket);

  // If already connected, it's easy
  if (gSocketState == kSocketState_Connected) return false;


  // If we closed the socket due to an error from a previous read,
  //  re-open it now. 
  if (gSocketState == kSocketState_Closed)
	{
	  SerSockStart( gConnectingSocket);
	}


  // From the sockets manual for the select() function:
  //
  // The parameter readfds identifies the sockets that are to be
  // checked for readability. If the socket is currently in the
  // listen state, it will be marked as readable if an incoming
  // connection request has been received such that an accept is
  // guaranteed to complete without blocking. For other sockets,
  // readability means that queued data is available for reading
  // such that a call to recv, WSARecv, WSARecvFrom, or recvfrom
  // is guaranteed not to block.

  bool	havePendingConnect = false;

  if (gSocketState == kSocketState_Listening)
	{
	  assert (gListeningSocket != prvInvalidSocket);
	  assert (gDebuggerSocket == prvInvalidSocket);
	  
	  struct fd_set readfds;
	  
	  FD_ZERO (&readfds);
	  FD_SET (gListeningSocket, &readfds);
	  
	  timeval	timeout = {0, 0};	// Don't wait at all
	  
	  int	result = select (0, &readfds, NULL, NULL, &timeout);
	  
	  if (result > 0 && FD_ISSET (gListeningSocket, &readfds))
		havePendingConnect = true;
	  else if (result == prvSocketErr)
		result = PrvGetTptError ();
	}

  return havePendingConnect;
}


/***********************************************************************
*
* FUNCTION: 	SerSockAcceptConnection
*
* DESCRIPTION:	Accepts a pending connection.  Assumes that
*				ConnectPending has already been called and returned
*				true.  After this, the socket connection should be open
*				for business.
*
* PARAMETERS:	None.
*
* RETURNED: 	0 if no error.
**
* CALLED BY:	PalmOSEmulator in it's idle loop
*
* NOTES:
*
***********************************************************************/

long SerSockAcceptConnection (void)
{
  struct	linger	lingerVal = {1, 0};
  
  assert (gSocketState == kSocketState_Listening);
  assert (gListeningSocket != prvInvalidSocket);
  assert (gDebuggerSocket == prvInvalidSocket);
  
  gDebuggerSocket = accept (gListeningSocket, NULL, NULL);
  
  if (gDebuggerSocket == prvInvalidSocket)
	return PrvGetTptError();
  
  
  // Turn linger off so socket doesn't stick around after we close it
  int result;
  result = setsockopt (gDebuggerSocket, SOL_SOCKET, SO_LINGER, (char*) &lingerVal, 
	  sizeof(lingerVal));	
  
  
  gSocketState = kSocketState_Connected;

  return 0;
}


/***********************************************************************
*
* FUNCTION: SerSockIsConnected
*
* DESCRIPTION:	Returns whether or not we have a live connection to an
*				external debugger via a socket.  This function can be
*				called at any time.
*
* PARAMETERS:	None.
*
* RETURNED: Nothing.
*
*
* CALLED BY:	PalmOSEmulator in it's idle loop
*
* NOTES:
*
***********************************************************************/
bool SerSockIsConnected (void)
{
  // If we thought we were connected, see if we still are.
  
  if (gSocketState == kSocketState_Connected)
	{
	  // From the sockets manual for the select() function:
	  //
	  // If a socket is processing a connect call (nonblocking), a socket is
	  // writable if the connection establishment successfully completes. If
	  // the socket is not processing a connect call, writability means a send,
	  // sendto, or WSASendto are guaranteed to succeed.
	  
	  struct fd_set writefds;
	  
	  FD_ZERO (&writefds);
	  FD_SET (gDebuggerSocket, &writefds);
	  
	  timeval	timeout = {0, 0};	// Don't wait at all
	  
	  int		result = select (0, NULL, &writefds, NULL, &timeout);
	  
	  bool	isConnected = result > 0 && FD_ISSET (gDebuggerSocket, &writefds);
	  
	  // If we're no longer connected, update our internal state.
	  
	  if (!isConnected)
		gSocketState = kSocketState_NeedConnect;
	}
  
  return gSocketState == kSocketState_Connected;
}




/***********************************************************************
*
* FUNCTION: SerSockSend
*
* DESCRIPTION:	Writes data out to the buffer.
*
* PARAMETERS:	buffer - pointer to the data to be written.
*
*				amountToWrite - number of bytes in the buffer.
*
*				errP - error code returned here
*
* RETURNED:		# of bytes actually written
*
*				On any error, the sub-system will be closed down.  All
*				sockets will be closed and the state will be set to
*				unconnected.
*
* CALLED BY:	TptSend() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/
long SerSockSend (const void* buffer, long amountToWrite, short* errP)
{
  int amountWritten = 0;
  
  
  // If we closed the socket due to an error from a previous read,
  //	re-open it now. 
  if (gSocketState == kSocketState_Closed)
	SerSockStart( gConnectingSocket);
  
  // If we're a connecting socket and not connected yet, try
  //	and connect now
  if (gSocketState == kSocketState_NeedConnect)
	{
	  sockaddr	addr;
	  int 		result;
	  
	  result = connect (gDebuggerSocket, PrvFillAddress (&addr, true), 
		  sizeof(addr));

	  if (result == 0)
		  gSocketState = kSocketState_Connected;
	  else
		{
		  *errP = PrvGetTptError ();
		  return 0;
		}
	}
  
  
  
  while (amountToWrite)
	{
	  int sentThisTime = send (gDebuggerSocket, &((char*) buffer)[amountWritten],
								amountToWrite, 0);
	  
	  if (sentThisTime == prvSocketErr )
		{
		  *errP = PrvGetTptError ();
		  return amountWritten;
		}
	  
	  amountWritten += sentThisTime;
	  amountToWrite -= sentThisTime;
	}
  
  return amountWritten;
}


/***********************************************************************
*
* FUNCTION:		SerSockReceivePartial
*
* DESCRIPTION:	Receive up to count bytes from the socket.	It is assumed that 
*				HasUnreadData has already been called and returned true.
*
* PARAMETERS:	buffer	- pointer to the buffer to put the data.
*
*				count	-	 size of the buffer.
*				timeout	- timeout, -1 means wait forever
*				errP	- error returned here
*
* RETURNED: The actual number of bytes read, or -1 on error
*
*
* CALLED BY:	Apps that wish to read as much data as possible
*
* NOTES:
*
***********************************************************************/

long SerSockReceivePartial (void* buffer, unsigned long count, long timeout,
								short* errP)
{
  unsigned long	tickStart;
  long			timeLeft;
  
  // Return err if not connected yet
  if (gSocketState != kSocketState_Connected)
	{
	  *errP = tptErrTimeOut;
	  return 0;
	}
  
  
  tickStart = PrvGetTicks();
  if (timeout < 0) timeLeft = -1;
  else
	{ 
	  timeLeft = (timeout + tickStart - PrvGetTicks());
	  if (timeLeft < 0) timeLeft = 0;
	}
  
  
  // Wait for data to arrive before the timeout, unless timeout is infinite
  if (timeout >= 0)
	{
	  *errP = (short)SerSockReceiveWait (1, timeLeft);
	  if (*errP) return 0;
	}
  else *errP = 0;
  
  
  // Receive the data
  int amountRead = recv (gDebuggerSocket,  ((char*) buffer), count, 0);
  if (amountRead == prvSocketErr || amountRead == 0)
	{
	  *errP = PrvGetTptError();
	  if (amountRead == 0) 
		  *errP = tptErrTimeOut;
	  else
		  *errP = PrvGetTptError();
	  
	  SerSockStop();
	  return 0;
	}
  
  return amountRead;
}


/***********************************************************************
*
* FUNCTION: SerSockReceive
*
* DESCRIPTION:	Don't return until count bytes of data have been read
*				from the socket, or until an error occurs. 
*
* PARAMETERS:	buffer	  IN  pointer to the buffer to put the data.
*				count	  IN  number of bytes to read
*				timeout   IN  timeout in ticks
*				*errP	  OUT error, if any returned here
*
* RETURNED: The actual number of bytes read.
*
*
* CALLED BY:	TptReceive() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/
long	SerSockReceive	(void* buffer, unsigned long count, long timeout, 
							 short* errP)
{
  unsigned long	amountRead = 0;
  unsigned long	tickStart;
  long			timeLeft;
  int 			actBytes;
  
  
  tickStart = PrvGetTicks();
  while (count - amountRead > 0)
	{
	  if (timeout < 0) timeLeft = -1;
	  else
		{ 
		  timeLeft = (timeout + tickStart - PrvGetTicks());
		  if (timeLeft < 0) timeLeft = 0;
		}
	  
	  
	  // Receive next partial bit
	  actBytes = SerSockReceivePartial (((char*) buffer) + amountRead, 
		  count - amountRead, timeLeft, errP);
	  
	  if (*errP) return amountRead;
	  
	  
	  // Increment amount read
	  amountRead += actBytes;
	  if (amountRead == count) break;
	  
	}
  
  return amountRead;
}



/***********************************************************************
*
* FUNCTION: SerSockReceiveWait
*
* DESCRIPTION:	Don't return until at least 1 byte is available
*					on the socket, or until error occurrs
*
* PARAMETERS:	
*				bytes	  IN  ignored
*				timeout   IN  timeout in ticks, -1 means wait forever
*
* RETURNED: 0 if no error
*
*
* CALLED BY:	TptReceiveWait() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/
long	SerSockReceiveWait	(unsigned long bytes, long timeout)
{
  
  
  // Get rid of warnings
  bytes = bytes;
  
  // If not connected, return error
  if (gSocketState != kSocketState_Connected )
	  return tptErrTimeOut;
  
  // From the sockets manual for the select() function:
  //
  // The parameter readfds identifies the sockets that are to be
  // checked for readability. If the socket is currently in the
  // listen state, it will be marked as readable if an incoming
  // connection request has been received such that an accept is
  // guaranteed to complete without blocking. For other sockets,
  // readability means that queued data is available for reading
  // such that a call to recv, WSARecv, WSARecvFrom, or recvfrom
  // is guaranteed not to block.
  
  struct fd_set	readfds;
  
  FD_ZERO (&readfds);
  FD_SET (gDebuggerSocket, &readfds);
  
  timeval sockTimeout = {0, timeout*1000L};	// timeout in microseconds
  
  int result = select (0, &readfds, NULL, NULL, &sockTimeout);
  
  if (result > 0 && FD_ISSET (gDebuggerSocket, &readfds))
	  return 0;
  else  
	  return tptErrTimeOut;
}


/***********************************************************************
*
* FUNCTION: SerSockReceiveFlush
*
* DESCRIPTION:	Flush all data available in socket 
*
* PARAMETERS:	
*				bytes	  IN  ignored
*				timeout   IN  not used
*
* RETURNED: 0 if no error
*
*
* CALLED BY:	TptReceiveFlush() in SerTransport.cpp
*
* NOTES:
*
***********************************************************************/
long	SerSockReceiveFlush ( long timeout)
{
  char	buffer[500];
  short	err;
  
  unsigned long	startTime = PrvGetTicks ();
  
  do 
	{
	if (!SerSockReceiveWait (1, timeout))
		SerSockReceivePartial (buffer, sizeof (buffer), timeout, &err);
	else
		break;
	} 
  while (timeout < 0  
		  || (PrvGetTicks () - startTime <= (unsigned long)timeout));
	  
  return 0;
}



/***********************************************************************
*
* FUNCTION: PrvNewSocket
*
* DESCRIPTION:	Creates and returns a new socket configured the way we
*				like it.  The way we like it is: reusable after it's
*				closed, uses "keep alive" packets, and non-blocking.
*
* PARAMETERS:	None.
*
* RETURNED: The created socket, or prvInvalidSocket on error.
*
*				On any error, the sub-system will be closed down.  All
*				sockets will be closed and the state will be set to
*				unconnected.
*
***********************************************************************/
static SOCKET PrvNewSocket (void)
{
  BOOL	tmp;
  int 	result;
  struct	linger	lingerVal = {1, 0};
  
  SOCKET	newSocket = socket (AF_INET, SOCK_STREAM, 0);
  if (newSocket == prvInvalidSocket)
	  goto Error;
  
  // Allow rapid reuse of this port.
  tmp = 1;
  result = setsockopt (newSocket, SOL_SOCKET, SO_REUSEADDR, (char*) &tmp, sizeof(tmp));
  if (result != 0)
	  goto Error;
  
  
  // Turn linger off so socket doesn't stick around after we close it
  result = setsockopt (newSocket, SOL_SOCKET, SO_LINGER, (char*) &lingerVal, 
	  sizeof(lingerVal));
  if (result != 0)
	  goto Error;
  
  
  // Send all packets immediately
  // Set the socket to not use the "Nagle delay algorithm" that clumps
  // batches of small writes together into one big send.
  tmp = 1;
  result = setsockopt (newSocket, IPPROTO_TCP, TCP_NODELAY, 
  				  (char*) &tmp, sizeof(tmp));
  if (result)
	{
	  assert (!result);
	}
  

  // Enable TCP keep alive process.
  tmp = 1;
  result = setsockopt (newSocket, SOL_SOCKET, SO_KEEPALIVE, (char*) &tmp, sizeof(tmp));
  if (result != 0)
	  goto Error;
  
  return newSocket;
  
Error:
  if (newSocket != prvInvalidSocket)
	  closesocket (newSocket);

  assert (false);
  
  return prvInvalidSocket;
}




/***********************************************************************
*
* FUNCTION: PrvFillAddress
*
* DESCRIPTION:	Fill in a sockaddr data structure with values
*				appropriate for connecting.  For now, this is an
*				Internet address on the local machine with port 2000.
*
* PARAMETERS:	addr - the sockaddr to fill in.
*
* RETURNED: The same sockaddr passed in.
*
***********************************************************************/

static struct sockaddr* PrvFillAddress (struct sockaddr* addr, bool forConnect)
{
  const long		portNumber = 2000;
  sockaddr_in*		addr_in = (sockaddr_in*) addr;
  
  addr_in->sin_family 		= AF_INET;
  addr_in->sin_port			= htons (portNumber);
  addr_in->sin_addr.s_addr	= htonl (forConnect ? INADDR_LOOPBACK : INADDR_ANY);
  
  return addr;
}


/***********************************************************************
*
* FUNCTION: PrvGetError
*
* DESCRIPTION:	Returns the error code for the last operation in
*				  native OS format
*
* PARAMETERS:	None.
*
* RETURNED: The error code as returned by OS
*
***********************************************************************/
static int PrvGetError (void)
{
	
#if defined (_WIN32)
	return ::WSAGetLastError ();
#else
	return errno;
#endif
}


/***********************************************************************
*
* FUNCTION: PrvGetTptError
*
* DESCRIPTION:	Returns the error code for the last operation.
*
* PARAMETERS:	None.
*
* RETURNED: The error code mapped to a tptErr constant
*
***********************************************************************/
static short PrvGetTptError (void)
{
  int 	err;
  
  err = PrvGetError();
  
#if defined (_WIN32)
  switch (err)
	{
	  case WSAENOTCONN:
	  case WSAECONNREFUSED:
	  case WSAESHUTDOWN:
		  return tptErrTimeOut;
	  default:
		  return tptErrInternal;
	}
		
#else
  return tptErrInternal;
#endif
}



/************************************************************
*
*  FUNCTION: PrvGetTicks
*
*  DESCRIPTION: Returns TickCount
*
*  PARAMETERS: void
*
*  RETURNS: long tickcount
*
*************************************************************/
static unsigned long PrvGetTicks(void)
{
#if defined (_WIN32)
  return GetTickCount();
#else
  #error
#endif
}
