/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "PalmHeap.h"

#include "CPU_REG.h"			// Emulator::HandleDlgButton
#include "ErrorHandling.h"		// Errors::ReportCorruptedHeap
#include "RAM_ROM.h"			// CEnableFullAccess


// Some handy macros lifted from MemoryPrvNew.h

#define	memHandleProtect(h)											\
			((VoidHand)((DWord)h | 0x80000000))
#define	memHandleUnProtect(h)										\
			((VoidHand)((DWord)h & 0x7FFFFFFF))


// ===========================================================================
//		 PalmHeap
// ===========================================================================

static uaecptr	gHeapPtrs[(8 * 1024L * 1024L) / (64 * 1024L) + 2];


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::Initialize
 *
 * DESCRIPTION:	Initializes the PalmHeap sub-system.  This function is
 *				called just once after a new ROM is loaded.
 *
 * PARAMETERS:	nothing
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void PalmHeap::Initialize (void)
{
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::Reset
 *
 * DESCRIPTION:	Re-initializes the PalmHeap sub-system.  This function
 *				is called when the ROM is "rebooted" (which includes
 *				the initial boot-up after a ROM is loaded).
 *
 * PARAMETERS:	nothing
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void PalmHeap::Reset (void)
{
	memset (gHeapPtrs, 0, sizeof (gHeapPtrs));
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::Dispose
 *
 * DESCRIPTION:	Cleans up the PalmHeap sub-system.  This function is
 *				called when a ROM is unloaded (including when the
 *				emulator quits).
 *
 * PARAMETERS:	nothing
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void PalmHeap::Dispose (void)
{
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetHeapInfo
 *
 * DESCRIPTION:	Returns information about the given heap.
 *
 * PARAMETERS:	heapID - the Palm OS ID of the heap to get information
 *				on.
 *
 *				info - a pointer to a HeapInfo structure.  This
 *				structure will be filled with the returned
 *				information.
 *
 * RETURNED:	Information is returned in the HeapInfo structure.
 *
 ***********************************************************************/

void PalmHeap::GetHeapInfo (UInt heapID, HeapInfo* info)
{
	GetHeapInfo (GetHeapPtr (heapID), info);
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetHeapInfo
 *
 * DESCRIPTION:	Returns information about the given heap.
 *
 * PARAMETERS:	heapPtr - a pointer to the base of the heap to get
 *				information on.  The pointer points to the Palm OS
 *				heap header (that is, the Mem1HeapHeaderType,
 *				Mem2HeapHeaderType, or MemHeapHeaderType).
 *
 *				info - a pointer to a HeapInfo structure.  This
 *				structure will be filled with the returned
 *				information.
 *
 * RETURNED:	Information is returned in the HeapInfo structure.
 *
 ***********************************************************************/

void PalmHeap::GetHeapInfo (uaecptr heapPtr, HeapInfo* info)
{
	CEnableFullAccess	munge;	// Remove blocks on memory access.


	// Save the pointer to the beginning of the heap.  This pointer
	// points to a struct, the definition of which changes depending
	// on the heap version.

	info->heapHdrStart	= heapPtr;


	// Get the heap version so we know the layout of the header.

	info->version = GetHeapVersion (info->heapHdrStart);


	/*
		Heap layout

			flags:					// Always here, always 2 bytes
			size					// 2 bytes in V1, 4 otherwise
			firstFreeChunkOffset	// Only in V3
			mstrPtrTbl
				numEntries			// always 2 bytes long
				nextTblOffset		// 2 bytes in V1, 4 otherwise
				mstrP [numEntries]
			...chunks...
	*/

	MPTInfo	mptInfo;

	switch (info->version)
	{
		case 1:
			info->flags			= get_word (info->heapHdrStart);
			info->size			= get_word (info->heapHdrStart + offsetof (Mem1HeapHeaderType, size));
			info->firstFree		= UAE_NULL;
			info->chunkHdrSize	= sizeof (Mem1ChunkHeaderType);

			GetMPTInfo (info->heapHdrStart + offsetof (Mem1HeapHeaderType, mstrPtrTbl), info->version, &mptInfo);
			info->heapHdrSize	= sizeof (Mem1HeapHeaderType) - mptInfo.mptHdrSize;
			info->firstChunk	= info->heapHdrStart + info->heapHdrSize + mptInfo.mptHdrSize + mptInfo.numEntries * sizeof (Ptr);

			// In version 1 heaps, a zero in the size field means 64K.

			if (info->size == 0)
			{
				info->size = 64 * 1024L;
			}
			break;

		case 2:
			info->flags			= get_word (info->heapHdrStart);
			info->size			= get_long (info->heapHdrStart + offsetof (Mem2HeapHeaderType, size));
			info->firstFree		= UAE_NULL;
			info->chunkHdrSize	= sizeof (MemChunkHeaderType);

			GetMPTInfo (info->heapHdrStart + offsetof (Mem2HeapHeaderType, mstrPtrTbl), info->version, &mptInfo);
			info->heapHdrSize	= sizeof (Mem2HeapHeaderType) - mptInfo.mptHdrSize;
			info->firstChunk	= info->heapHdrStart + info->heapHdrSize + mptInfo.mptHdrSize + mptInfo.numEntries * sizeof (Ptr);
			break;

		case 3:
			info->flags			= get_word (info->heapHdrStart);
			info->size			= get_long (info->heapHdrStart + offsetof (MemHeapHeaderType, size));
			info->firstFree		= get_long (info->heapHdrStart + offsetof (MemHeapHeaderType, firstFreeChunkOffset));
			info->chunkHdrSize	= sizeof (MemChunkHeaderType);

			GetMPTInfo (info->heapHdrStart + offsetof (MemHeapHeaderType, mstrPtrTbl), info->version, &mptInfo);
			info->heapHdrSize	= sizeof (MemHeapHeaderType) - mptInfo.mptHdrSize;
			info->firstChunk	= info->heapHdrStart + info->heapHdrSize + mptInfo.mptHdrSize + mptInfo.numEntries * sizeof (Ptr);
			break;

		default:
			assert (false);
	}
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetMPTInfo
 *
 * DESCRIPTION:	Returns information about the given master pointer table.
 *
 * PARAMETERS:	mptHdrStart - a pointer to the master pointer table to
 *				get information on (that is, the Mem1MstrPtrTableType
 *				or MemMstrPtrTableType).  This pointer can be
 *				obtained by adding HeapInfo::heapHdrSize to to
 *				HeapInfo::heapHdrStart (where HeapInfo is returned
 *				by GetHeapInfo).  Since there can be more than one
 *				master pointer table in a heap, subsequent pointers
 *				can be found by adding MPTInfo::nextTblOffset to
 *				HeapInfo::heapHdrStart (if MPTInfo::nextTblOffset is
 *				NULL, there is no next table).
 *
 *				info - a pointer to an MPTInfo structure.  This
 *				structure will be filled with the returned
 *				information.
 *
 * RETURNED:	Information is returned in the MPTInfo structure.
 *
 ***********************************************************************/

void PalmHeap::GetMPTInfo (uaecptr mptHdrStart, int heapVersion, MPTInfo* info)
{
	CEnableFullAccess	munge;	// Remove blocks on memory access.

	switch (heapVersion)
	{
		case 1:
			info->version		= 1;

			info->mptHdrStart	= mptHdrStart;
			info->mptHdrSize	= sizeof (Mem1MstrPtrTableType);

			info->numEntries	= get_word (mptHdrStart + offsetof (Mem1MstrPtrTableType, numEntries));
			info->nextTblOffset	= get_word (mptHdrStart + offsetof (Mem1MstrPtrTableType, nextTblOffset));
			break;

		case 2:
		case 3:
			info->version		= 2;

			info->mptHdrStart	= mptHdrStart;
			info->mptHdrSize	= sizeof (MemMstrPtrTableType);

			info->numEntries	= get_word (mptHdrStart + offsetof (MemMstrPtrTableType, numEntries));
			info->nextTblOffset	= get_long (mptHdrStart + offsetof (MemMstrPtrTableType, nextTblOffset));
			break;

		default:
			assert (false);
	}
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetChunkInfo
 *
 * DESCRIPTION:	Returns information about the given chunk.
 *
 * PARAMETERS:	hdrP - a pointer to the header information about the
 *				chunk to return information on.
 *
 *				heapVersion - the heap version, as returned by
 *				GetHeapInfo (HeapInfo::version).
 *
 *				info - a pointer to a ChunkInfo structure.  This
 *				structure will be filled with the returned
 *				information.
 *
 * RETURNED:	Information is returned in the ChunkInfo structure.
 *
 ***********************************************************************/

void PalmHeap::GetChunkInfo (	uaecptr		hdrP,
								int			heapVersion,
								ChunkInfo*	info)
{
	CEnableFullAccess	munge;	// Remove blocks on memory access.

	// Clear out all the fields in case we have to make a hasty retreat.

	memset (info, 0, sizeof (*info));

	switch (heapVersion)
	{
		case 1:
		{
			info->version		= 1;
			info->chunkHdrStart	= hdrP;
			info->chunkHdrSize	= sizeof (Mem1ChunkHeaderType);

			Word	size		= get_word (hdrP + offsetof (Mem1ChunkHeaderType, size));

			if (size == 0)
			{
				// Leave now.  We're at the end of the heap.  If we're at the end of RAM,
				// too, attempts to access subsequent fields will cause a bus error.

				break;
			}

			Byte	lockOwner	= get_byte (hdrP + offsetof (Mem1ChunkHeaderType, lockOwner));
			Byte	flags		= get_byte (hdrP + offsetof (Mem1ChunkHeaderType, flags));
			SWord	hOffset		= get_word (hdrP + offsetof (Mem1ChunkHeaderType, hOffset));

			info->free			= (flags & memChunkFlagFree) != 0;
			info->moved			= (flags & memChunkFlagUnused1) != 0;
			info->unused2		= (flags & memChunkFlagUnused2) != 0;
			info->unused3		= (flags & memChunkFlagUnused3) != 0;
			info->sizeAdj		= (flags & mem1ChunkFlagSizeAdj);
			info->size			= size;
			info->lockCount		= (lockOwner & mem1ChunkLockMask) >> 4;
			info->owner			= (lockOwner & mem1ChunkOwnerMask);
			info->hOffset		= hOffset;

			break;
		}

		case 2:
		case 3:
		{
			info->version		= 2;
			info->chunkHdrStart	= hdrP;
			info->chunkHdrSize	= sizeof (MemChunkHeaderType);

			DWord	part1		= get_long (hdrP);

			if ((part1 & 0x00FFFFFF) == 0)
			{
				// Leave now.  We're at the end of the heap.  If we're at the end of RAM,
				// too, attempts to access subsequent fields will cause a bus error.

				break;
			}

			DWord	part2		= get_long (hdrP + 4);

			info->free			= (part1 & 0x80000000) != 0;
			info->moved			= (part1 & 0x40000000) != 0;
			info->unused2		= (part1 & 0x20000000) != 0;
			info->unused3		= (part1 & 0x10000000) != 0;
			info->sizeAdj		= (part1 & 0x0F000000) >> 24;
			info->size			= (part1 & 0x00FFFFFF);
			info->lockCount		= (part2 & 0xF0000000) >> 28;
			info->owner			= (part2 & 0x0F000000) >> 24;
			info->hOffset		= (part2 & 0x00FFFFFF);

			// hOffset is a signed value; make sure we keep it that way
			// when we extract it by hand.

			if ((info->hOffset & 0x00800000) != 0)
			{
				info->hOffset |= 0xFF000000;
			}

			break;
		}

		default:
			assert (false);
	}
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetChunkHeader
 *
 * DESCRIPTION:	Given a pointer to the body of a memory chunk (that is,
 *				the user-accessible part of it), return a pointer to
 *				the chunk header.
 *
 *				This function does not assume anything about the
 *				pointer passed to it; it just subtracts a fixed amount
 *				from it.  This function *does* require a valid heap
 *				header, however, so that that fixed amount can be
 *				determined from the heap version number.
 *
 * PARAMETERS:	p - the pointer to the chunk body.
 *
 * RETURNED:	A pointer to the chunk header.
 *
 ***********************************************************************/

uaecptr PalmHeap::GetChunkHeader (VoidPtr p, const HeapInfo& heapInfo)
{
	uaecptr	chunkHdrPtr = ((uaecptr) p) - heapInfo.chunkHdrSize;

	return chunkHdrPtr;
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetChunkHeader
 *
 * DESCRIPTION:	Given a handle to the body of a memory chunk (that is,
 *				the user-accessible part of it), return a pointer to
 *				the chunk header.
 *
 *				This function does not assume anything about the
 *				handle passed to it; it just dereferences it and passes
 *				the resulting VoidPtr to the other version of
 *				GetChunkHeader.
 *
 * PARAMETERS:	h - the handle to the chunk.
 *
 * RETURNED:	A pointer to the chunk header.
 *
 ***********************************************************************/

uaecptr PalmHeap::GetChunkHeader (VoidHand h, const HeapInfo& heapInfo)
{
	return GetChunkHeader (GetPointer (h), heapInfo);
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::ValidateChunk
 *
 * DESCRIPTION:	Validate a chunk, ensuring that the data in the header
 *				appears to make sense.  Aspects that are checked:
 *
 *					- chunk is in the right heap.
 *					- chunk's size is not outrageous.
 *					- unused2 and unused3 flags are clear.
 *					- the hOffset field of an allocated, movable
 *					  chunk points to a pointer back to the chunk.
 *					- an allocated, non-movable chunk has the
 *					  appropriate lock-count.
 *
 * PARAMETERS:	heapInfo - a HeapInfo block on the heap that should
 *				contain the chunk.
 *
 *				chunkInfo - a ChunkInfo block containing the information
 *				to be validated.
 *
 *				validateHOffset - ensure that heapStart + hOffset is
 *				some sensible value before trying to use it.  Since
 *				this operation can be expensive (it involves making
 *				sure that the resulting pointer is in some master
 *				pointer table somewhere), it is optional.
 *
 * RETURNED:	An error code if an improper condition is detected.
 *
 ***********************************************************************/

PalmHeap::EPalmHeapResult PalmHeap::ValidateChunk (	const HeapInfo&		heapInfo,
													const ChunkInfo&	chunkInfo,
													Bool				validateHOffset)
{
	EPalmHeapResult	result;

	// Make sure the chunk is in the heap (this should always be true).

	if (chunkInfo.chunkHdrStart < heapInfo.heapHdrStart ||
		chunkInfo.chunkHdrStart > heapInfo.heapHdrStart + heapInfo.size)
	{
		return kCorruptedHeap_ChunkNotInHeap;
	}

	// Check the size.

	if (chunkInfo.chunkHdrStart + chunkInfo.size >
		heapInfo.heapHdrStart + heapInfo.size)
	{
		char	buffer[20];

		// !!! There's a problem here.  These variables are set before the
		// variable that uses them.  Thus, replacements based on these variables
		// are made before they're ready.  The result is that %chunk_size and
		// %chunk_max still appear in the final message.  I'm not sure how, but
		// this needs to be addressed and fixed...

		sprintf (buffer, "%0x08X", chunkInfo.size);
		Errors::SetParameter ("%chunk_size", buffer);

		sprintf (buffer, "%0x08X", heapInfo.size);
		Errors::SetParameter ("%chunk_max", buffer);

		return kCorruptedHeap_ChunkTooLarge;
	}

	if (chunkInfo.unused2 || chunkInfo.unused3)
	{
		return kCorruptedHeap_InvalidFlags;
	}

	if (!chunkInfo.free)
	{
		// If it's a movable chunk, validate that the handle offset
		// references the handle that points back to the chunk.

		if (chunkInfo.hOffset)
		{
			// Get the master pointer to this block.  If hOffset is bogus, then
			// this pointer may be bogus, too.

			uaecptr	h = chunkInfo.chunkHdrStart - chunkInfo.hOffset * 2;

			if (validateHOffset)
			{
				// Make sure it is in a master pointer block.

				// Get information about the first master pointer table.

				MPTInfo		mptInfo;
				GetMPTInfo (heapInfo.heapHdrStart + heapInfo.heapHdrSize, heapInfo.version, &mptInfo);

				// Loop over the MPT's.  Usually there's just one, but there could
				// be more (linked together in a single-linked list).

				while (1)
				{
					// Skip over the MPT header and get to the array of master pointers.

					uaecptr	mptArray = mptInfo.mptHdrStart + mptInfo.mptHdrSize;

					// See if "h" is in this array.  If so, break from the loop.

					if (h >= mptArray && h < mptArray + mptInfo.numEntries * sizeof (Ptr))
					{
						break;
					}

					// If there isn't another MPT, return with failure.

					if (mptInfo.nextTblOffset == 0)
					{
						return kCorruptedHeap_HOffsetNotInMPT;
					}

					// Get information on the next MPT.

					GetMPTInfo (heapInfo.heapHdrStart + mptInfo.nextTblOffset, heapInfo.version, &mptInfo);
				}
			}

			// "h" is in a master pointer table (or we've skipped the test). Make
			// sure "h" points back to the block it's supposed to.

			CEnableFullAccess	munge;	// Remove blocks on memory access.

			if (get_long (h) != chunkInfo.chunkHdrStart + chunkInfo.chunkHdrSize)
			{
				return kCorruptedHeap_HOffsetNotBackPointing;
			}
		}

		// If it's not a movable chunk, it must have a max lock count

		else
		{
			if (chunkInfo.lockCount != memPtrLockCount)
			{
				return kCorruptedHeap_InvalidLockCount;
			}
		}
	}

	// !!! Walk the heap and make sure that chunkHdrStart is found?

	return kChunkOK;
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::ValidatePointer
 *
 * DESCRIPTION:	Ensure that the given pointer points to the body of
 *				a memory manager chunk.
 *
 * PARAMETERS:	p - the pointer to test.
 *
 * RETURNED:	An error code if an improper condition is detected.
 *
 ***********************************************************************/

PalmHeap::EPalmHeapResult PalmHeap::ValidatePointer (VoidPtr p, const HeapInfo& heapInfo)
{
	// See if the pointer is NULL or not.

	if (!p)
		return kNULLPointerOrHandle;

	// Get the chunk header for this chunk.

	uaecptr		chunkHdr = GetChunkHeader (p, heapInfo);
	ChunkInfo	chunkInfo;
	GetChunkInfo (chunkHdr, heapInfo.version, &chunkInfo);

	// Validate the chunk.

	PalmHeap::EPalmHeapResult	result = ValidateChunk (heapInfo, chunkInfo, true);

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::ValidateHandle
 *
 * DESCRIPTION:	Ensure that the given handle is indeed a handle and that
 *				it references a valid memory manager chunk.
 *
 * PARAMETERS:	h - the handle to test.
 *
 * RETURNED:	An error code if an improper condition is detected.
 *
 ***********************************************************************/

PalmHeap::EPalmHeapResult PalmHeap::ValidateHandle (VoidHand h, const HeapInfo& heapInfo)
{
	// See if the handle is NULL or not.

	if (!h)
		return kNULLPointerOrHandle;

	// Dereference the handle to get the pointer.
	// !!! Need to validate this operation.

	VoidPtr	p = GetPointer (h);

	// Call the pointer-based version of this function to do the next
	// set of work.

	return ValidatePointer (p, heapInfo);
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::ValidateHeaps
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

void PalmHeap::ValidateHeaps (void)
{
	static Bool	validateInProgress;

	if (!validateInProgress)
	{
		validateInProgress = true;

		static long gValidateHeapsCounter;
		gValidateHeapsCounter++;

		UInt	numHeaps = GetNumHeaps (0);

		for (UInt heapID = 0; heapID < numHeaps; ++heapID)
		{
//			HeapInfo	info;
//			GetHeapInfo (heapID, &info);

			// !!! Need to validate the heap and the master pointer table, too.

			WalkHeap (heapID, ValidateChunkCallback, NULL);
		}

		validateInProgress = false;
	}
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::WalkHeap
 *
 * DESCRIPTION:	"Walk" the heap.  That is, iterate over the chunks in
 *				the heap, calling a user-defined callback function for
 *				each one.
 *
 *				If an error occurs while iterating (usually because of
 *				some heap corruption), an error dialog is presented.
 *
 * PARAMETERS:	heapID - the Palm OS ID of the heap to iterate over.
 *
 *				fn - a pointer to the function to call for each chunk.
 *				This function is passed a HeapInfo block, a ChunkInfo
 *				block, and the userData value.  The function should
 *				return either kKeepIterating or kStopIterating.
 *
 * RETURNED:	The result of the iteration.  If a heap corruption is
 *				detected while iterating, an error value is returned.
 *				Otherwise, a value indicating whether or not the full
 *				heap was iterated is returned.
 *
 ***********************************************************************/

PalmHeap::EPalmHeapResult PalmHeap::WalkHeap (UInt heapID, WalkerProc fn, void* userData)
{
	EPalmHeapResult	result;

	// Get information on the heap we're to iterate over.

	HeapInfo		heapInfo;
	GetHeapInfo (heapID, &heapInfo);

	uaecptr			prevChunkHdr = UAE_NULL;
	uaecptr			chunkHdr = heapInfo.firstChunk;
	ChunkInfo		chunkInfo;

	while (1)
	{
		// Get information about the current chunk.

		GetChunkInfo (chunkHdr, heapInfo.version, &chunkInfo);

		// If the size is zero, we've reached the sentinel at the end.

		if (!chunkInfo.size)
		{
			result = kWalkHeapFinished;
			break;
		}

		// See if this chunk looks valid.

		result = ValidateChunk (heapInfo, chunkInfo, true);
		if (result >= kWalkHeapError)
		{
			break;
		}

		// Call the callback proc for this chunk.  If it returns kStopIterating,
		// it's signalling that it wants us to stop iterating.

		if (fn (heapInfo, chunkInfo, userData) == kStopIterating)
		{
			result = kWalkHeapAborted;
			break;
		}

		// Go on to next chunk

		prevChunkHdr = chunkHdr;	// Keep this, for debugging.
		chunkHdr += chunkInfo.size;
	}

	// If an error occured, handle it now.

	if (result >= kWalkHeapError)
	{
		int	button = Errors::ReportCorruptedHeap (result);

		Emulator::HandleDlgButton (button, m68k_getpc ());
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetHeapContaining
 *
 * DESCRIPTION:	Return information about the heap containing the given
 *				test address.
 *
 * PARAMETERS:	testAddress - the address that should be contained
 *				by the returned heap.
 *
 * RETURNED:	A HeapInfo structure describing the heap.  If the heap
 *				could not be found, the HeapInfo members are all zero.
 *
 ***********************************************************************/

PalmHeap::HeapInfo PalmHeap::GetHeapContaining (uaecptr testAddress)
{
	HeapInfo	heapInfo;
	UInt		numHeaps = GetNumHeaps (0);

	for (UInt heapID = 0; heapID < numHeaps; ++heapID)
	{
		uaecptr	heapPtr = GetHeapPtr (heapID);

		if (heapPtr != UAE_NULL)
		{
			GetHeapInfo (heapID, &heapInfo);

			if (testAddress >= heapInfo.heapHdrStart &&
				testAddress < heapInfo.heapHdrStart + heapInfo.size)
			{
				return heapInfo;
			}
		}
	}

	memset (&heapInfo, 0, sizeof (heapInfo));

	return heapInfo;
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetChunkContaining
 *
 * DESCRIPTION:	Return information about the chunk containing the given
 *				test address.
 *
 * PARAMETERS:	testAddress - the address that should be contained by
 *				the returned chunk.
 *
 * RETURNED:	A ChunkInfo structure describing the chunk.  If the chunk
 *				could not be found, the ChunkInfo members are all zero.
 *
 ***********************************************************************/

PalmHeap::ChunkInfo PalmHeap::GetChunkContaining (uaecptr testAddress)
{
	GetChunkData	data;
	UInt			numHeaps = GetNumHeaps (0);

	for (UInt heapID = 0; heapID < numHeaps; ++heapID)
	{
		data.testAddress				= testAddress;
		data.chunkInfo.chunkHdrStart	= UAE_NULL;

		WalkHeap (heapID, GetChunkCallback, &data);

		if (data.chunkInfo.chunkHdrStart)
		{
			return data.chunkInfo;
		}
	}

	memset (&data.chunkInfo, 0, sizeof (data.chunkInfo));

	return data.chunkInfo;
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetNumHeaps
 *
 * DESCRIPTION:	Return the number of heaps created by the emulated
 *				Palm OS.
 *
 * PARAMETERS:	cardNo - the ID of the "card".
 *
 * RETURNED:	The number of heaps.
 *
 ***********************************************************************/

UInt PalmHeap::GetNumHeaps (UInt cardNo)
{
	return MemNumRAMHeaps (cardNo);
//	return 2;
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetHeapPtr
 *
 * DESCRIPTION:	Return the pointer to the heap header structure for
 *				the given heap.
 *
 * PARAMETERS:	heapID - the ID of the heap whose base pointer is to
 *				be returned.
 *
 * RETURNED:	The base pointer to the given heap.
 *
 ***********************************************************************/

uaecptr PalmHeap::GetHeapPtr (UInt heapID)
{
	assert (heapID < (sizeof (gHeapPtrs) / sizeof (gHeapPtrs [0])));

	if (gHeapPtrs[heapID] == UAE_NULL)
	{
		gHeapPtrs[heapID] = (uaecptr) MemHeapPtr (heapID);
	}

	return gHeapPtrs[heapID];

//	return (uaecptr) MemHeapPtr (heapID);
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetHeapVersion
 *
 * DESCRIPTION:	Determine and return the version number of the given
 *				heap.
 *
 * PARAMETERS:	heapPtr - a pointer to the header for the heap whose
 *				version is to be determined.
 *
 * RETURNED:	The version number of the heap.  Currently, these values
 *				are 1 (original heap format), 2 (allows for >64K heaps),
 *				and 3 (uses a free list for faster allocations).
 *
 ***********************************************************************/

int PalmHeap::GetHeapVersion (uaecptr heapPtr)
{
	CEnableFullAccess	munge;	// Remove blocks on memory access.

	int heapVersion;

	// Get the heap version.  If bit 14 is set, it's a version 3 heap.
	// If bit 15 is set, it's a version 2 heap.  Otherwise, it's a
	// version 1 heap.

	uae_u16	flags = get_word (heapPtr);
	
	if ((flags & memHeapFlagVers3) != 0)
	{
		heapVersion = 3;	// has free list
	}
	else if ((flags & memHeapFlagVers2) != 0)
	{
		heapVersion = 2;	// > 64K
	}
	else
	{
		heapVersion = 1;	// original
	}

	return heapVersion;
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetPointer
 *
 * DESCRIPTION:	Dereferences a handle and returns the pointer
 *
 * PARAMETERS:	h - the handle to dereference
 *
 * RETURNED:	The pointer associated with the handle.
 *
 ***********************************************************************/

VoidPtr PalmHeap::GetPointer (VoidHand h)
{
	if (!h)
		return NULL;

	CEnableFullAccess	munge;	// Remove blocks on memory access.

	h = memHandleUnProtect(h);
	return 	(VoidPtr) get_long ((uaecptr) h);
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::GetChunkCallback
 *
 * DESCRIPTION:	Callback function passed to WalkHeap by GetChunkContaining.
 *
 * PARAMETERS:	heapInfo, chunkInfo, userData - standard WalkerProc
 *				parameters.
 *
 * RETURNED:	A EWalkerProcResult indicating whether or not the
 *				iteration should continue.
 *
 ***********************************************************************/

PalmHeap::EWalkerProcResult PalmHeap::GetChunkCallback (	const HeapInfo& heapInfo,
															const ChunkInfo& chunkInfo,
															void* userData)
{
	GetChunkData*	data = (GetChunkData*) userData;

	// See if the test pointer is even in the heap.  If not,
	// we can quit right now.

	if ((data->testAddress < heapInfo.heapHdrStart) ||
		(data->testAddress >= heapInfo.heapHdrStart + heapInfo.size))
	{
		return kStopIterating;
	}

	// See if the test pointer is within the given chunk.
	// If so, stop iterating and return information on that chunk.

	uaecptr	bodyStart = chunkInfo.chunkHdrStart;	// + chunkInfo.chunkHdrSize;
	uaecptr	bodyEnd = chunkInfo.chunkHdrStart + chunkInfo.size;	// - chunkInfo.sizeAdj;

	if ((data->testAddress >= bodyStart) && (data->testAddress < bodyEnd))
	{
		data->chunkInfo = chunkInfo;
		return kStopIterating;
	}

	return kKeepIterating;
}


/***********************************************************************
 *
 * FUNCTION:	PalmHeap::ValidateChunkCallback
 *
 * DESCRIPTION:	Callback function passed to WalkHeap by ValidateHeaps.
 *
 * PARAMETERS:	heapInfo, chunkInfo, userData - standard WalkerProc
 *				parameters.
 *
 * RETURNED:	A EWalkerProcResult indicating whether or not the
 *				iteration should continue.
 *
 ***********************************************************************/

PalmHeap::EWalkerProcResult PalmHeap::ValidateChunkCallback (	const HeapInfo& heapInfo,
																const ChunkInfo& chunkInfo,
																void* userData)
{
	EPalmHeapResult	result = ValidateChunk (heapInfo, chunkInfo, true);

	if (result >= kWalkHeapError)
	{
		return kStopIterating;
	}

	return kKeepIterating;
}


