/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "RAM_ROM.h"

#include "Bank_DRAM.h"			// DRAMBank::Initialize
#include "Bank_Dummy.h"			// DummyBank::Initialize
#include "Bank_MC68328.h"		// MC68328Bank::Initialize
#include "Bank_MC68681.h"		// MC68681Bank::Initialize
#include "Bank_MC68EZ328.h"		// MC68EZ328Bank::Initialize


#include "Bank_ROM.h"			// ROMBank::Initialize
#include "Bank_SRAM.h"			// SRAMBank::Initialize
#include "Byteswapping.h"		// Canonical
#include "DebugMgr.h"			// Debug::CheckStepSpy
#include "ErrorHandling.h"		// kError_ stuff
#include "MetaMemory.h"			// MetaMemory
#include "Miscellaneous.h"		// NextPowerOf2
#include "Platform.h"			// AllocateMemory, DisposeMemory
#include "Strings.r.h"			// kStr_ values
#include "TrapPatches.h"		// RemoveAllTailpatches, InstallAllTailpatches, gSysDozing, etc.

#include "Crc.h"				// Crc16CalcBigBlock

#if HAS_PROFILING
#include "Profiling.h"		// profiling stuff
#endif


/*
	Hitchhiker's Guide To Accessing Memory

	When emulating CPU instructions, all memory access must be emulated
	as well.  In addition, the memory location that the emulated instruction
	is accessing is not necessarily the same as the memory location in
	the emulator's address space.  That is, memory location 0x00001234 in
	the emulated process's address space is not the same as memory location
	0x00001234 in the emulator's address space.  Instead, the emulated
	0x00001234 is actually 0x00001234 bytes into some buffer that we've
	allocate for the purpose of holding the emulated address space's
	contents.

	There are several ranges of Palm OS memory that we need to emulate:

		- Low-memory RAM (for the dynamic heap)
		- High-memory RAM (for the storage heap)
		- ROM
		- Dragonball registers
		- Everything else

	Each of these ranges (except for the last one) is managed by a set of
	functions for reading from and writing to that memory and a chunk of
	memory that we allocate as a "backing store" for that memory.
	
	Thus, for each of these ranges, we need to quickly determine what set
	of functions need to be called.  The UAE-prescribed method for doing this
	is to divide the entire 4GB address space into 64K banks, each 64K long.

	Each 64K bank is managed by a set of functions.  These functions
	handle reading from and writing to that 64K bank, handling 1-byte, 2-byte,
	and 4-byte accesses.  They also manage mapping from the emulated addres
	space to the physical space as viewed by the emulator, and validate
	memory addresses.

	Because there are 64K of these 64K banks, the functions that manage
	each bank are stored in an array of 64K elements.  When we need to invoke
	a function for a particular bank, the upper word of the prospective
	memory address is used as a "bank index" into this array.  From that,
	the right set of functions is found, and the desired function from that
	set of functions is called.

	This 64K array is very sparse.  There are a one or two entries for
	managing the dynamic heap (depending on whether the dynamic heap is
	64K, 96K or 128K), 2 to 128 entries for the storage heap (depending on
	whether the storage heap is 128K to 8Meg), 8 to 32 entries for the ROM
	(again, depending on its size), and 1 for the Dragonball registers.
	That leaves the remaining 64K - 12 entries to be managed by a set of
	functions that do little but signal an address error if the memory they
	manage is being accessed.

	Pictorally, it looks like this:

	Emulated 4G space:				Memory function array							Address Bank functions

	0x00000000	+--------------+	+--------------------------------+
				|      64K     |<---| Ptr to fns that mng this range |------+
	0x00010000	+--------------+	+--------------------------------+		|
				|      64K     |<---| Ptr to fns that mng this range |------+
	0x00020000	+--------------+	+--------------------------------+		|
				|      64K     |											+----->	DRAM::GetLong, DRAM::SetLong
				\/\/\/\/\/\/\/\/													DRAM::GetWord, DRAM::SetWord
				\/\/\/\/\/\/\/\/													DRAM::GetByte, DRAM::SetByte
				|      64K     |													etc.
	0x10000000	+--------------+	+--------------------------------+
				|      64K     |<---| Ptr to fns that mng this range |------+
	0x10010000	+--------------+	+--------------------------------+		|
				|      64K     |<---| Ptr to fns that mng this range |------+
	0x10020000	+--------------+	+--------------------------------+		|
				|      64K     |											+----->	SRAM::GetLong, SRAM::SetLong
				\/\/\/\/\/\/\/\/													SRAM::GetWord, SRAM::SetWord
				\/\/\/\/\/\/\/\/													SRAM::GetByte, SRAM::SetByte
				|      64K     |													etc.
	0x10C00000	+--------------+	+--------------------------------+
				|      64K     |<---| Ptr to fns that mng this range |------+
	0x10C10000	+--------------+	+--------------------------------+		|
				|      64K     |<---| Ptr to fns that mng this range |------+
	0x10C20000	+--------------+	+--------------------------------+		|
				|      64K     |											+----->	ROM::GetLong, ROM::SetLong
				\/\/\/\/\/\/\/\/													ROM::GetWord, ROM::SetWord
				\/\/\/\/\/\/\/\/													ROM::GetByte, ROM::SetByte
				|      64K     |													etc.
	0xFFFE0000	+--------------+	+--------------------------------+
				|      64K     |<---| Ptr to fns that mng this range |
	0xFFFF0000	+--------------+	+--------------------------------+
				|      64K     |<---| Ptr to fns that mng this range |------+
	0xFFFFFFFF	+--------------+	+--------------------------------+		|
																			+----->	Regs::GetLong, Regs::SetLong
																					Regs::GetWord, Regs::SetWord
																					Regs::GetByte, Regs::SetByte
																					etc.

	All of the remaining memory function array entries are initialized with
	pointers to the Dummy::GetLong, Dummy::SetLong, etc., set of functions.

	As an extension to this UAE model, we have added the facility for mapping
	memory from the emulator's address space directly into the emulated address
	space.  Thus, if we have some data at memory location 0x01234567 (for
	instance, some data from a .prc file that we're installing), we can make
	that data also show up at memory location 0x01234567 within the emulated
	address space.  This mechanism makes it easy for us to "inject" data
	into the emulated address space.  It also makes it easier to call emulated
	ROM functions as subroutines.  Some ROM functions take as parameters
	pointers to memory locations that are to receive values (for example,
	pointers to local variables).  For those functions, we just map the
	emulator's stack into the emulated address space.  After that's done,
	we can pass pointers to stack-based variables with impunity.

	Mapping from emulated addresses to "real" addresses (that is, the corresponding
	memory locations within the buffers we've allocated to "stand in" for the
	emulated memory) is performed via UAE-defined macros, described below.

	Executive summary:

		- use get_foo if you want the address checked for even/oddness.
		- use fooget to get the stuff that jumps through the address banks
		- use do_get_mem_foo to actually get the value using a physical address.

	get_long, get_word, get_byte, put_long, put_word, put_byte
		Inline functions that check for bus error (if necessary) and
		then either call one of the following or handle the bus error.

	longget_1, wordget_1, byteget_1, longput_1, wordput_1, byteput_1
		If MD_HAVE_MEM_1_FUNCS is not defined, these are just macros
		to the following functions.  Otherwise, the machdep/memory.h
		file is assumed to have defined them.

	longget, wordget, byteget, longput, wordput, byteput
		If NO_INLINE_MEMORY_ACCESS is not defined, these are macros
		to "call_mem_get_func(get_mem_bank(addr).lget, addr)", etc.
		Otherwise, they are out-of-line functions.  machdep_memory.h
		undefs this, so we get the inlined functions.

	get_mem_bank
		Gets the right addrbank object, with the help of the bankindex
		macro.

	bankindex
		Returns a bank index for the given address.

	call_mem_get_func
		Merely calls the given function through the addrbank fn ptr.

	check_addr
		Macro called by get_long, etc., to check the address for a
		bus error.  Always returns true (OK) if NO_EXCEPTION_3 is
		defined or CPU_LEVEL indicates 68020 or better.  Otherwise
		returns false (error) if low-bit of address is set. config.h
		defines NO_EXCEPTION_3, so this stubs to "true".

	do_get_mem_long, do_get_mem_word, do_get_mem_byte,
	do_put_mem_long, do_put_mem_word, do_put_mem_byte
		Very low-level memory access. They return the value at the
		given physical location, byte-swapping if needed.

	m68k_setpc
		Given a virtual address, sets regs.pc and updates regs.pc_p
		to point to hold the physical address.

	m68k_getpc
		Returns the current virtual address (which is not always
		up-to-date in regs.pc, but _is_ up-to-date in regs.pc_p).

	nextilong, nextiword, nextibyte
		Get the value pointed to by the pc and then bump the pc.
*/

// Types.

#pragma mark Types

// Snippet from SystemPrv.h

/************************************************************
 * Common Equates between Native and Emulation mode
 *************************************************************/
#define	sysCardSignature		0xFEEDBEEFL		// card signature long word
#define sysStoreSignature		0xFEEDFACEL		// store signature long word


// Private function declarations


class Card
{
	public:
		static Bool		CheckCardHeader			(const CardHeaderType& cardHdr);
		static Bool		CheckChecksum			(const void* p, uae_u32 romLength);
		static Bool		Supports328				(const void* p);
		static Bool		SupportsEZ				(const void* p);
};


// Globals.

#pragma mark Globals

uae_u32			gRAMBank_Size;
uae_u32			gROMBank_Size;

uae_u32			gRAMBank_Mask;
uae_u32			gROMBank_Mask;

char*			gRAM_Memory;
char*			gROM_Memory;

int				buserr;					// (normally defined in memory.c)
AddressBank*	mem_banks[65536];		// (normally defined in memory.c)

MemAccessFlags	gMemAccessFlags =
{
	MASTER_RUNTIME_VALIDATE_SWITCH,
	MASTER_RUNTIME_VALIDATE_SWITCH,
	MASTER_RUNTIME_VALIDATE_SWITCH,
	MASTER_RUNTIME_VALIDATE_SWITCH,
	MASTER_RUNTIME_VALIDATE_SWITCH,
	MASTER_RUNTIME_VALIDATE_SWITCH,
	MASTER_RUNTIME_VALIDATE_SWITCH,
	MASTER_RUNTIME_VALIDATE_SWITCH,
	MASTER_RUNTIME_VALIDATE_SWITCH,
	MASTER_RUNTIME_VALIDATE_SWITCH,

	// User prevention flags.

	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH,
	false,	// SRAM-get
	true,	// SRAM-set
	false,	// ROM-get
	false,	// ROM-set (HACK: We really want it to be true!)
	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH,

	false,	// fCheck_UserChunkGet
	false,	// fCheck_UserChunkSet
	false,	// fCheck_SysChunkGet
	false,	// fCheck_SysChunkSet

	// System prevention flags.

	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH,
	false,	// SRAM-get
	true,	// SRAM-set
	false,	// ROM-get
	false,	// ROM-set (HACK: We really want it to be true!)
	MASTER_RUNTIME_PREVENT_SWITCH,
	MASTER_RUNTIME_PREVENT_SWITCH
};

MemAccessFlags	kZeroMemAccessFlags;


#if defined (_DEBUG)

	// In debug mode, define a global variable that points to the
	// Palm ROM's low-memory globals.  That makes it easier to find
	// out what's going wrong when something goes wrong.

	LowMemHdrType*	gLowMemory;

#endif

#if PROFILE_MEMORY
/*
	After 12 Million instructions executed...

	kDRAMLongRead,		// 2,797,111
	kDRAMLongRead2,		//
	kDRAMWordRead,		// 1,334,295
	kDRAMByteRead,		//   427,936

	kDRAMLongWrite,		// 1,960,933
	kDRAMLongWrite2,	//
	kDRAMWordWrite,		//   800,780
	kDRAMByteWrite,		//   213,029

	kSRAMLongRead,		//     4,167
	kSRAMLongRead2,		//     4,167
	kSRAMWordRead,		//    10,768
	kSRAMByteRead,		//     6,295 (jumped WAY up (5-fold) after playing with Address Book / 20 Million instructions)

	kSRAMLongWrite,		//   516,246 (stable after boot?)
	kSRAMLongWrite2,	//
	kSRAMWordWrite,		//     5,112
	kSRAMByteWrite,		//     2,183

	kROMLongRead,		//    10,156
	kROMLongRead2,		//
	kROMWordRead,		//   557,062
	kROMByteRead,		//    15,516

	kROMLongWrite,		//         0
	kROMLongWrite2,		//
	kROMWordWrite,		//         0
	kROMByteWrite,		//         0

	kREGLongRead,		//     16,638
	kREGLongRead2,		//
	kREGWordRead,		//     30,477
	kREGByteRead,		//      3,553

	kREGLongWrite,		//      9,472
	kREGLongWrite2,		//
	kREGWordWrite,		//     20,709
	kREGByteWrite,		//     12,656


	After 32,439,154 instructions:

	kDRAMLongRead,		// 6,466,064
	kDRAMWordRead,		// 3,342,826
	kDRAMByteRead,		//   940,345

	kDRAMLongWrite,		// 4,136,635
	kDRAMWordWrite,		// 1,462,995
	kDRAMByteWrite,		//   509,189

	kSRAMLongRead,		//    24,355
	kSRAMWordRead,		//    32,190
	kSRAMByteRead,		//    75,917

	kSRAMLongWrite,		// 2,074,137 (8-Meg ROM)
	kSRAMWordWrite,		//    42,050
	kSRAMByteWrite,		//    11,292

	kROMLongRead,		//    25,806
	kROMWordRead,		// 1,302,897
	kROMByteRead,		//    63,590

	kROMLongWrite,		//         0
	kROMWordWrite,		//         0
	kROMByteWrite,		//         0

	kREGLongRead,		//    32,037
	kREGWordRead,		//    74,299
	kREGByteRead,		//    91,930

	kREGLongWrite,		//    20,358
	kREGWordWrite,		//    57,879
	kREGByteWrite,		//    22,592
*/

long	gMemoryAccess[kLastEnum];
#endif


#pragma mark -


// ===========================================================================
//		 Memory
// ===========================================================================

// ---------------------------------------------------------------------------
//		 Memory::Initialize
// ---------------------------------------------------------------------------
// Initializes the RAM, ROM, and special memory areas of the emulator.

void Memory::Initialize (	FileReference&	iRAMFileRef,
							FileReference&	iROMFileRef,
							uae_s32			iRAMSize)
{
	assert (gRAM_Memory == NULL);
	assert (gROM_Memory == NULL);

	// Reset all our variables.

	gRAMBank_Size = 0;
	gROMBank_Size = 0;

	gRAMBank_Mask = 0;
	gROMBank_Mask = 0;

	gRAM_Memory = NULL;
	gRAM_Memory = NULL;

	memset (mem_banks, 0, sizeof (mem_banks));

	try
	{
		// Load the ROM file.

		Memory::InitializeROM (iROMFileRef);

		// Load the RAM file.

		Memory::InitializeRAM (iRAMFileRef, iRAMSize);

		// Initialize the shadow memory bits.

		MetaMemory::Initialize (gRAMBank_Size);

		// Clear the memory banks.

		DummyBank::Initialize ();

		// Initialize the valid memory banks.

		// Memory banks 0x1000 to <mumble> are managed byt the functions
		// in SRAMBank.  <mumble> is based on the amount of RAM being emulated.

		SRAMBank::Initialize ();

		// Memory banks 0 and 1 are managed by the functions in DRAMBank.
		// Initialize this after initializing the SRAMBank. The order is
		// important for DragonballEZ. On Dragonball devices, DRAM is
		// at 0x00000000, and SRAM is at 0x10000000. But on EZ devices,
		// both start at 0x00000000. By calling DRAMBank::Initialize
		// second, we allow it to overwrite the AddressBank handlers
		// for the part of memory where they overlap.

		DRAMBank::Initialize ();

		// Memory banks 0x10C0 to <mumble> are managed byt the functions
		// in ROMBank.  <mumble> is based on the amount of ROM being emulated.

		ROMBank::Initialize ();

		// If the device has Flash, re-initialize the ROM range for that.

		uae_u32	device = Emulator::GetHardwareDevice ();

		if (device == Emulator::kPalmVII ||
			device == Emulator::kPalmV ||
			device == Emulator::kPalmIIIx)
		{
			FlashBank::Initialize ();
		}

		// Memory bank 0x10E0 is managed by the functions in MC68681Bank.
		// (Note: taken out because we don't have any ROMs that use the
		// 68681 any more. The Debug 2.0 ROM used to, but we've now posted
		// a version that doesn't.)

//		MC68681Bank::Initialize ();

		// Memory bank 0xFFFF is managed by the functions in MC68328Bank.

		if (Emulator::EZMode ())
			MC68EZ328Bank::Initialize ();
		else
			MC68328Bank::Initialize ();
			
	}
	catch (...)
	{
		Memory::Dispose ();
		throw;
	}

	assert (gRAM_Memory != NULL);
	assert (gROM_Memory != NULL);
}


// ---------------------------------------------------------------------------
//		 Memory::Reset
// ---------------------------------------------------------------------------

void Memory::Reset (void)
{
	buserr = 0;

	MetaMemory::Reset ();
}


// ---------------------------------------------------------------------------
//		 Memory::Dispose
// ---------------------------------------------------------------------------
// Deallocates the memory buffers created by Memory::Initialize.

void Memory::Dispose (void)
{
	Platform::DisposeMemory (gRAM_Memory);
	Platform::DisposeMemory (gROM_Memory);

	MetaMemory::Dispose ();
}


// ---------------------------------------------------------------------------
//		 Memory::Save
// ---------------------------------------------------------------------------
// Saves the current RAM image to a file. iMemoryFile is a reference to a
// file where the memory should be saved. 

void Memory::Save (FileHandle& ramFile, const AppPreferences::Configuration& cfg)
{
	assert (gRAM_Memory);

	// Temporarily remove any munging we may have done.

	if (!Debug::InDebugger ())
	{
		Patches::RemoveAllTailpatches ();
		Debug::RemoveBreakpoints ();
	}

	// Write the file.

	SessionFile::WriteRAM (ramFile);

	// Restore memory munging.

	if (!Debug::InDebugger ())
	{
		Debug::InstallBreakpoints ();
		Patches::InstallAllTailpatches ();
	}


	// Write out the device type.

	assert (cfg.fDeviceType != Emulator::kUnspecified);
	SessionFile::WriteInt (ramFile, SessionFile::kDeviceType, cfg.fDeviceType);

	// Write out the silkscreen type.

	assert (cfg.fSilkscreenType != AppPreferences::kSilkscreenUnspecified);
	SessionFile::WriteInt (ramFile, SessionFile::kSilkscreenType, cfg.fSilkscreenType);

	// Save the name of the ROM file.

	string	romFileName = cfg.fROMFile.GetFileName ();
	SessionFile::WriteChunk (ramFile, SessionFile::kROMNameTag,
							romFileName.size (), romFileName.c_str ());
}


// ---------------------------------------------------------------------------
//		 Memory::InitializeBanks
// ---------------------------------------------------------------------------
// Initializes the specified memory banks with the given data.

void Memory::InitializeBanks (	AddressBank&	iBankInitializer,
								uae_s32			iStartingBankIndex,
								uae_s32			iNumberOfBanks)
{
	for (uae_s32 aBankIndex = iStartingBankIndex;
		aBankIndex < iStartingBankIndex + iNumberOfBanks;
		aBankIndex++)
	{
		mem_banks[aBankIndex] = &iBankInitializer;
	}
}


// ---------------------------------------------------------------------------
//		 Memory::InitializeROM
// ---------------------------------------------------------------------------
// Loads the ROM file into memory and initializes it.

void Memory::InitializeROM (FileReference& iROMFileRef)
{
	// Open the file.

	FileHandle	romFile (iROMFileRef, kOpenExisting | kOpenRead);

	// Get the file size.

	uae_s32	fileSize = romFile.GetLength ();

	// Make sure the file is big enough to have a card header.

	if (fileSize < sizeof (CardHeaderType))
		Errors::Throw (kError_BadROM);

	// Read the card header.

	CardHeaderType	cardHeader;
	romFile.Read (sizeof (CardHeaderType), &cardHeader);
	Canonical (cardHeader);

	// Validate the card header.

	if (!Card::CheckCardHeader (cardHeader))
		Errors::Throw (kError_BadROM);

	// The ROM is made up of two parts: a Big ROM and a Small ROM.  ROM image
	// files can contain both parts or just the Big ROM part.  If it contains
	// both parts, then everything is just fine, and we can use it no problem.
	// If the Small ROM part is missing, we have to detect that, dummy up a
	// small ROM for correct operation, and relocate the Big ROM higher in memory.

	// Both the Big ROM and Small ROM start off with a CardHeaderType struct.
	// We just read in one of them.  Let's look at where the Big ROM _should_
	// be to see if there's another CardHeader there.  If so, we have both parts.
	// If not, we have just the Big ROM.

	uae_u32	bigROMOffset;

	if (cardHeader.hdrVersion == 1)
	{
		bigROMOffset = 0x3000;
	}
	else
	{
		// Version 2 CardHeaders contain a "bigROMOffset" field that contains the
		// offset from the beginning of the card to where the ROM should
		// appear in memory.  Normally the card appears at 0x10000000 (which
		// would make bigROMOffset something like 0x00C08000), but that
		// isn't always the case.  The problem is, we don't always know
		// where the card beginning will be.  For now, we'll assume that
		// the ROM wants to live at 0x10C00000, and that the Big ROM will
		// want to begin somewhere soon after that.  This means we can get by
		// with using only the lower bits of bigROMOffset.

		bigROMOffset = cardHeader.bigROMOffset;
		bigROMOffset &= 0x000FFFFF;		// Allows for 1 Meg offset.
	}

	// Make sure the file is big enough to have a Big ROM.

	if (fileSize < bigROMOffset)
		Errors::Throw (kError_BadROM);

	// We'll use bigROMOffset for later to get to some Big ROM data.
	// But we'll also need a value when we read the ROM image into
	// memory.  This value will be used as an offset into a buffer
	// where we'll be reading the ROM image.  If we have a Small ROM,
	// the offset will be zero.  If we have only a Big ROM, the offset
	// will be the same value as bigROMOffset.

	uae_u32	bufferOffset = bigROMOffset;

	// See if there's a card header there, too.

	CardHeaderType	cardHeader2;
	romFile.SetPos (bigROMOffset, kSeekSet);
	romFile.Read (sizeof (CardHeaderType), &cardHeader2);
	Canonical (cardHeader2);

	if (Card::CheckCardHeader (cardHeader2))
	{
		// Looks like, so we don't have to relocate the ROM image.

		bufferOffset = 0;
	}

	// The ROM size is now the size of the file plus any offset for
	// any small ROM that we have to add and dummy up.

	uae_u32	romSize = fileSize + bufferOffset;

	// Read in the ROM image.

	char*	romImage = (char*) Platform::AllocateMemory (romSize);

	romFile.SetPos (0, kSeekSet);
	romFile.Read (fileSize, romImage + bufferOffset);

	// If we only had a Big ROM, dummy up the Small ROM.  All we really
	// need to do here is copy the Big ROM's card header to the Small
	// ROM area.
	//
	// Also, clear out this area to 0xFF to look like new Flash RAM.

	if (bufferOffset)
	{
		memset (romImage, 0xFF, bigROMOffset);

		*(CardHeaderType*) romImage = *(CardHeaderType*) (romImage + bigROMOffset);
	}

	// See if the checksum looks OK.

	if (!Card::CheckChecksum (romImage + bigROMOffset,romSize - bigROMOffset))
		Errors::Throw (kError_BadROM);

	// Tell the emulator what kind of ROM we've just loaded.

	if (Card::SupportsEZ (romImage))
		Emulator::SetHardwareDevice (Emulator::kPalmV);
	else
		Emulator::SetHardwareDevice (Emulator::kPalmPilotProfessional);

	// Byteswap all the words in the ROM (if necessary). Most accesses
	// are 16-bit accesses, so we optimize for that case.

	ByteswapWords (romImage, romSize);

	// Everything seems to be OK.  Save the ROM data in some global
	// variables for the CPU emulator to access. Make sure that
	// gROMBank_Size is a power-of-2. The ROMBank memory  routines
	// require this.

	gROM_Memory		= romImage;
	gROMBank_Size	= ::NextPowerOf2 (romSize);
	gROMBank_Mask	= gROMBank_Size - 1;
}


// ---------------------------------------------------------------------------
//		 Memory::InitializeRAM
// ---------------------------------------------------------------------------
// Loads a RAM image from a file.  Note: This function should only be called
// from Memory::Initialize.

void Memory::InitializeRAM (FileReference& iRAMFileRef, uae_s32 iRAMSize)
{
	// If a RAM file's been specified, load it.

	if (iRAMFileRef.IsSpecified ())
	{
		FileHandle	ramFile (iRAMFileRef, kOpenExisting | kOpenRead);
		SessionFile::ReadRAM (ramFile);
	}

	// If a RAM file was NOT specified or the attempt to load it fails, just
	// create a new buffer of the specified size.

	if (!gRAM_Memory)
	{
		gRAMBank_Size	= iRAMSize * 1024;
		gRAM_Memory		= (char*) Platform::AllocateMemory (gRAMBank_Size, true);
	}
	
	gRAMBank_Mask = gRAMBank_Size - 1;

#if defined (_DEBUG)
	// In debug mode, define a global variable that points to the
	// Palm ROM's low-memory globals.  That makes it easier to find
	// out what's doing wrong when something goes wrong.

	gLowMemory = (LowMemHdrType*) gRAM_Memory;
#endif
}


// ---------------------------------------------------------------------------
//		 Memory::MapPhysicalMemory
// ---------------------------------------------------------------------------
// Maps a range of physical memory to appear at the same location of the
// emulated Palm OS's virtual memory.

void Memory::MapPhysicalMemory (const void* addr, uae_u32 size)
{
	DummyBank::MapPhysicalMemory (addr, size);
}


// ---------------------------------------------------------------------------
//		 Memory::UnmapPhysicalMemory
// ---------------------------------------------------------------------------
// Unmaps a range of physical memory from appearing at the same location of
// the emulated Palm OS's virtual memory.

void Memory::UnmapPhysicalMemory (const void* addr)
{
	DummyBank::UnmapPhysicalMemory (addr);
}


// ---------------------------------------------------------------------------
//		 Memory::GetMappingInfo
// ---------------------------------------------------------------------------

void Memory::GetMappingInfo (const void* addr, void** start, uae_u32* len)
{
	DummyBank::GetMappingInfo (addr, start, len);
}


// ---------------------------------------------------------------------------
//		 Memory::PreventedAccess
// ---------------------------------------------------------------------------

void Memory::PreventedAccess (uaecptr iAddress, long size, Bool forRead, Errors::EAccessType kind)
{
	int button = Errors::ReportPreventedAccess (iAddress, size, forRead, kind);

	// (Exception is thrown if button == kDebug or kReset).

	Emulator::HandleDlgButton (button, m68k_getpc ());
}



#pragma mark -

// ===========================================================================
//		 SessionFile
// ===========================================================================

// ---------------------------------------------------------------------------
//		 SessionFile::ReadChunk
// ---------------------------------------------------------------------------

void SessionFile::ReadChunk (FileHandle& f, uae_u32 size, void* data)
{
	f.Read (size, data);
}


// ---------------------------------------------------------------------------
//		 SessionFile::WriteChunk
// ---------------------------------------------------------------------------

void SessionFile::WriteChunk (FileHandle& f, uae_u32 tag, uae_u32 size, const void* data)
{
	Canonical (tag);
	f.Write (sizeof (tag), &tag);

	Canonical (size);
	f.Write (sizeof (size), &size);
	Canonical (size);

	f.Write (size, data);
}


// ---------------------------------------------------------------------------
//		 SessionFile::FindChunk
// ---------------------------------------------------------------------------

long SessionFile::FindChunk (FileHandle& f, uae_u32 targetTag)
{
	long	len = -1;
	long	fileOffset = 0;
	long	fileLength = 0;

	fileLength = f.GetLength ();
	
	f.SetPos (0, kSeekSet);
	
	while (fileOffset < fileLength)
	{
		uae_u32	chunkTag;
		long	chunkLen;

		f.Read (sizeof (chunkTag), &chunkTag);
		f.Read (sizeof (chunkLen), &chunkLen);

		Canonical (chunkTag);
		Canonical (chunkLen);

		if (chunkTag == targetTag)
		{
			len = chunkLen;
			break;
		}

		f.SetPos (chunkLen, kSeekCur);
		fileOffset += sizeof (chunkTag) + sizeof (chunkLen) + chunkLen;
	}

	return len;
}

// ---------------------------------------------------------------------------
//		 SessionFile::ReadRAM
// ---------------------------------------------------------------------------

Bool SessionFile::ReadRAM (FileHandle& f)
{
	// Get the size of the file.

	gRAMBank_Size = SessionFile::FindChunk (f, kRAMDataTag);
	if (gRAMBank_Size == -1)
	{
		gRAMBank_Size = 0;
		return false;
	}

	// Use it to create a RAM buffer.

	gRAM_Memory		= (char*) Platform::AllocateMemory (gRAMBank_Size, false);

	// Read the RAM data into memory.

	SessionFile::ReadChunk (f, gRAMBank_Size, gRAM_Memory);

	// Byteswap the words (if necessary).

	ByteswapWords (gRAM_Memory, gRAMBank_Size);
	
	return true;
}


// ---------------------------------------------------------------------------
//		 SessionFile::WriteRAM
// ---------------------------------------------------------------------------

void SessionFile::WriteRAM (FileHandle& f)
{
	// Unbyteswap the words (if necessary). We save everything in
	// Big-Endian format.

	ByteswapWords (gRAM_Memory, gRAMBank_Size);

	try
	{
		// Write RAM image.

		SessionFile::WriteChunk (f, kRAMDataTag, gRAMBank_Size, gRAM_Memory);
	}
	catch (...)
	{
		// Rebyteswap the words (if necessary).

		ByteswapWords (gRAM_Memory, gRAMBank_Size);

		// Rethrow the exception.

		throw;
	}

	// Rebyteswap the words (if necessary).

	ByteswapWords (gRAM_Memory, gRAMBank_Size);
}


// ---------------------------------------------------------------------------
//		 SessionFile::ReadInt, et al.
// ---------------------------------------------------------------------------

Bool SessionFile::ReadInt (FileHandle& f, uae_u32 tag, uae_u8& val)
{
	if (SessionFile::FindChunk (f, tag) == -1)
		return false;

	SessionFile::ReadChunk (f, sizeof (val), &val);
	Canonical (val);

	return true;
}


Bool SessionFile::ReadInt (FileHandle& f, uae_u32 tag, uae_s8& val)
{
	if (SessionFile::FindChunk (f, tag) == -1)
		return false;

	SessionFile::ReadChunk (f, sizeof (val), &val);
	Canonical (val);

	return true;
}


Bool SessionFile::ReadInt (FileHandle& f, uae_u32 tag, uae_u16& val)
{
	if (SessionFile::FindChunk (f, tag) == -1)
		return false;

	SessionFile::ReadChunk (f, sizeof (val), &val);
	Canonical (val);

	return true;
}


Bool SessionFile::ReadInt (FileHandle& f, uae_u32 tag, uae_s16& val)
{
	if (SessionFile::FindChunk (f, tag) == -1)
		return false;

	SessionFile::ReadChunk (f, sizeof (val), &val);
	Canonical (val);

	return true;
}


Bool SessionFile::ReadInt (FileHandle& f, uae_u32 tag, uae_u32& val)
{
	if (SessionFile::FindChunk (f, tag) == -1)
		return false;

	SessionFile::ReadChunk (f, sizeof (val), &val);
	Canonical (val);

	return true;
}


Bool SessionFile::ReadInt (FileHandle& f, uae_u32 tag, uae_s32& val)
{
	if (SessionFile::FindChunk (f, tag) == -1)
		return false;

	SessionFile::ReadChunk (f, sizeof (val), &val);
	Canonical (val);

	return true;
}


Bool SessionFile::ReadString (FileHandle& f, uae_u32 tag, char* s)
{
	long	len = SessionFile::FindChunk (f, tag);
	if (len == -1)
		return false;

	SessionFile::ReadChunk (f, len, s);

	return true;
}


// ---------------------------------------------------------------------------
//		 SessionFile::WriteInt, et al.
// ---------------------------------------------------------------------------

void SessionFile::WriteInt (FileHandle& f, uae_u32 tag, uae_u8 val)
{
	Canonical (val);
	SessionFile::WriteChunk (f, tag, sizeof (val), &val);
}


void SessionFile::WriteInt (FileHandle& f, uae_u32 tag, uae_s8 val)
{
	Canonical (val);
	SessionFile::WriteChunk (f, tag, sizeof (val), &val);
}


void SessionFile::WriteInt (FileHandle& f, uae_u32 tag, uae_u16 val)
{
	Canonical (val);
	SessionFile::WriteChunk (f, tag, sizeof (val), &val);
}


void SessionFile::WriteInt (FileHandle& f, uae_u32 tag, uae_s16 val)
{
	Canonical (val);
	SessionFile::WriteChunk (f, tag, sizeof (val), &val);
}


void SessionFile::WriteInt (FileHandle& f, uae_u32 tag, uae_u32 val)
{
	Canonical (val);
	SessionFile::WriteChunk (f, tag, sizeof (val), &val);
}


void SessionFile::WriteInt (FileHandle& f, uae_u32 tag, uae_s32 val)
{
	Canonical (val);
	SessionFile::WriteChunk (f, tag, sizeof (val), &val);
}


void SessionFile::WriteString (FileHandle& f, uae_u32 tag, const char* s)
{
	SessionFile::WriteChunk (f, tag, strlen (s), s);
}



#pragma mark -

// ===========================================================================
//		 Card
// ===========================================================================

// ---------------------------------------------------------------------------
//		 Card::CheckCardHeader
// ---------------------------------------------------------------------------
// Determines if the memory pointed to looks like a card header.

Bool Card::CheckCardHeader (const CardHeaderType& cardHdr)
{
	// Make sure the stack size isn't something ludicrous.

	if (cardHdr.initStack == 0 || cardHdr.initStack > 0x000FFFFF)	// 1 Meg
		return false;

	// Make sure the signature is right.

	if (cardHdr.signature != sysCardSignature)
		return false;

	// Make sure the header version isn't ludicrous.

	if (cardHdr.hdrVersion == 0 || cardHdr.hdrVersion > 255)	// arbitrary value
		return false;

	// Make sure this isn't a "RAM only" card.

	if (cardHdr.flags & memCardHeaderFlagRAMOnly)
		return false;

	return true;
}


// ---------------------------------------------------------------------------
//		 Card::CheckChecksum
// ---------------------------------------------------------------------------
// Determines if the memory pointed to looks like a card header.

Bool Card::CheckChecksum (const void* p, uae_u32 fileLength)
{
	CardHeaderType	cardHdr = *(CardHeaderType*) p;

	Canonical (cardHdr);

	if (cardHdr.hdrVersion == 1)
	{
		static const Word	kDebug20 = 0x9BED;
		static const Word	kROMChecksums[] =
		{
			0x67C8,		// 1.0 - English - v30 ROMs - ROM5Rel - 1/9/96
			0x83FC,		// 1.0 - French - v10 ROMs - ROM5Rel - 3/28/96
			0xE1D9,		// 1.0 - French - v11 ROMs - ROM5Rel - 4/4/96
			0x41BE,		// 1.0 - French - v13 ROMs - ROM5Rel - 4/11/96
			0xC2AE,		// 1.0 - German - v10 ROMs - ROM5Rel - 4/1/96
			0x1FCE,		// 1.0 - German - v12 ROMs - ROM5Rel - 4/10/96
			0x3D33,		// 1.0 - German - v13 ROMs - ROM5Rel - 4/10/96

			kDebug20,	// 2.0 - Debug - 12/23/97
			0x3EA3,		// 2.0 - ROMA - ROM5Rel - 1/28/97
			0xEFB6,		// 2.0 - ROMB - English - ROM.2.0B.GM.US - 2/18/97
			0x11A9,		// 2.0 - ROMB - WorkPad - ROMf2wProFatRel - 7/7/97
			0x9573,		// 2.0 - International ROMA - French - ROMf2fRel - 4/29/97
			0x2422,		// 2.0 - International ROMA - German - ROMf2gRel - 3/7/97
			0x0BFC,		// 2.0 - International ROMB - French - ROMf1fProRel - 5/28/97
			0xFCBB,		// 2.0 - International ROMB - German - ROMf2gProRel - 5/29/97
			0x2E10		// 2.0 - International ROMB - Spanish - ROMf1sFatRel - 11/18/97
		};

		Word	checksumValue = Crc16CalcBigBlock ((/*non-const*/ void*) p, fileLength, 0);

		// Special hack for 2.0 Debug ROMs.  The version with this checksum was
		// bogus, so let's make sure we never load it.

		if (checksumValue == kDebug20)
		{
			// Throw a special error for this one.

			Errors::Throw (kError_UnsupportedROM);
		}

#if 0
		// Can't do this, yet, as the above set of checksums may not match
		// what's actually in shipping devices (I only know that they match
		// what we have on our servers around here).

		int	ii;
		for (int ii = 0; ii < (sizeof (kROMChecksums) / sizeof (kROMChecksums[0])); ++ii)
		{
			if (kROMChecksums[ii] == checksumValue)
			{
				return true;
			}
		}
#else
		return true;
#endif
	}
	else
	{
		// The checksum is the cumulative checksum of the ROM image before
		// the stored checksum value and the ROM image following the checksum
		// value.  First, calculate the first part.

		DWord	chunkSize = offsetof (CardHeaderType, checksumValue);
		Word	checksumValue = Crc16CalcBigBlock ((/*non-const*/ void*) p, chunkSize, 0);

		// Now calculate the second part.

		checksumValue = Crc16CalcBigBlock (
			((char*) p) + chunkSize + sizeof (cardHdr.checksumValue),
			cardHdr.checksumBytes - chunkSize - sizeof (cardHdr.checksumValue),
			checksumValue);

		if (checksumValue == cardHdr.checksumValue)
		{
			return true;
		}
	}

	return false;
}


// ---------------------------------------------------------------------------
//		 Card::Supports328
// ---------------------------------------------------------------------------

Bool Card::Supports328 (const void* p)
{
	CardHeaderType	cardHdr = *(CardHeaderType*) p;
//	ByteswapWords (&cardHdr, sizeof (CardHeaderType));
	Canonical (cardHdr);

	Bool	dbMode = false;

	if (cardHdr.hdrVersion < 3 || (cardHdr.flags & memCardHeaderFlag328))
	{
		dbMode = true;
	}

	return dbMode;
}


// ---------------------------------------------------------------------------
//		 Card::SupportsEZ
// ---------------------------------------------------------------------------

Bool Card::SupportsEZ (const void* p)
{
	CardHeaderType	cardHdr = *(CardHeaderType*) p;
//	ByteswapWords (&cardHdr, sizeof (CardHeaderType));
	Canonical (cardHdr);

	Bool	ezMode = false;

	if (cardHdr.hdrVersion >= 3 && (cardHdr.flags & memCardHeaderFlagEZ))
	{
		ezMode = true;
	}

	return ezMode;
}


#pragma mark -

// ===========================================================================
//		 CEnableFullAccess
// ===========================================================================

long	CEnableFullAccess::fgAccessCount = 0;


// ---------------------------------------------------------------------------
//		 CEnableFullAccess::CEnableFullAccess
// ---------------------------------------------------------------------------

CEnableFullAccess::CEnableFullAccess (void) :
	fOldMemAccessFlags (gMemAccessFlags)
#if HAS_PROFILING
	, fOldProfilingCounted (gProfilingCounted)
#endif
{
	gMemAccessFlags = kZeroMemAccessFlags;

#if HAS_PROFILING
	gProfilingCounted = false;
#endif

	++fgAccessCount;
}


// ---------------------------------------------------------------------------
//		 CEnableFullAccess::~CEnableFullAccess
// ---------------------------------------------------------------------------

CEnableFullAccess::~CEnableFullAccess (void)
{
	gMemAccessFlags = fOldMemAccessFlags;

#if HAS_PROFILING
	gProfilingCounted = fOldProfilingCounted;
#endif

	--fgAccessCount;
}


// ---------------------------------------------------------------------------
//		 CEnableFullAccess::AccessOK
// ---------------------------------------------------------------------------

Bool CEnableFullAccess::AccessOK (void)
{
	return fgAccessCount > 0;
}
