/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "CPU_MT.h"

#include "CPU_REG.h"			// Emulator::Execute, etc.
#include "Platform.h"			// gEnableSounds
#include "Strings.r.h"			// kStr_ values


// ======================================================================
//	Globals and constants
// ======================================================================

HANDLE	gCPUThread;
HANDLE	gCPUBeginEvent;			// Signal this to start the CPU loop
HANDLE	gCPUStoppedEvent;		// Gets signaled when CPU loop exits.
HANDLE	gCPUDelayEvent;			// Gets signaled to wake up Platform::Delay.
Bool	gQuitCPUThread;			// True when time to quit the thread.


/***********************************************************************
 *
 * FUNCTION:	CPU
 *
 * DESCRIPTION:	Thread procedure for the CPU thread.  The CPU thread
 *				just executes emulated 68K instructions.  The only time
 *				it ever stops is if an exception occurs (or when it's
 *				requested).  If either of these occurs, the CPU loop
 *				exits back to here.  CPU() signals the event by
 *				signalling the event -- the gCPUStoppedEvent event, that
 *				is.  Anyone waiting for the CPU thread to pause can look
 *				for or wait on this condition.  CPU() then blocks itself
 *				on gCPUBeginEvent, waiting for some external source to
 *				start it up again.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

DWORD WINAPI CPU (void*)
{
	while (1)
	{
		::WaitForSingleObject (gCPUBeginEvent, INFINITE);	// Wait for someone to start us.

		if (gQuitCPUThread)
			break;

		Emulator::Execute ();							// Run the CPU loop.

		::SetEvent (gCPUStoppedEvent);					// Signal that we've stopped.
	}

	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	CPU_init
 *
 * DESCRIPTION:	Initializes the thread safe interface to the CPU
 *				emulator subsystem.  First initializes the subsytem
 *				itself, and then creates a thread and some synchro-
 *				nization objects to manage it.
 *
 * PARAMETERS:	deviceType - type of hardware to emulate.
 *
 *				ramFile - reference to the file containing the RAM
 *					image (and other details) to use.  May not refer
 *					to an actual file, in which case a fresh RAM
 *					image is created.
 *
 *				romFile - reference to a ROM image file to use.
 *
 *				ramSize - size of RAM (in K) to use.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

ErrCode CPU_init (	int deviceType,
					FileReference& ramFile,
					FileReference& romFile,
					uae_s32 ramSize)
{
	// Try initializing the emulator with all this information.

	try
	{
		Emulator::Initialize (deviceType, ramFile, romFile, ramSize);
	}
	catch (ErrCode errCode)
	{
		return errCode;
	}

	// Create the event objects.
	//
	// gCPUBeginEvent is used to block the CPU loop until we're ready to
	// run.  It gets unblocked in CPU_run, which gets called (among other
	// places) in DoDisplayThread just before we enter the event handling
	// loop.
	//
	// gCPUStoppedEvent gets signaled whenever the CPU loop exits.  It's
	// a manual-reset event, which let's us poll at any time to see if the
	// CPU loop has stopped or not.

	gCPUBeginEvent		= ::CreateEvent (	NULL,		// attributes
											FALSE,		// auto-reset
											FALSE,		// non-signalled
											NULL);		// name

	gCPUStoppedEvent	= ::CreateEvent (	NULL,		// attributes
											TRUE,		// manual-reset
											TRUE,		// signalled
											NULL);		// name

	gCPUDelayEvent		= ::CreateEvent (	NULL,		// attributes
											TRUE,		// manual-reset
											FALSE,		// non-signalled
											NULL);		// name

	gQuitCPUThread = false;

	// Create the CPU thread.  Because gCPUBeginEvent is created as
	// non-signalled, the CPU thread will block immediately.  The
	// UI thread will unblock it when it's ready.

	DWORD	tid;
	gCPUThread = ::CreateThread (	NULL,	// attributes
									0,		// stack size
									CPU,	// func
									NULL,	// param
									0,		// flags
									&tid);	// id

	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:	CPU_dispose
 *
 * DESCRIPTION:	Reverse the effects of CPU_initialize.  Stops the CPU
 *				thread and destroys the OS handles managing it.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void CPU_dispose (void)
{
	// Should already be stopped, but let's just make sure:

	::CPU_stop ();

	// Tell the thread to quit.

	gQuitCPUThread = true;

	// OK, start it up again so that it can die.

	::CPU_run ();

	// Wait for it to die.

	::WaitForSingleObject (gCPUThread, INFINITE);

	// Close it out.

	::CloseHandle (gCPUThread);			gCPUThread = NULL;
	::CloseHandle (gCPUDelayEvent);		gCPUDelayEvent = NULL;
	::CloseHandle (gCPUStoppedEvent);	gCPUStoppedEvent = NULL;
	::CloseHandle (gCPUBeginEvent);		gCPUBeginEvent = NULL;

	// Free up the memory.

	Emulator::Dispose ();
}


/***********************************************************************
 *
 * FUNCTION:	CPU_reset
 *
 * DESCRIPTION:	Thread-safe interface to Emulator::Reset.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void CPU_reset (void)
{
	::CPU_stop ();

	Emulator::Reset ();

	::CPU_run ();
}


/***********************************************************************
 *
 * FUNCTION:	CPU_run
 *
 * DESCRIPTION:	Start up the CPU thread.  If the thread is already
 *				running, do nothing.  Otherwise, unsignal the event
 *				that flags when the CPU stops, and unblock the thread
 *				by signalling gCPUBeginEvent.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void CPU_run (void)
{
	// If the CPU is stopped, restart it by signalling the gCPUBeginEvent object.

	if (::WaitForSingleObject (gCPUStoppedEvent, 0) == WAIT_OBJECT_0)
	{
		::ResetEvent (gCPUStoppedEvent);	// Unsignal this guy
		::SetEvent (gCPUBeginEvent);		// Unblock CPU()
	}
}


/***********************************************************************
 *
 * FUNCTION:	CPU_stop
 *
 * DESCRIPTION:	Stop the CPU thread, returning whether or not the thread
 *				was running when this function was called (TRUE == it
 *				was running, FALSE == it was already stopped).  If the
 *				thread is running, this function asks it to stop, and
 *				then waits FOREVER for it to stop.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	Previous state of thread.
 *
 ***********************************************************************/

Bool CPU_stop (void)
{
	// See if the CPU is running. If not, return false (saying that it
	// wasn't running).

	if (::WaitForSingleObject (gCPUStoppedEvent, 0) == WAIT_OBJECT_0)
		return false;

	// Ask the CPU loop to exit.

	Emulator::SetBreakReason (kBreak_StopRequest);

	// If it was dozing, give it a nudge.

	::SetEvent (gCPUDelayEvent);

	// Wait for the CPU thread to stop.

	::WaitForSingleObject (gCPUStoppedEvent, INFINITE);

	::ResetEvent (gCPUDelayEvent);

	return true;	// Say that we stopped it.
}


/***********************************************************************
 *
 * FUNCTION:	CPU_wait
 *
 * DESCRIPTION:	Merely wait for the CPU thread to stop.  Wait for the
 *				specified amount of time.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	If the thread stops in time, return TRUE, else FALSE.
 *
 ***********************************************************************/

Bool CPU_wait (DWORD timeout)
{
	return ::WaitForSingleObject (gCPUStoppedEvent, timeout) == WAIT_OBJECT_0;
}


/***********************************************************************
 *
 * FUNCTION:	CPU_executeuntilatrap
 *
 * DESCRIPTION:	Continue emulation until an "A-trap" is executed.  By
 *				"A-trap", I mean "system function invoked via a TRAP $F,
 *				0xA??? sequence".  Times out after 1 second.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	TRUE if emulation was halted. 
 *
 ***********************************************************************/

Bool CPU_executeuntilatrap (void)
{
	// Tell the CPU emulator to exit on the next "ATrap"

	Bool	old = Emulator::SetBreakOnException (kException_Trap15, true);

	// Wait for it to do so (one second).

	Bool	ok = ::CPU_wait (1000);

	// Restore the old setting.

	(void) Emulator::SetBreakOnException (kException_Trap15, old);

	return ok;
}


/***********************************************************************
 *
 * FUNCTION:	CPU_pen
 *
 * DESCRIPTION:	Thread-safe interface to Hardware::PenEvent.
 *
 * PARAMETERS:	down - true if the pen/mouse is down, false if up.
 *
 *				x, y - LCD coordinates of the event.  If down==FALSE,
 *					x and y are ignored, but are typically (-1,-1).
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void CPU_pen (Bool down, int x, int y)
{
	// Don't do anything if the CPU is already stopped.
	// It might, for example, be in the debugger.

	if (::CPU_stop ())
	{
		Hardware::PenEvent (down, x, y);
		::CPU_run ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	CPU_button
 *
 * DESCRIPTION:	Thread-safe interface to Hardware::ButtonEvent.
 *
 * PARAMETERS:	down - true if the button is pressed, false if up.
 *
 *				button - number indicating what hard button is pressed.
 *					The value is one of kButton_Power, etc.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void CPU_button (Bool down, int button)
{
	// Don't do anything if the CPU is already stopped.
	// It might, for example, be in the debugger.

	if (::CPU_stop ())
	{
		Hardware::ButtonEvent (down, button);
		::CPU_run ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	CPU_hotsync
 *
 * DESCRIPTION:	Thread-safe interface to Hardware::HotSyncEvent.
 *
 * PARAMETERS:	down - true if HotSync button pressed, false if up.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void CPU_hotsync (Bool down)
{
	// Don't do anything if the CPU is already stopped.
	// It might, for example, be in the debugger.

	if (::CPU_stop ())
	{
		Hardware::HotSyncEvent (down);
		::CPU_run ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	CPU_putkey
 *
 * DESCRIPTION:	Thread-safe interface to Hardware::KeyboardEvent.
 *
 * PARAMETERS:	key - ASCII value of key to enter into the Palm OS
 *					event system
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void CPU_putkey (BYTE key)
{
	// Don't do anything if the CPU is already stopped.
	// It might, for example, be in the debugger.

	if (::CPU_stop ())
	{
		Hardware::KeyboardEvent (key);
		::CPU_run ();
	}
}


// ---------------------------------------------------------------------------
//		 CPU_setsounds
// ---------------------------------------------------------------------------

void CPU_setsounds(Bool enablesounds)
{
	gEnableSounds = enablesounds;
}


// ---------------------------------------------------------------------------
//		 CPUStopper::CPUStopper
// ---------------------------------------------------------------------------

CPUStopper::CPUStopper (HWND parent, const char* msg)	// !!! Localize calls to this c'tor.
{
	ok = ::CPU_executeuntilatrap();	// Stop the CPU on an ATrap instruction.

	if (!ok)
	{
		/*	!!!
			CPU_MT.cpp: CPUStopper::CPUStopper: The warning message in this
			function is the one that's most often quoted back to me in messages from
			people asking for help with Copilot. Generally the problem is they've
			got half a ROM file and the boot sequence hangs, at which time they
			attempt to load an application. This message is sort of inappropriate at
			this point, because it leads them to believe they should turn the
			emulator "on" first. Not knowing how to do that, they click on the
			nearest mailto: link and I hear about it. Perhaps Poser users will run
			into this problem less often because of the better startup code, but
			nevertheless I think this message could use a bit more explanation.
							
							--Greg
		*/

		char	buffer[200];
		string	format (Platform::GetString (kStr_EmulatorOff));
		sprintf (buffer, format.c_str (), msg);
		string	app (Platform::GetString (kStr_AppName));
		::MessageBox(parent, buffer, app.c_str (), MB_OK);
	}
}


// ---------------------------------------------------------------------------
//		 CPUStopper::~CPUStopper
// ---------------------------------------------------------------------------

CPUStopper::~CPUStopper ()
{
	if (ok)
	{
		::CPU_run ();
	}
}


