/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Document.h"

#include <windowsx.h>			// ComboBox_SetCurSel

#include "CPU_REG.h"			// Emulator::kPilot1000, etc.
#include "Emulator.h"			// gInstance
#include "LCD.h"				// CenterWindow
#include "CPU_MT.h"				// CPU_init
#include "RAM_ROM.h"			// SessionFile
#include "Settings.h"			// gSettings
#include "Platform.h"			// Platform::ReportError

#include "resource.h"


// ======================================================================
//	Globals and constants
// ======================================================================

AppPreferences::Configuration	gCurrentConfiguration;
FileReference					gCurrentRAMRef;

AppPreferences::Configuration	gCfg;


// ======================================================================
//	Private functions
// ======================================================================

static BOOL CALLBACK	PrvQueryNewDocumentDlgProc (HWND hwnd,
													UINT msg,
													WPARAM wparam,
													LPARAM lparam);
static Bool				PrvBrowseForROM (HWND hwnd, FileReference& romFile);
static FileReference	PrvFindPairedROMFile (FileHandle& docFile);


/***********************************************************************
 *
 * FUNCTION:	QueryNewDocument
 *
 * DESCRIPTION:	Ask the user for the parameters for a new session.
 *				This information includes: devide type, country
 *				specific graphics, ram size, and rom file.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

Bool QueryNewDocument (HWND hwnd, AppPreferences::Configuration& cfg)
{
	gCfg = cfg;

	int	result = ::DialogBox (gInstance, MAKEINTRESOURCE (IDD_NEWFILE), hwnd,
								&::PrvQueryNewDocumentDlgProc); 

	if (result == IDOK)
	{
		cfg = gCfg;
	}

	return result == IDOK;
}


/***********************************************************************
 *
 * FUNCTION:	PrvQueryNewDocumentDlgProc
 *
 * DESCRIPTION:	DlgProc for QueryNewDocument
 *
 * PARAMETERS:	Standard DlgProc parameters.
 *
 *				Custom information is communications in the global
 *				gCfg variable.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

BOOL CALLBACK PrvQueryNewDocumentDlgProc (	HWND hwnd,
											UINT msg,
											WPARAM wparam,
											LPARAM lparam)
{
	// Must match the IDC_DEVICETYPE menu.

	static uae_u32 typesArray[] =
	{
		Emulator::kPilot1000,			// Emulator::kPilot5000,
		Emulator::kPalmPilotPersonal,	// Emulator::kPalmPilotProfessional,
		Emulator::kPalmIII,
		Emulator::kPalmIIIx,
		Emulator::kPalmV,
		Emulator::kPalmVII
	};


	// Must match the IDC_RAMSIZE menu.

	static uae_u32 memSizesArray[] =
	{
		128,
		256,
		512,
		1024,
		2048,
		3072,
		4096,
		8192
	};


	// Must match the Country menu

	static uae_u32 countryArray[] =
	{
		AppPreferences::kSilkscreenEnglish,
		AppPreferences::kSilkscreenJapanese
	};


	switch (msg)
	{
		case WM_INITDIALOG:
		{
			::CenterWindow (hwnd);

			HWND	dlgItem;

			const int kDeviceBase = 32;
			const int kRAMSizeBase = 16;

			dlgItem = ::GetDlgItem (hwnd, IDC_DEVICE);
			ComboBox_AddString (dlgItem, Platform::GetString (kDeviceBase + 0).c_str ());
			ComboBox_AddString (dlgItem, Platform::GetString (kDeviceBase + 1).c_str ());
			ComboBox_AddString (dlgItem, Platform::GetString (kDeviceBase + 2).c_str ());
			ComboBox_AddString (dlgItem, Platform::GetString (kDeviceBase + 3).c_str ());
			ComboBox_AddString (dlgItem, Platform::GetString (kDeviceBase + 4).c_str ());
			ComboBox_AddString (dlgItem, Platform::GetString (kDeviceBase + 5).c_str ());


			// Establish a default setting.
			ComboBox_SetCurSel (dlgItem, 2);	// Palm III

			for (uae_u32 ii = 0; ii < countof (typesArray); ++ii)
			{
				if (typesArray[ii] == gCfg.fDeviceType)
				{
					ComboBox_SetCurSel (dlgItem, ii);
				}
			}

			dlgItem = ::GetDlgItem (hwnd, IDC_COUNTRY);
			ComboBox_AddString (dlgItem, "English");
			ComboBox_AddString (dlgItem, "Japanese");

			// Establish a default setting.
			ComboBox_SetCurSel (dlgItem, 0);	// English

			for (ii = 0; ii < countof (countryArray); ++ii)
			{
				if (countryArray[ii] == gCfg.fSilkscreenType)
				{
					ComboBox_SetCurSel (dlgItem, ii);
				}
			}

			dlgItem = ::GetDlgItem (hwnd, IDC_RAMSIZE);
			ComboBox_AddString (dlgItem, (Platform::GetString (kRAMSizeBase + 0).c_str ()));
			ComboBox_AddString (dlgItem, (Platform::GetString (kRAMSizeBase + 1).c_str ()));
			ComboBox_AddString (dlgItem, (Platform::GetString (kRAMSizeBase + 2).c_str ()));
			ComboBox_AddString (dlgItem, (Platform::GetString (kRAMSizeBase + 3).c_str ()));
			ComboBox_AddString (dlgItem, (Platform::GetString (kRAMSizeBase + 4).c_str ()));
			ComboBox_AddString (dlgItem, (Platform::GetString (kRAMSizeBase + 5).c_str ()));
			ComboBox_AddString (dlgItem, (Platform::GetString (kRAMSizeBase + 6).c_str ()));
			ComboBox_AddString (dlgItem, (Platform::GetString (kRAMSizeBase + 7).c_str ()));

			// Establish a default setting.
			ComboBox_SetCurSel (dlgItem, 3);	// 1024K

			for (ii = 0; ii < countof (memSizesArray); ++ii)
			{
				if (memSizesArray[ii] == gCfg.fRAMSize)
				{
					ComboBox_SetCurSel (dlgItem, ii);
				}
			}

			dlgItem = ::GetDlgItem (hwnd, IDC_ROMFILE);
			::SetWindowText (dlgItem, gCfg.fROMFile.GetFilePath ());

			break;
		}

		case WM_COMMAND:
		{
			switch (LOWORD (wparam))
			{
				case IDOK:
				{
					HWND	dlgItem;
					LONG	index;

					// Get the chosen device type.

					dlgItem = ::GetDlgItem (hwnd, IDC_DEVICE);
					index = ComboBox_GetCurSel (dlgItem);
					gCfg.fDeviceType = typesArray [index];

					// Get the chosen country graphics.

					dlgItem = ::GetDlgItem (hwnd, IDC_COUNTRY);
					index = ComboBox_GetCurSel (dlgItem);
					gCfg.fSilkscreenType = countryArray [index];

					// Get the chosen ram size.

					dlgItem = ::GetDlgItem (hwnd, IDC_RAMSIZE);
					index = ComboBox_GetCurSel (dlgItem);
					gCfg.fRAMSize = memSizesArray [index];

					// Get the chosen ROM filename.

					char	fileName[MAX_PATH];
					dlgItem = ::GetDlgItem (hwnd, IDC_ROMFILE);
					::GetWindowText (dlgItem, fileName, MAX_PATH);
					gCfg.fROMFile = fileName;
				}
				// Fall thru...

				case IDCANCEL:
					::EndDialog (hwnd, LOWORD (wparam));
					break;

				case IDC_BROWSE_ROMFILE:
				{
					FileReference	romFile;
					if (::PrvBrowseForROM (hwnd, romFile))
					{
						HWND	dlgItem = ::GetDlgItem (hwnd, IDC_ROMFILE);
						::SetWindowText (dlgItem, romFile.GetFilePath ());
					}
					break;
				}
			}

			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}


/***********************************************************************
 *
 * FUNCTION:	CreateNewDocument
 *
 * DESCRIPTION:	Create a new session based on the given configuration.
 *
 * PARAMETERS:	cfg - configuration containing the specifics of
 *					the new session.
 *
 * RETURNED:	Error code for any errors that may occur.
 *
 ***********************************************************************/

ErrCode CreateNewDocument (const AppPreferences::Configuration& cfg)
{
	FileReference	dummyRef;
	FileReference	romFile (cfg.fROMFile);	// !!! Need a non-const object?

	// Try initializing the emulator with all this information.

	ErrCode	error = CPU_init (cfg.fDeviceType, dummyRef, romFile, cfg.fRAMSize);

	if (error == 0)
	{
		gCurrentConfiguration = cfg;
		gPrefs->SetLastConfiguration (cfg);

		// Zap this one out, so that we give preference to a New operation
		// at startup and not an Open operation.

		gCurrentRAMRef = dummyRef;
		gPrefs->SetLastRAMFile (dummyRef);
	}

	return error;
}


/***********************************************************************
 *
 * FUNCTION:	QueryOpenDocument
 *
 * DESCRIPTION:	Ask the user for an existing document to open.
 *
 * PARAMETERS:	hwnd - parent window.
 *
 *				fileRef - object to receive the user's choice.
 *
 * RETURNED:	True if the user made a choice, false otherwise.
 *
 ***********************************************************************/

Bool QueryOpenDocument (HWND hwnd, FileReference& fileRef)
{
	// Ask the user to choose a .PSF file.

	static const char	filters[] =
	{
		"Emulator Session Files (*.psf)\0"		"*.psf\0"
		"All Files (*.*)\0"						"*.*\0"
		""
	};

	char	fileName[MAX_PATH];

	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	reply.lStructSize	= sizeof (OPENFILENAME);
	reply.hInstance		= gInstance;
	reply.lpstrFilter	= filters;
	reply.nFilterIndex	= 1;
	reply.lpstrFile		= fileName;
	reply.nMaxFile		= MAX_PATH;
	reply.Flags			= OFN_ENABLEHOOK /*| OFN_ENABLESIZING*/ | OFN_EXPLORER;
	reply.lpfnHook		= &::CenterDlgProc;

	fileName[0] = 0;

	BOOL	result = ::GetOpenFileName (&reply);

	if (result)
	{
		fileRef = FileReference (fileName);
	}

	return result != 0;
}


/***********************************************************************
 *
 * FUNCTION:	OpenOldDocument
 *
 * DESCRIPTION:	Create a new session based on an existing .psf file.
 *
 * PARAMETERS:	ramFileRef - reference to the .psf file.
 *
 * RETURNED:	Error code for any errors that may occur.
 *
 ***********************************************************************/

ErrCode OpenOldDocument (FileReference& ramFileRef)
{
	ErrCode	result = 0;

	AppPreferences::Configuration	cfg;

	// Create new scope so that the ram file is closed before
	// jumping off into Emulator::Initialize.

	try
	{
		// Open the RAM file itself.

		FileHandle	ramFile (ramFileRef, kOpenExisting | kOpenRead);

		// Read from it the associated ROM file.

		cfg.fROMFile = ::PrvFindPairedROMFile (ramFile);

		// Read from the RAM file the device type.

		SessionFile::ReadInt (ramFile, SessionFile::kDeviceType, cfg.fDeviceType);

		// Read from the RAM file the silkscreen type.

		SessionFile::ReadInt (ramFile, SessionFile::kSilkscreenType, cfg.fSilkscreenType);
	}
	catch (ErrCode errCode)
	{
		result = errCode;

		gPrefs->SetLastRAMFile (FileReference ());
	}

	// Try initializing the emulator with all this information.

	if (result == 0)
	{
		result = CPU_init (cfg.fDeviceType, ramFileRef, cfg.fROMFile, 0);
	}

	// If everything succeeded, save the configuration for next time.

	if (result == 0)
	{
		gCurrentRAMRef			= ramFileRef;
		gCurrentConfiguration	= cfg;
		gPrefs->SetLastConfiguration (cfg);

		gPrefs->SetLastRAMFile (ramFileRef);
		gPrefs->UpdateRAMMRU (ramFileRef);
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	SaveRAMImage
 *
 * DESCRIPTION:	Save the current session to a .psf file.
 *
 * PARAMETERS:	hwnd - parent window (used when asking the user for
 *					a destination file name).
 *
 *				ramFileRef - reference to a currently existing file.
 *					Optional; may be unspecified if the current
 *					session has never been saved before.
 *
 *				askUser - True if this function should always ask the
 *					user for a destination filename.  Otherwise, the
 *					user is asked only if ramFileRef is unspecified.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void SaveRAMImage (HWND hwnd, FileReference& ramFileRef, Bool askUser)
{
	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	char			fileName[MAX_PATH] = {0};

	// See if there's a given name we should use.

	if (ramFileRef.IsSpecified ())
	{
		strcpy (fileName, ramFileRef.GetFilePath ());
	}

	if (askUser || *fileName == 0)
	{
		string	saveRAMPrompt (Platform::GetString (IDS_SAVE_RAM_PROMPT));

		reply.lStructSize	= sizeof (OPENFILENAME);
		reply.hInstance		= gInstance;
		reply.lpstrFilter	= "Emulator Session Files (*.psf)\0*.psf\0";
		reply.nFilterIndex	= 1;
		reply.lpstrFile		= fileName;
		reply.nMaxFile		= MAX_PATH;
		reply.lpstrTitle	= saveRAMPrompt.c_str ();
		reply.Flags			= OFN_ENABLEHOOK | OFN_OVERWRITEPROMPT | OFN_EXPLORER;
		reply.lpstrDefExt	= "psf";
		reply.lpfnHook		= CenterDlgProc;

		if (!::GetSaveFileName (&reply))
		{
			DWORD error = ::CommDlgExtendedError ();
			return;
		}
	}

	ramFileRef = FileReference (fileName);
	FileHandle	ramFile (ramFileRef, kCreateAlways | kOpenReadWrite);

	// Memory::Save will report an error and throw an exception if it
	// couldn't save the file.  However, it's too late to do anything
	// about it now, so just swallow the exception.  !!! TBD, revisit
	// this later so that developer can free up some disk space before
	// quitting.

	try
	{
		// Save the ram image, ROM name, device type, and graphics type.
		Memory::Save (ramFile, gCurrentConfiguration);
	}
	catch (...)
	{
	}

	// Save the ROM path.

	SessionFile::WriteString (ramFile, SessionFile::kROMPathTag, gCurrentConfiguration.fROMFile.GetFilePath ());

	// Remember this file for next time.

	gPrefs->SetLastRAMFile (ramFileRef);
	gPrefs->UpdateRAMMRU (ramFileRef);
}


/***********************************************************************
 *
 * FUNCTION:	PrvBrowseForROM
 *
 * DESCRIPTION:	Ask the user for a ROM image file.
 *
 * PARAMETERS:	hwnd - parent window
 *
 *				romFile - receives the user's choice.
 *
 * RETURNED:	True if the user chose a ROM file, false otherwise.
 *
 ***********************************************************************/

Bool PrvBrowseForROM (HWND hwnd, FileReference& romFile)
{
	// Look for files ending in ".rom" or " rom", or starting with "rom.".

	static const char	filters[] =
	{
		"Palm OS ROM Files (*.rom)\0"		"*.rom;* rom;rom.*\0"
		"All Files (*.*)\0"					"*.*\0"
		""
	};

	char	fileName[MAX_PATH];

	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	reply.lStructSize       = sizeof (OPENFILENAME);
	reply.hInstance			= gInstance;
	reply.lpstrFilter       = filters;
	reply.nFilterIndex      = 1;
	reply.lpstrFile         = fileName;
	reply.nMaxFile          = MAX_PATH;
	reply.Flags				= OFN_ENABLEHOOK /*| OFN_ENABLESIZING*/ | OFN_EXPLORER;
	reply.lpfnHook			= CenterDlgProc;

	fileName[0] = 0;

	BOOL	result = ::GetOpenFileName (&reply);

	if (result)
	{
		romFile = FileReference (fileName);
	}

	return result != 0;
}


/***********************************************************************
 *
 * FUNCTION:	PrvFindPairedROMFile
 *
 * DESCRIPTION:	Gets the ROM file associated with the given RAM file.
 *				The RAM file is first scanned for a full path to the
 *				ROM file.  If that can't be found, the RAM file is
 *				scanned for a simple name (which will be assumed to be
 *				the name of a ROM file in the current directory).
 *
 * PARAMETERS:	docFile - handle to the open .psf file.
 *
 * RETURNED:	ROM file reference stored and found in the .psf file.
 *
 ***********************************************************************/

FileReference PrvFindPairedROMFile (FileHandle& docFile)
{
	// First look for a ROM file path.

	long	len = SessionFile::FindChunk (docFile, SessionFile::kROMPathTag);
	
	if (len > 0)
	{
		char	path[MAX_PATH];
		SessionFile::ReadChunk (docFile, len, path);
		path [len] = 0;
		return FileReference (path);
	}

	// If a path can't be found, look for a simple ROM name.

	len = SessionFile::FindChunk (docFile, SessionFile::kROMNameTag);

	if (len > 0)
	{
		char	path[_MAX_PATH], drive[_MAX_DRIVE], dir[_MAX_DIR], name[MAX_PATH];

		SessionFile::ReadChunk (docFile, len, name);

		::GetModuleFileName (gInstance, path, _MAX_PATH);
		_splitpath (path, drive, dir, NULL, NULL);

		_makepath (path, drive, dir, name, NULL);

		return FileReference (path);
	}

	return FileReference ();
}
