/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

/*
	Notes on the Xmodem/Ymodem implementation used in this file:

	Basic Xmodem is dead simple:

		1. Sender waits for NAK from receiver.
		2. Sender sends SOH, block#, complement of block#
		3. Sender sends 128 bytes of data.
		3. Sender sends one-byte sum of those 128 bytes as checksum.
		4. Receiver sends ACK or NAK.
		5. If NAK, go back to step 2.
		6. If ACK, increment block# and go back to step 2.
		7. At end of file, sender sends EOT.
		8. Receiver acknowledges EOT with ACK.

	There is an Xmodem variant called Ymodem that sends an additional block at
	the start of the protocol - this block contains the file name and size,
	among other things.
*/


#include "EmulatorCommon.h"
#include "DownloadROM.h"

#include "Crc.h"				// Crc16CalcBlock
#include "Emulator.h"			// gInstance
#include "ErrorHandling.h"		// kError_CantDownloadROM_BadBaudRate, etc.
#include "LCD.h"				// CenterWindow
#include "Platform.h"			// Platform::GetString, CloseSerialPort
#include "Settings.h"			// gSettings

#include "resource.h"			// IDC_PORT, IDS_DOWNLOAD_DONE, etc.

#include <CommCtrl.h>			// PropSheet_SetWizButtons (actually, in PrSht.h), etc.
#include <windowsx.h>			// ComboBox_AddString


// ======================================================================
//	Globals and constants
// ======================================================================

#define kXmodemBlockSize	1024	// 1k-Xmodem variant
#define kXmodemSoh			'\002'	// start of block header
#define kXmodemEof			'\004'	// end of file signal
#define kXmodemAck			'\006'	// acknowledge
#define kXmodemNak			'\025'	// negative acknowledge (resend packet)
#define kXmodemCan			'\030'	// cancel
#define kXmodemNakCrc		'C'		// used instead of NAK for initial block

struct XmodemPacket
{
	BYTE	startOfHeader;
	BYTE	blockNumber;
	BYTE	blockComplement;
	BYTE	data[kXmodemBlockSize];
	BYTE	crcHi;
	BYTE	crcLo;
};

struct DownloadData
{
	long		OK;					// Which button exited the dialog?
	char		portName[MAX_PATH];	// port to use
	long		speed;				// Baud rate (0 = 57,600, etc.)
	const char*	defaultName;		// Save ROM to this file (NULL = ask for name)

	long		romRead;			// Amount of ROM transferred
	long		romSize;			// Total ROM size
	BYTE*		romImage;			// ROM image buffer
	DWORD		result;				// Result of transfer operation

	BYTE*		readBuffer;			// Data from ReadFile is placed here for processing
	long		readBufferOffset;	// Amount of data in "readBuffer"

	BYTE		xmodemBlock;		// Xmodem block number
	DWORD		xmodemTimeoutBase;  // Last time we sent a response to the sender

	Bool		reopenSerialPort;	// Need to re-open the serial port for
									// "serial port emulation" for the ROM.

	HANDLE		downloadThread;		// Handle to thread doing the downloading
	HANDLE		quitEvent;			// Event to tell thread to quit
};


// ======================================================================
//	Private functions
// ======================================================================

// Handy macro to send a single byte (that's all we'll ever need to send).

#define SendByte(b) \
	do { \
		BYTE _b = b; \
		DWORD _n; \
		WriteFile(port, &_b, 1, &_n, &ov); \
		GetOverlappedResult(port, &ov, &_n, TRUE); \
	} while (0)


/***********************************************************************
 *
 * FUNCTION:	PrvSaveROMImage
 *
 * DESCRIPTION:	Asks the user for a file to which to save the
 *				downloaded ROM, and saves the ROM image to that file.
 *
 * PARAMETERS:	hdlg - parent window
 *
 *				romImage - pointer to the ROM image
 *
 *				romSize - size of the ROM image
 *
 *				name - default file name. May be NULL.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

static void PrvSaveROMImage (HWND hDlg, const void* romImage, long romSize, const char* name)
{
	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	char			fileName[MAX_PATH] = {0};

	// See if there's a given name we should use.

	if (name)
	{
		strcpy (fileName, name);
	}

	// Ask the user for one.

	string	saveROMPrompt (Platform::GetString (IDS_SAVE_ROM_PROMPT));

	reply.lStructSize	= sizeof (OPENFILENAME);
	reply.hInstance		= gInstance;
	reply.lpstrFilter	= "Palm OS ROM Files (*.rom)\0*.rom\0";
	reply.nFilterIndex	= 1;
	reply.lpstrFile		= fileName;
	reply.nMaxFile		= MAX_PATH;
	reply.lpstrTitle	= saveROMPrompt.c_str ();
	reply.Flags			= OFN_ENABLEHOOK | OFN_OVERWRITEPROMPT | OFN_EXPLORER;
	reply.lpstrDefExt	= "rom";
	reply.lpfnHook		= &::CenterDlgProc;

	if (!::GetSaveFileName (&reply))
	{
		DWORD error = ::GetLastError ();
		return;
	}

	// Open the file and write the ROM image to it.

	FILE*	f = fopen (fileName, "wb");
	if (f)
	{
		fwrite (romImage, 1, romSize, f);
		fclose (f);

		// Let's also save this as the next default name.

		AppPreferences::Configuration	cfg;
		cfg = gPrefs->GetLastConfiguration ();
		cfg.fROMFile = FileReference (fileName);
		gPrefs->SetLastConfiguration (cfg);
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvSetDownloadError
 *
 * DESCRIPTION:	Bottleneck location for remembering that an error
 *				occurred and what error it was.  Handy place for
 *				setting a breakpoint.
 *
 * PARAMETERS:	data - "globals" used while downloading
 *
 *				error - error code to remember
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

static void PrvSetDownloadError (DownloadData* data, DWORD error)
{
	data->result = error;
}


/***********************************************************************
 *
 * FUNCTION:	PrvValidXmodemBlock
 *
 * DESCRIPTION:	Determine if the given buffer contains a valid XMODEM
 *				packet.
 *
 * PARAMETERS:	block - pointer to the buffer containing the packet.
 *
 * RETURNED:	True if the buffer contains a valid packet.
 *
 ***********************************************************************/

static Bool PrvValidXmodemBlock (BYTE* block)
{
	/*
	 * Xmodem-1k block layout is as follows:
	 *
	 * +--------------+
	 * | SOH = '\002' |
	 * +--------------+
	 * | Block number |
	 * +--------------+
	 * | Block compl. | Complement of block number
	 * +--------------+
	 * | 1024 bytes   |
	 * | of data      |
	 * .      .       .
	 * .      .       .
	 * +--------------+
	 * | CRC hi byte  | CRC-16 of preceding 1024 bytes of data,
	 * +--------------+ plus two zero bytes
	 * | CRC lo byte  |
	 * +--------------+
	 */

	if (block[0] != kXmodemSoh)
	{
		return false;
	}

	if ((block[1] ^ block[2]) != 0xFF)
	{
		return false;
	}

	WORD crc = ::Crc16CalcBlock (&block[3], kXmodemBlockSize, 0);

	Byte	zeros[2] = {0, 0};
	crc = ::Crc16CalcBlock (zeros, 2, crc);

	WORD	storedCRC = (block[3 + kXmodemBlockSize] << 8) | block[3 + kXmodemBlockSize + 1];

	if (storedCRC != crc)
	{
		return false;
	}

	return true;
}


/***********************************************************************
 *
 * FUNCTION:	HandleDownloadedData
 *
 * DESCRIPTION:	Some data has been received...scope it out.  If we have
 *				a full packet of data containing part of the ROM image,
 *				append that data to the full ROM image.
 *
 * PARAMETERS:	data - "globals" used while downloading
 *
 * RETURNED:	If the protocol requires that some response is
 *				required, this function returns the character that
 *				is to be sent back.  Otherwise, it returns zero.
 *
 ***********************************************************************/

BYTE HandleDownloadedData (DownloadData* data)
{
	// First check to see whether we have enough data for a full block, and
	// that the block is valid, and it's the block number we're expecting

	if (data->readBufferOffset >= 1+2+kXmodemBlockSize+2
		&& PrvValidXmodemBlock (data->readBuffer)
		&& data->readBuffer[1] == data->xmodemBlock)
	{
		if (data->romImage == NULL)
		{
			// We must be at the first block. Skip past the file name,
			// and get the file (ROM) size, and allocate space for the ROM image.

			char* p = (char*) &data->readBuffer[3];
			p += strlen (p) + 1;
			data->romSize = atoi (p);
			data->romImage = (BYTE*) calloc (data->romSize, 1);
		}
		else
		{
			// Got a good block of data. Copy it into the ROM image.

			memcpy (data->romImage + data->romRead, &data->readBuffer[3], kXmodemBlockSize);
			data->romRead += kXmodemBlockSize;
		}

		// Prepare the read buffer for the next block of data.
		// Acknowledge this packet as good.

		data->readBufferOffset = 0;
		data->xmodemBlock++;
		return kXmodemAck;
	}
	else if (data->readBuffer[0] == kXmodemEof)
	{
		// EOF signal. We may or may not actually get here because the main loop
		// may drop out as soon as it receives enough data. Acknowledge and move on.

		return kXmodemAck;
	}

	if (::GetTickCount() - data->xmodemTimeoutBase > 5000)
	{
		// We haven't received a good packet in five seconds. Probably a dropped
		// character, or we're just starting the protocol. If just starting, nak
		// with kXmodemNakCrc 'C', else just use regular nak '\025'.

		data->readBufferOffset = 0;
		return data->romRead > 0 ? kXmodemNak : kXmodemNakCrc;
	}

	// Nothing interesting happened. Return zero indicating don't do anything.

	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	DownloadThread
 *
 * DESCRIPTION:	Thread for monitoring the serial port while a ROM
 *				download is in progress.
 *
 * PARAMETERS:	param - pointer to our "download globals"
 *
 * RETURNED:	Thread result code.
 *
 ***********************************************************************/

DWORD CALLBACK DownloadThread(void* param)
{
	DownloadData*	data = (DownloadData*) param;

	// Serial buffer size.
	const long kBufferSize = 64 * 1024L;

	// Initialize our download globals.
	data->romRead			= 0;
	data->romSize			= 0;
	data->romImage			= NULL;
	data->result			= 0;

	data->readBuffer		= (BYTE*) Platform::AllocateMemory (kBufferSize, true);
	data->readBufferOffset	= 0;

	// If the serial port is already in use by an emulated ROM,
	// remember its settings and close it off so that we can re-use it.
	Bool			wasOpen			= false;
	UART::Parity	oldParity;
	int				oldStopBits;
	int				oldDataBits;
	uae_u32			oldBaud;
	Bool			oldHwHandshaking;

	wasOpen = Platform::SerialPortOpen ();
	if (wasOpen)
	{
		Platform::GetSerialPortSettings (oldParity, oldStopBits, oldDataBits,
										 oldBaud, oldHwHandshaking);
		Platform::CloseSerialPort ();
	}

	// Re-open the serial port for our own uses.
	DWORD	error;

	HANDLE	port = ::CreateFile (	data->portName,					// name
									GENERIC_READ | GENERIC_WRITE,	// access
									0,								// share mode
									NULL,							// security attributes
									OPEN_EXISTING,					// create action
									FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,	// flags & attributes
									NULL);							// template handle

	if (port == INVALID_HANDLE_VALUE)
	{
		port = NULL;
		::PrvSetDownloadError (data, ::GetLastError ());
		if (data->result == ERROR_ACCESS_DENIED)
		{
			::PrvSetDownloadError (data, kError_CantDownloadROM_SerialPortBusy | (1 << 29));
		}
		goto Exit;
	}

	if (!::SetCommMask (port, EV_ERR))
	{
		::PrvSetDownloadError (data, ::GetLastError ());
		goto Exit;
	}

	if (!::SetupComm (port, 4096, 4096))
	{
		::PrvSetDownloadError (data, ::GetLastError ());
		goto Exit;
	}

	DCB	dcb;
	dcb.DCBlength = sizeof (DCB);
	if (!::GetCommState (port, &dcb))
	{
		::PrvSetDownloadError (data, ::GetLastError ());
		goto Exit;
	}

	switch (data->speed)
	{
		case 0:		dcb.BaudRate = CBR_115200;	break;
		case 1:		dcb.BaudRate = CBR_57600;	break;
		case 2:		dcb.BaudRate = CBR_38400;	break;
		case 3:		dcb.BaudRate = CBR_19200;	break;
		case 4:		dcb.BaudRate = CBR_14400;	break;
		case 5:		dcb.BaudRate = CBR_9600;	break;
	}

	dcb.fBinary			= TRUE;
	dcb.ByteSize		= 8;
	dcb.Parity			= NOPARITY;
	dcb.fParity			= FALSE; 
	dcb.StopBits		= ONESTOPBIT;
	dcb.fOutX			= 0;
	dcb.fInX			= 0;
	dcb.fOutxDsrFlow	= FALSE;	
	dcb.fDtrControl		= DTR_CONTROL_ENABLE;

	if (dcb.BaudRate >= CBR_19200)
	{
		dcb.fOutxCtsFlow = TRUE;              
		dcb.fRtsControl  = RTS_CONTROL_HANDSHAKE;
	}
	else
	{
		dcb.fOutxCtsFlow = FALSE;
		dcb.fRtsControl  = RTS_CONTROL_ENABLE; 
	}

	COMMTIMEOUTS	cto;

	cto.ReadIntervalTimeout			= 50;
	cto.ReadTotalTimeoutMultiplier	= 0;
	cto.ReadTotalTimeoutConstant	= 100;
	cto.WriteTotalTimeoutMultiplier	= 0;
	cto.WriteTotalTimeoutConstant	= 0;
	::SetCommTimeouts (port, &cto);

	if (!SetCommState (port, &dcb))
	{
		::PrvSetDownloadError (data, ::GetLastError ());
		goto Exit;
	}

	// OVERLAPPED structure to handle overlapped i/o. Required for reliable
	// serial communications under Windows 95.

	OVERLAPPED ov;
	::ZeroMemory (&ov, sizeof (ov));
	ov.hEvent = ::CreateEvent (NULL, TRUE, FALSE, NULL);

	// Initialize Xmodem state and kick off the protocol.

	data->xmodemBlock		= 0;
	data->xmodemTimeoutBase = GetTickCount ();
	SendByte (kXmodemNakCrc);

	// Loop for as long as we're not told to quit.

	while (::WaitForSingleObject (data->quitEvent, 0) != WAIT_OBJECT_0)
	{
		// Get any data in the serial buffer.

		DWORD	bytesRead;
		if (!::ReadFile (port, data->readBuffer + data->readBufferOffset,
						kBufferSize-data->readBufferOffset, &bytesRead, &ov))
		{
			if (::GetLastError() != ERROR_IO_PENDING ||
				!::GetOverlappedResult (port, &ov, &bytesRead, TRUE))
			{
				::PrvSetDownloadError (data, GetLastError ());
				goto Exit;
			}
		}

		// Hand off the data to the function that processes it.
		// Call this function even if bytesRead is 0, as it determines
		// if a resend is in order.

		data->readBufferOffset += bytesRead;
		BYTE	c = ::HandleDownloadedData (data);

		if (c != 0)
		{
			// If HandleDownloadedData returns nonzero, that's the byte we need
			// to send to continue the protocol.

			SendByte (c);
			data->xmodemTimeoutBase = ::GetTickCount ();
		}

		// If there was an error in HandleDownloadedData, return.

		if (data->result)
		{
			break;
		}

		// If we've read all the data, return.

		if (data->romSize && data->romRead == data->romSize)
		{
			SendByte (kXmodemAck); // in case we haven't sent this yet
			break;
		}
	}

Exit:

	// If there was an error, get rid of any partially accumulated ROM image.

	if (data->result)
	{
		Platform::DisposeMemory (data->romImage);
		data->romImage = NULL;
		data->romSize = 0;
	}

	// Get rid of our temporary intermediate buffer.

	Platform::DisposeMemory (data->readBuffer);

	// Close the serial port.

	::CloseHandle (port);

	// If we closed the serial port being used for serial port emulation
	// of the ROM, re-open it.

	if (wasOpen)
	{
		Platform::OpenSerialPort ();
		Platform::SetSerialPortSettings (oldParity, oldStopBits, oldDataBits,
										 oldBaud, oldHwHandshaking);
	}

	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	DownloadROMProc1
 *
 * DESCRIPTION:	DlgProc for the first Wizard page.  Centers the
 *				dialog and handles clicks on Next.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

BOOL CALLBACK DownloadROMProc1 (HWND hDlg, UINT message, UINT wParam, LONG lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
		{
			::SetWindowLong (hDlg, DWL_MSGRESULT, 0);
			::CenterWindow (::GetParent (hDlg));
			return TRUE;
		}

		case WM_NOTIFY:
		{
			switch (((PSHNOTIFY*) lParam)->hdr.code) 
			{
				case PSN_SETACTIVE:
					PropSheet_SetWizButtons (::GetParent (hDlg), PSWIZB_NEXT);
					return 0;
			}
			break;
		}
	}

	return FALSE; 
}


/***********************************************************************
 *
 * FUNCTION:	DownloadROMProc2
 *
 * DESCRIPTION:	DlgProc for the second Wizard page.  Sets up the port
 *				and baud menus and handles selections on them.  Handles
 *				the Next and Previous buttons.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

BOOL CALLBACK DownloadROMProc2 (HWND hDlg, UINT message, UINT wParam, LONG lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
		{
			PROPSHEETPAGE*	page = (PROPSHEETPAGE*) lParam;
			DownloadData*	data = (DownloadData*) page->lParam;
			::SetWindowLong (hDlg, DWL_USER, (LONG) data);

			HWND	port = ::GetDlgItem (hDlg, IDC_PORT);
			TSettings::PopulatePortMenu (port);
			ComboBox_SetCurSel (port, 0);

			HWND	speed = ::GetDlgItem (hDlg, IDC_SPEED);
			ComboBox_AddString (speed, Platform::GetString (IDS_115200).c_str ());
			ComboBox_AddString (speed, Platform::GetString (IDS_57600).c_str ());
			ComboBox_AddString (speed, Platform::GetString (IDS_38400).c_str ());
			ComboBox_AddString (speed, Platform::GetString (IDS_19200).c_str ());
			ComboBox_AddString (speed, Platform::GetString (IDS_14400).c_str ());
			ComboBox_AddString (speed, Platform::GetString (IDS_9600).c_str ());
			ComboBox_SetCurSel (speed, 0);

			return TRUE;
		}

		case WM_NOTIFY:
		{
			switch (((PSHNOTIFY*) lParam)->hdr.code) 
			{
				case PSN_SETACTIVE:
				{
					PropSheet_SetWizButtons (GetParent (hDlg), PSWIZB_BACK | PSWIZB_NEXT);

					DownloadData*	data = (DownloadData*) ::GetWindowLong (hDlg, DWL_USER);

					ComboBox_SelectString (::GetDlgItem (hDlg, IDC_PORT), -1, data->portName);
					ComboBox_SetCurSel (::GetDlgItem (hDlg, IDC_SPEED), data->speed);

					return 0;
				}

				case PSN_KILLACTIVE:
				{
					DownloadData*	data = (DownloadData*) ::GetWindowLong (hDlg, DWL_USER);

					HWND	portCombo = ::GetDlgItem (hDlg, IDC_PORT);
					LONG	index = ComboBox_GetCurSel (portCombo);
					ComboBox_GetLBText (portCombo, index, data->portName);

					data->speed = ComboBox_GetCurSel (::GetDlgItem (hDlg, IDC_SPEED));

					return FALSE;
				}
			}
			break;
		}
	}

	return FALSE; 
}


/***********************************************************************
 *
 * FUNCTION:	DownloadROMProc3
 *
 * DESCRIPTION:	DlgProc for Wizard pages 3, 4, and 5.  Handles the
 *				Next and Previous buttons.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

BOOL CALLBACK DownloadROMProc3 (HWND hDlg, UINT message, UINT wParam, LONG lParam)
{
	switch (message)
	{
		case WM_NOTIFY:
			switch (((PSHNOTIFY*) lParam)->hdr.code) 
			{
				case PSN_SETACTIVE:
					PropSheet_SetWizButtons (::GetParent (hDlg), PSWIZB_BACK | PSWIZB_NEXT);
					return 0;
			}
			break;
	}

	return FALSE; 
}


/***********************************************************************
 *
 * FUNCTION:	DownloadROMProc6
 *
 * DESCRIPTION:	DlgProc for Wizard page 6.  Kicks off the download
 *				process, and updates the progres indicator as the
 *				ROM is received.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

BOOL CALLBACK DownloadROMProc6 (HWND hDlg, UINT message, UINT wParam, LONG lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
		{
			PROPSHEETPAGE*	page = (PROPSHEETPAGE*) lParam;
			DownloadData*	data = (DownloadData*) page->lParam;
			::SetWindowLong (hDlg, DWL_USER, (LONG) data);

			return TRUE;
		}

		case WM_NOTIFY:
		{
			PSHNOTIFY*	notify = (PSHNOTIFY*) lParam;
			switch (notify->hdr.code) 
			{
				case PSN_SETACTIVE:
				{
					DownloadData*	data = (DownloadData*) ::GetWindowLong (hDlg, DWL_USER);

					PropSheet_SetWizButtons (::GetParent (hDlg), PSWIZB_BACK);

					DWORD	tid;
					data->quitEvent = ::CreateEvent (NULL, TRUE, FALSE, NULL);
					data->downloadThread = ::CreateThread (NULL, 0, DownloadThread, (LPVOID) data, 0, &tid);

					::SetTimer (hDlg, (UINT) hDlg, 100, NULL);
					return 0;
				}

				case PSN_KILLACTIVE:	// Get this when we switch panels
			//	case PSN_QUERYCANCEL:	// Get this on cancel (no deactivate! how hokey!)
				{
					DownloadData*	data = (DownloadData*) ::GetWindowLong (hDlg, DWL_USER);

					if (data->quitEvent)
					{
						::KillTimer (hDlg, (UINT) hDlg);

						::SetEvent (data->quitEvent);
						::WaitForSingleObject (data->downloadThread, 5 * 1000L);

						::CloseHandle (data->quitEvent);
						::CloseHandle (data->downloadThread);

						data->quitEvent = NULL;
						data->downloadThread = NULL;
					}

					return FALSE;
				}
			}
			break;
		}

		case WM_TIMER:
		{
			DownloadData*	data = (DownloadData*) GetWindowLong (hDlg, DWL_USER);

			if (data->result)
			{
				KillTimer (hDlg, (UINT) hDlg);

				PropSheet_SetWizButtons (GetParent (hDlg), PSWIZB_BACK | PSWIZB_NEXT);
				PropSheet_PressButton (GetParent (hDlg), PSBTN_NEXT);

				break;
			}

			if (WaitForSingleObject (data->downloadThread, 0) == WAIT_OBJECT_0)
			{
				// The thread is signalled, which means that it's exited.

				KillTimer (hDlg, (UINT) hDlg);

				PrvSaveROMImage (hDlg, data->romImage, data->romSize, data->defaultName);

				PropSheet_SetWizButtons (GetParent (hDlg), PSWIZB_BACK | PSWIZB_NEXT);
				PropSheet_PressButton (GetParent (hDlg), PSBTN_NEXT);

				break;
			}

			if (data->romSize > 0)
			{
				SendDlgItemMessage (hDlg, IDC_PROGRESS, PBM_SETRANGE, 0, MAKELPARAM (0, data->romSize / 256));
				SendDlgItemMessage (hDlg, IDC_PROGRESS, PBM_SETPOS, data->romRead / 256, 0);
			}
		}
	}

	return FALSE; 
}


/***********************************************************************
 *
 * FUNCTION:	DownloadROMProc7
 *
 * DESCRIPTION:	DlgProc for Wizard page 7.  Displays either a success
 *				or error message.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

BOOL CALLBACK DownloadROMProc7 (HWND hDlg, UINT message, UINT wParam, LONG lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
		{
			PROPSHEETPAGE*	page = (PROPSHEETPAGE*) lParam;
			DownloadData*	data = (DownloadData*) page->lParam;

			::SetWindowLong (hDlg, DWL_USER, (LONG) data);

			return TRUE;
		}

		case WM_NOTIFY:
			switch (((PSHNOTIFY*) lParam)->hdr.code) 
			{
				case PSN_SETACTIVE:
				{
					DownloadData*	data = (DownloadData*) ::GetWindowLong (hDlg, DWL_USER);

					PropSheet_SetWizButtons (::GetParent (hDlg), PSWIZB_BACK | PSWIZB_FINISH);

					if (data->result)
					{
						string	error;

						if ((data->result & (1 << 29)) != 0)
						{
							error = Platform::GetString (
								Errors::GetIDForError (data->result & ~(1 << 29)));
						}
						else
						{
							error = Platform::GetString (
								Errors::GetIDForError (kError_CantDownloadROM_Generic));
						}

						::SetDlgItemText (hDlg, IDC_MESSAGE, error.c_str ());
					}
					else
					{
						string	msg (Platform::GetString (IDS_DOWNLOAD_DONE));
						::SetDlgItemText (hDlg, IDC_MESSAGE, msg.c_str ());
					}

					return 0;
				}

				case PSN_WIZFINISH:
				{
					DownloadData*	data = (DownloadData*) ::GetWindowLong (hDlg, DWL_USER);
					data->OK = TRUE;
					return 0;
				}
			}
			break;
	}

	return FALSE; 
}


/***********************************************************************
 *
 * FUNCTION:	DownloadROM
 *
 * DESCRIPTION:	Performs the whole process of downloading a ROM image
 *				over the serial port via a Windows Wizard.
 *
 * PARAMETERS:	parent - parent window.
 *
 *				defaultName - default name for ROM image file (may be NULL)
 *
 * RETURNED:	True if the download succeeded and the ROM was saved
 *				to a file.
 *
 ***********************************************************************/

BOOL DownloadROM (HWND parent, const char* defaultName)
{
	PROPSHEETHEADER	psh;
	PROPSHEETPAGE	psp[7];
	DLGPROC			funcs[7] = {
		DownloadROMProc1,
		DownloadROMProc2,
		DownloadROMProc3,
		DownloadROMProc3,
		DownloadROMProc3,
		DownloadROMProc6,
		DownloadROMProc7
	};

	DownloadData	dlData;
	dlData.OK			= FALSE;
	dlData.portName[0]	= 0;
	dlData.speed		= 0;
	dlData.defaultName	= defaultName;

	memset (psp, 0, sizeof (psp));

	for (int ii = 0; ii < 7; ++ii)
	{
		psp[ii].dwSize		= sizeof (PROPSHEETPAGE);
		psp[ii].dwFlags		= 0;
		psp[ii].hInstance	= gInstance;
		psp[ii].pszTemplate	= MAKEINTRESOURCE (IDD_DOWNLOAD_WIZ1 + ii);
		psp[ii].pfnDlgProc	= funcs[ii];
		psp[ii].lParam		= (LPARAM) &dlData;
	}

	string	downloadROM (Platform::GetString (IDS_DOWNLOAD_ROM));

	psh.dwSize		= sizeof (PROPSHEETHEADER);
	psh.dwFlags		= PSH_PROPSHEETPAGE | PSH_WIZARD | PSH_NOAPPLYNOW;
	psh.hwndParent	= parent;
	psh.hInstance	= gInstance;
	psh.pszIcon		= NULL;
	psh.pszCaption	= downloadROM.c_str ();
	psh.nPages		= sizeof (psp) / sizeof (PROPSHEETPAGE);
	psh.nStartPage	= 0;
	psh.ppsp		= (LPCPROPSHEETPAGE) &psp;
	psh.pfnCallback	= NULL;

	// Launch the wizard

	::PropertySheet (&psh);

	// Return TRUE if the user clicked on Finish and there was no error.

	return dlData.OK && (dlData.result == 0);
}

