/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "Emulator.h"

#include "Application.h"		// StartUIThread
#include "CPU_MT.h"				// CPU_dispose
#include "CPU_REG.h"			// Emulator::kUnspecified
#include "Document.h"			// QueryNewDocument, QueryOldDocument, etc.
#include "DownloadROM.h"		// DownloadROM
#include "ErrorHandling.h"		// Errors::ReportIfError
#include "LCD.h"				// CenterWindow
#include "Settings.h"			// gSettings
#include "Strings.r.h"			// kStr_CmdNew, kStr_CmdOpen

#include "resource.h"


// ======================================================================
//	Globals and constants
// ======================================================================

HINSTANCE	gInstance;


/***********************************************************************
 *
 * FUNCTION:	PrvStartupDlgProc
 *
 * DESCRIPTION:	DlgProc for the startup dialog (the one with the New,
 *				Open, etc. buttons).  Centers the window, and ends
 *				the dialog once any of the buttons are clicked.
 *
 * PARAMETERS:	Standard DlgProc parameters
 *
 * RETURNED:	Standard DlgProc result
 *
 ***********************************************************************/

BOOL CALLBACK PrvStartupDlgProc (	HWND hwnd,
									UINT msg,
									WPARAM wparam,
									LPARAM lparam)
{
	switch (msg)
	{
		case WM_INITDIALOG:
		{
			::CenterWindow (hwnd);
			break;
		}

		case WM_COMMAND:
		{
			switch (LOWORD (wparam))
			{
				case IDC_NEW:
				case IDC_OPEN:
				case IDC_DOWNLOAD:
				case IDC_EXIT:
				case IDCANCEL:
					::EndDialog (hwnd, LOWORD (wparam));
					break;
			}

			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}


/***********************************************************************
 *
 * FUNCTION:	PrvStartupScreen
 *
 * DESCRIPTION:	Presents the user with the New/Open/etc dialog.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	Button the user clicked on.
 *
 ***********************************************************************/

int PrvStartupScreen (void)
{
	int	choice = ::DialogBox (	gInstance,
								MAKEINTRESOURCE (IDD_STARTUP),
								NULL,	// hwndParent
								&::PrvStartupDlgProc);
	return choice;
}


/***********************************************************************
 *
 * FUNCTION:	PrvForceStartupScreen
 *
 * DESCRIPTION:	See if the user has requested that the Startup dialog
 *				be presented instead of attempting to use the latest
 *				session file or creation settings.
 *
 *				The current signal is to toggle the CAPSLOCK key.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	True if the user has signalled that the dialog should
 *				be presented.
 *
 ***********************************************************************/

Bool PrvForceStartupScreen (void)
{
	return ::GetKeyState (VK_CAPITAL) != 0;
}

/***********************************************************************
 *
 * FUNCTION:	PrvSetRegKey
 *
 * DESCRIPTION:	Set an entry in the Windows system registry.
 *
 * PARAMETERS:	lpszKey - key for the entry
 *
 *				lpszValue - value of the entry
 *
 *				lpszValueName - name for the value.  Could be NULL,
 *					in which case the value is assigned as the value
 *					of the key itself.
 *
 * RETURNED:	True if the operation was a success; false if the
 *				patient died.
 *
 ***********************************************************************/

static Bool PrvSetRegKey (LPCTSTR lpszKey, LPCTSTR lpszValue, LPCTSTR lpszValueName = NULL)
{
	if (lpszValueName == NULL)
	{
		if (::RegSetValue (	HKEY_CLASSES_ROOT,
							lpszKey, REG_SZ,
							lpszValue,
							lstrlen(lpszValue) * sizeof(TCHAR)) != ERROR_SUCCESS)
		{
			return FALSE;
		}

		return TRUE;
	}

	HKEY hKey;

	if (::RegCreateKey (HKEY_CLASSES_ROOT, lpszKey, &hKey) == ERROR_SUCCESS)
	{
		LONG	lResult = ::RegSetValueEx (	hKey,
											lpszValueName,
											0,
											REG_SZ,
											(CONST BYTE*) lpszValue,
											(lstrlen(lpszValue) + 1) * sizeof(TCHAR));

		if (::RegCloseKey (hKey) == ERROR_SUCCESS && lResult == ERROR_SUCCESS)
		{
			return TRUE;
		}
	}

	return FALSE;

}


/***********************************************************************
 *
 * FUNCTION:	PrvRegisterTypes
 *
 * DESCRIPTION:	Register our file types and icons.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

static void PrvRegisterTypes (void)
{
	/*
		The following is from Eric Cloninger:

			Key                             Default Value
			HKEY_CLASSES_ROOT\\.psf         "POSE_file"
			HKEY_CLASSES_ROOT\\POSE_file    "POSE File"    <<< Shows up in Explorer
			HKEY_CLASSES_ROOT\\POSE_file\\shell\open\command  "c:\pose\emulator.exe"
			HKEY_CLASSES_ROOT\\POSE_file\\DefaultIcon         "c:\pose\emulator.exe,1"

		For more details, see MSDN Article ID: Q122787. (Also Article ID: Q142275)

		Also:

			Finding the path

			Rather than jam your application's files into Windows directories, you can
			now place them in a more logical location, and the system can still find
			them. For example, you register the path where you locate your executable
			(.EXE) files as well as dynamic-link library (DLL) files by using the
			AppPaths subkey under
			HKEY_LOCAL_MACHINE\Software\Microsoft\Windows\CurrentVersion. Add a subkey
			using your application executable filename and set its default value to
			the path for your .EXE file. Add an additional value named Path for this
			subkey, and set its value to the path of your DLLs. The system will
			automatically update the path for these entries if a user moves or renames
			the application's executable file using the standard system UI.
	*/

	char	moduleName[_MAX_PATH];
	::GetModuleFileName (gInstance, moduleName, _MAX_PATH);

	char	icon[_MAX_PATH];
	::sprintf (icon, "%s,-%d", moduleName, IDI_DOCUMENT);

	char	open[_MAX_PATH];
	::sprintf (open, "\"%s \"%%1\"\"", moduleName);

	::PrvSetRegKey (".psf", "psffile");
	::PrvSetRegKey ("psffile", "Palm Emulator Session");

	::PrvSetRegKey (".rom", "romfile");
	::PrvSetRegKey ("romfile", "Palm Emulator ROM");

	::PrvSetRegKey ("psffile\\DefaultIcon", icon);
	::PrvSetRegKey ("romfile\\DefaultIcon", icon);

	::PrvSetRegKey ("psffile\\shell\\open\\command", open);
	::PrvSetRegKey ("romfile\\shell\\open\\command", open);
}


/***********************************************************************
 *
 * FUNCTION:	WinMain
 *
 * DESCRIPTION:	Application entry point.  Attempts to restart a
 *				previous session, or asks the user if they want to
 *				start a new one.
 *
 * PARAMETERS:	Standard WinMain parameters
 *
 * RETURNED:	Standard WinMain result
 *
 ***********************************************************************/

int WINAPI WinMain (HINSTANCE instance, HINSTANCE, LPSTR, int)
{
	try
	{
		gInstance = instance;


		// When run under the VC++ debugger, the default directory appears
		// to be set to S:\emulator\windows\source, instead of the expected
		// \Debug or \Release sub-directories.  Jam the current directory
		// to the app directory so that we can find our files.

		{
			char	path[_MAX_PATH], drive[_MAX_DRIVE], dir[_MAX_DIR];

			::GetModuleFileName (gInstance, path, _MAX_PATH);
			_splitpath (path, drive, dir, NULL, NULL);
			_makepath (path, drive, dir, NULL, NULL);
			::SetCurrentDirectory (path);
		}


		::PrvRegisterTypes ();


		// Our main application loop merely loops between getting commands
		// and executing them.  The commands are specified in a parameter
		// block called QuitParams.  QuitParams contains a command number
		// and any parameters for the command.
		//
		// Commands can come from one of two sources: the user and the user.
		// Uh...by that I mean that...hmmm, let me start again.
		//
		// The application normally starts up in a "non running" state.
		// That is, it is not emulating a ROM. In that state, it is
		// displaying a dialog asking the user what to do.  That dialog
		// allows the user to select from "New document", "Open document",
		// "Download ROM", and "Exit".  When the user selects one of these
		// options, the QuitParams parameter block is filled out and
		// dispatched on.
		//
		// After the user has started an emulation session, he can
		// choose to close that session or exit the emulator altogether
		// ("He can choose to close that session or exit the emulator").
		// He can also choose to create a new session or open a previously
		// saved one.  Any of those choices will cause the UI thread to
		// fill out a QuitParams parameter block and exit back to the
		// main thread (this one).
		//
		// We start off by filling out the parameter block with no command.
		// That will force our main loop to ask the use for something to do.

		QuitParams	quitParams;
		quitParams.fCommand = QuitParams::kNone;


		// See if there is some "auto startup" action to do.  If the user had
		// previously saved a session file, attempt to reload it.  Otherwise,
		// if the user had previously created a session file (but didn't
		// save it), attempt to create a new session based on those settings.
		// If neither of those cases is true, or if the use is holding down
		// the Caps Lock key, then keep the QuitParams command at "none".

		// See if the user wants to force the appearance of the Startup
		// screen (by holding down the ShiftLock key) and skip the auto-
		// loading of the previous session file.

		if (!::PrvForceStartupScreen ())
		{
			AppPreferences::Configuration	cfg = gPrefs->GetLastConfiguration ();
			FileReference	ramFileRef	= gPrefs->GetLastRAMFile ();


			// See if there was a previously saved RAM file.  If so, open it.

			if (ramFileRef.IsSpecified ())
			{
				quitParams.fCommand		= QuitParams::kOpen;
				quitParams.fRAMFileRef	= ramFileRef;
			}


			// See if there was a previously created document.  If so,
			// create a new document based on its settings.

			else if (cfg.fDeviceType != Emulator::kUnspecified &&
					cfg.fSilkscreenType != AppPreferences::kSilkscreenUnspecified &&
					cfg.fRAMSize != 0 &&
					cfg.fROMFile.IsSpecified ())
			{
				quitParams.fCommand		= QuitParams::kNew;
				quitParams.fCfg			= cfg;
			}


			// See if there is a file specified on the command line.  A file
			// there takes precedence over any default auto-startup sequence.

			if (__argc > 1)
			{
				const char*	fileName = __argv[1];

				if (::EndsWith (fileName, ".psf"))
				{
					// It's a session file.  Open it up.

					quitParams.fCommand		= QuitParams::kOpen;
					quitParams.fRAMFileRef	= FileReference (fileName);
				}
				else if (::EndsWith (fileName, ".rom"))
				{
					// It's a ROM file.  Bring up the New Session dialog with
					// the chosen ROM file written in there.

					quitParams.fCommand			= QuitParams::kNone;
					quitParams.fCfg				= cfg;
					quitParams.fCfg.fROMFile	= FileReference (fileName);
					if (::QueryNewDocument (NULL, quitParams.fCfg))
					{
						quitParams.fCommand = QuitParams::kNew;
					}
				}
				else
				{
					// It's an unknown file type.  Bring up the dialog with
					// the New/Open/Download/Exit buttons.

					quitParams.fCommand = QuitParams::kNone;
				}
			}
		}


		// Now loop, alternating between getting a command from the user
		// via the Startup dialog, and dispatching on that command.  Most
		// of the time, we just create the CPU and UI threads and let them
		// do the rest of the work.

		Bool	quitOuterLoop = false;

		while (!quitOuterLoop)
		{
			Bool	doUI		= false;
			ErrCode	error		= 0;
			int		operation	= 0;


			// If there's nothing to do in QuitParams, get a command from the user.

			if (quitParams.fCommand == QuitParams::kNone)
			{
				int	command	= ::PrvStartupScreen ();

				switch (command)
				{
					case IDC_NEW:

						// Get the last defaults in this area.

						quitParams.fCfg = gPrefs->GetLastConfiguration ();

						if (::QueryNewDocument (NULL, quitParams.fCfg))
						{
							quitParams.fCommand = QuitParams::kNew;
						}
						break;

					case IDC_OPEN:
						if (::QueryOpenDocument (NULL, quitParams.fRAMFileRef))
						{
							quitParams.fCommand = QuitParams::kOpen;
						}
						break;

					case IDC_DOWNLOAD:
						::DownloadROM (NULL, "Palm.ROM");
						break;

					case IDC_EXIT:
					case IDCANCEL:
						quitParams.fCommand = QuitParams::kQuit;
						break;
				}
			}


			// Now switch on QuitParams::fCommand. It's possible that
			// that it's still "kNone" if the user had cancelled a
			// QueryNewDocument or QueryOpenDocument dialog, or had
			// just downloaded a ROM.

			switch (quitParams.fCommand)
			{
				case QuitParams::kNew:
					error = ::CreateNewDocument (quitParams.fCfg);
					doUI = true;
					operation = kStr_CmdNew;
					break;

				case QuitParams::kOpen:
					error = ::OpenOldDocument (quitParams.fRAMFileRef);
					doUI = true;
					operation = kStr_CmdOpen;
					break;

				case QuitParams::kNone:
					break;

				case QuitParams::kQuit:
					quitOuterLoop = true;
					break;
			}


			// If there was some sort of error starting up the CPU thread,
			// report it.

			if (error)
			{
				Errors::ReportIfError (operation, error, 0, false);
				quitParams.fCommand = QuitParams::kNone;
			}


			// Otherwise, if we need to fire up the UI thread, do it.

			else if (doUI)
			{
				// Start up the UI thread. This thread responds to user events and
				// periodically updates the LCD display.  WaitForSingleObject will block
				// the current thread until the UI thread object is signalled (which
				// occurs when the thread is terminated).

				HANDLE	uiThread = ::StartUIThread (&quitParams);
				::WaitForSingleObject (uiThread, INFINITE);
				::CloseHandle (uiThread);

				::CPU_dispose ();
			}
		}
	}
	catch (...)
	{
		::MessageBox (NULL, "Unhandled exception caught in WinMain.\nDying like a log...", "Fatal Error", MB_OK);
	}

	return 0;
}
