/* ================================================================================== */
/* Copyright (c) 1998-1999 3Com Corporation or its subsidiaries. All rights reserved. */
/* ================================================================================== */

#include "EmulatorCommon.h"
#include "LCD.h"

#include <WindowsX.h>			// SelectBitmap, SelectBrush, SelectPen

#include "Application.h"		// DisplayWndProc
#include "CPU_REG.h"			// Screen::LCDIsOn, Screen::BacklightIsOn, Screen::GetBits, etc.
#include "CPU_MT.h"				// CPU_ functions
#include "Emulator.h"			// gInstance
#include "Settings.h"			// gSettings
#include "Strings.r.h"			// kStr_AppName

#include "resource.h"			// IDB_ENGLISH_PP_SMALL, etc.

#include <malloc.h>				// _alloca


// ======================================================================
//	Globals and constants
// ======================================================================

const char*	kClassName			= "PalmOSEmulatorDisplay";

const int kPlasticWidth1x		= 219;
const int kPlasticHeight1x		= 333;
const int kPlasticBaseX1x		= 32;
const int kPlasticBaseY1x		= 32;

const int kPlasticWidth2x		= 436;
const int kPlasticHeight2x		= 664;
const int kPlasticBaseX2x		= kPlasticBaseX1x * 2;
const int kPlasticBaseY2x		= kPlasticBaseY1x * 2;


// ======================================================================
//	Structure and data for defining the outlines for our windows.  Each
//	outline is defined as an array of "tripoints".  A "tripoint" is a
//	structure with three fields: a y coordinate and two x coordinates.
//	These three values define a scanline subrange that belongs to the
//	inside of the region.  This subrange remains in effect for
//	subsequent scanlines until we reach the scanline indicated by the
//	next tripoint.
// ======================================================================

struct TriPoint
{
	int	y;
	int	x1;
	int x2;
};

TriPoint gSmallPoints [] =
{
	{   0,   2, 217 },
	{   1,   1, 218 },
	{   2,   0, 219 },
	{ 305,   1, 219 },
	{ 306,   2, 219 },
	{ 307,   8, 219 },
	{ 309,   8, 218 },
	{ 310,   8, 216 },
	{ 311,   8, 214 },
	{ 312,   8, 212 },
	{ 313,   8, 211 },
	{ 331,   9, 210 },
	{ 332,  10, 209 },
	{ 333,  -1,  -1 }
};

TriPoint gLargePoints [] =
{
	{   0,   7, 429 },
	{   1,   5, 431 },
	{   2,   4, 432 },
	{   3,   3, 433 },
	{   4,   2, 434 },
	{   5,   1, 435 },
	{   6,   1, 435 },
	{   7,   0, 436 },
	{ 605,   1, 436 },
	{ 607,   2, 436 },
	{ 608,   3, 436 },
	{ 609,   4, 436 },
	{ 610,   6, 436 },
	{ 611,   8, 436 },
	{ 612,  16, 436 },
	{ 614,  16, 435 },
	{ 616,  16, 434 },
	{ 617,  16, 433 },
	{ 618,  16, 431 },
	{ 619,  16, 429 },
	{ 620,  16, 426 },
	{ 621,  16, 423 },
	{ 622,  16, 420 },
	{ 656,  17, 420 },
	{ 657,  17, 419 },
	{ 658,  18, 419 },
	{ 659,  18, 418 },
	{ 660,  19, 417 },
	{ 661,  20, 416 },
	{ 662,  21, 415 },
	{ 663,  23, 413 },
	{ 664,  -1,  -1 }
};


const RECT		kHardButtons [] =
{
	{ 1, 274, 1 + 16, 274 + 24 },		// Power button
	{ 96, 272, 96 + 32, 272 + 16 },		// Up button
	{ 96, 293, 96 + 32, 293 + 16 },		// Down button
	{ 23, 270, 23 + 32, 270 + 32 },		// App1 button
	{ 60, 270, 60 + 32, 270 + 32 },		// App2 button
	{ 131, 270, 131 + 32, 270 + 32 },	// App3 button
	{ 168, 270, 168 + 32, 270 + 32 },	// App4 button
	{ 0, 266, 0 + 204, 266 + 47 }		// Dead area around the buttons
};

HWND			gEmulatorWindow;

HBITMAP			gBitmap;
HPALETTE		gPalette;

BOOL			gMoveDrag;
RECT			gDragOrigRect;
POINT			gDragBase;
BOOL			gPenTracking;
BOOL			gButtonTracking;
int				gCurrentButton;


// ======================================================================
//	Private functions
// ======================================================================

static void		PrvSetColorRamp (RGBQUAD colors[], int numColors, Screen::RGBType* palette);
static HBITMAP	PrvLoadResourceBitmap (HINSTANCE hInstance, LPSTR lpString, HPALETTE* lphPalette);
static HPALETTE	PrvCreateDIBPalette (LPBITMAPINFO lpbmi, LPINT lpiNumColors);


#define WIDTH(r)	((r).right - (r).left)
#define HEIGHT(r)	((r).bottom - (r).top)


/***********************************************************************
 *
 * FUNCTION:	CenterWindow
 *
 * DESCRIPTION:	Centers the given window the desktop window.
 *
 *				(This routine doesn't need to be in this file, but
 *				 there's no better place for it.)
 *
 * PARAMETERS:	hwnd - window to center.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void CenterWindow (HWND hwnd)
{
	RECT	workArea;
	::SystemParametersInfo (SPI_GETWORKAREA, NULL, &workArea, NULL);

	RECT	dRect;
	::GetWindowRect (hwnd, &dRect);

	int xLeft	= (workArea.left + workArea.right) / 2 - (dRect.right - dRect.left) / 2;
	int yTop	= (workArea.top + workArea.bottom) / 2 - (dRect.bottom - dRect.top) / 2;

	::SetWindowPos (hwnd, NULL, xLeft, yTop, -1, -1, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
}


/***********************************************************************
 *
 * FUNCTION:	CenterDlgProc
 *
 * DESCRIPTION:	DlgProc for centering dialogs.
 *
 *				(This routine doesn't need to be in this file, but
 *				 there's no better place for it.)
 *
 * PARAMETERS:	Standard DlgProc parameters.
 *
 * RETURNED:	Standard DlgProc result.
 *
 ***********************************************************************/

UINT CALLBACK CenterDlgProc (	HWND hwnd,
								UINT msg,
								WPARAM wparam,
								LPARAM lparam)
{
	switch (msg)
	{
		case WM_INITDIALOG:
		{
			::CenterWindow (::GetParent (hwnd));
			break;
		}

		default:
			return FALSE;
	}

	return TRUE;
}


/***********************************************************************
 *
 * FUNCTION:	PrvSetColorRamp
 *
 * DESCRIPTION:	Helper routine for LCD_ScreenShot.  Fills in a color
 *				table with appropriate gray values.
 *
 * PARAMETERS:	colors - RGBQUAD array to receive the colors.
 *
 *				numColors - number of colors to write into "colors".
 *
 *				foreground, background - starting and ending colors of
 *				the ramp to create.
 *
 * RETURNED:	New colors are written into "colors".
 *
 ***********************************************************************/

static void PrvSetColorRamp (RGBQUAD colors[], int numColors, 
				   Screen::RGBType* palette)
{
	for (int color = 0; color < numColors; color++)
	{
		colors[color].rgbBlue		= palette[color].b;	
		colors[color].rgbGreen		= palette[color].g;
		colors[color].rgbRed		= palette[color].r;
		colors[color].rgbReserved	= 0;
	}
}



/***********************************************************************
 *
 * FUNCTION:	PaintScreen
 *
 * DESCRIPTION:	Called by LCD_Refresh and LCD_Paint to do the real
 *				dirty work.
 *
 * PARAMETERS:	dc - HDC to draw to.
 *
 *				always - true to redraw the whole screen, false to draw
 *				just the parts that were changed by the emulated process
 *				since the last time we updated.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void PaintScreen (HDC dc, BOOL always)
{
	// If we're doing a paint in response to an update message, tell the
	// Screen class to return *all* the bits, not just the ones that changed.

	if (always)
	{
		Screen::InvalidateAll ();
	}

	// Get information about the screen.  Return immediately if nothing's changed.

	Screen::BufferInfo	info;
	if (!Screen::GetBits (info, true /* incremental */))
		return;

	// Get the bounds of the area we'll be blitting from.

	RECT	srcRect;
	srcRect.top		= info.firstLine;
	srcRect.bottom	= info.lastLine;
	srcRect.left	= info.visibleBounds.left;
	srcRect.right	= info.visibleBounds.right;

	// There is a moment during boot-up when the LCD width is 1024.
	// We don't really want that as the width (we only allow for 160
	// pixels in our window), so clip to 160.

	if (WIDTH (srcRect) > 160)
	{
		srcRect.right = srcRect.left + 160;
	}

	// Use that as the start of the area we'll be blitting to.

	RECT	destRect = srcRect;

	// Normalize the dest rect to 0,0.

	destRect.right -= destRect.left;
	destRect.left = 0;

	// If we're using the gray background, we'll need to offset the dest rect.

	if (gPrefs->GrayBackground ())
	{
		destRect.top	+= kPlasticBaseY1x;
		destRect.bottom	+= kPlasticBaseY1x;
		destRect.left	+= kPlasticBaseX1x;
		destRect.right	+= kPlasticBaseX1x;
	}

	// Now scale according to the user's preferences.

	destRect.top	*= gPrefs->GetScale ();
	destRect.bottom	*= gPrefs->GetScale ();
	destRect.left	*= gPrefs->GetScale ();
	destRect.right	*= gPrefs->GetScale ();

	// If the screen is off, just paint the background color.

	if (!info.lcdOn)
	{
		COLORREF	background = RGB (	info.myCLUT[0].r,
										info.myCLUT[0].g,
										info.myCLUT[0].b);
		::SetBkColor (dc, background);
		::ExtTextOut (dc, 0, 0, ETO_OPAQUE, &destRect, NULL, 0, NULL);

		return;
	}


	// Create the GDI objects we'll need to copy the bitmap to the screen.
	//
	// Interestingly, caching the objects we create here and using them in
	// subsequent calls gets us nothing.  So, for simplicity's sake, we'll
	// keep the straightforward implementation of creating the objects
	// every time.

	int	numColors = 1 << info.lcdDepth;

	// Fill in a BITMAPINFO structure accordingly.

	BITMAPINFO*	bmi = (BITMAPINFO*) _alloca (sizeof (BITMAPINFOHEADER) +
											numColors * sizeof (RGBQUAD));
	::ZeroMemory (bmi, sizeof (BITMAPINFOHEADER));
	bmi->bmiHeader.biSize		= sizeof (BITMAPINFOHEADER);
	bmi->bmiHeader.biWidth		= info.myRowBytes * 8 / info.myDepth;
	bmi->bmiHeader.biHeight		= -info.height;
	bmi->bmiHeader.biPlanes		= 1;
	bmi->bmiHeader.biBitCount	= info.myDepth;
	bmi->bmiHeader.biClrUsed	= numColors;

	::PrvSetColorRamp (bmi->bmiColors, numColors, info.myCLUT);

	// Create the DIB section.  The last two parameters to CreateDIBSection
	// are hSection and dwOffset; by setting them to NULL/0, we tell Windows
	// to allocate the memory for us.

	VOID*	bits;
	HBITMAP	bmp = ::CreateDIBSection (dc, bmi, DIB_RGB_COLORS, &bits, NULL, 0);

	if (!bmp || !bits)
	{
		// If we couldn't get that, just return.  I'm not sure why CreateDIBSection
		// would fail, but I've seen it happen.  It usually happens when another
		// copy of the emulator contacts this one in order to download the ROM.
		// Perhaps the intense serial port activity is getting in the way.
		// Whatever.  But subsequent attempts to call CreateDIBSection appear to
		// work.  GetLastError is returning zero.

		DWORD	error = ::GetLastError ();

		return;
	}

	// Copy our bits into it. They should already be in the right format
	// from the Screen::GetBits call.
	//
	// Note: modifying GetBits to copy directly into the "bits" buffer (instead
	// of using an intermediate buffer) doesn't get us anything (according
	// to timing tests).

	memcpy (((char*) bits) + info.firstLine * info.myRowBytes,
			((char*) info.myBuffer) + info.firstLine * info.myRowBytes,
			(info.lastLine - info.firstLine) * info.myRowBytes);

	// Create a DC for blitting purposes.

	HDC		mdc = ::CreateCompatibleDC (NULL);
	HBITMAP	oldBitmap = (HBITMAP) ::SelectObject (mdc, bmp);

	// Blit the image to the screen.

	if (gPrefs->GetScale () == 1)
	{
		::BitBlt (	dc,						// handle to destination device context
					destRect.left,			// x-coordinate of destination rectangle's upper-left corner
					destRect.top,			// y-coordinate of destination rectangle's upper-left corner
					WIDTH (destRect),		// width of destination rectangle
					HEIGHT (destRect),		// height of destination rectangle
					mdc,					// handle to source device context
					srcRect.left,			// x-coordinate of source rectangle's upper-left corner
					srcRect.top,			// y-coordinate of source rectangle's upper-left corner
					SRCCOPY);				// raster operation code
	}
	else
	{
		::StretchBlt (	dc,					// handle to destination device context
						destRect.left,		// x-coordinate of upper-left corner of dest. rect.
						destRect.top,		// y-coordinate of upper-left corner of dest. rect.
						WIDTH (destRect),	// width of destination rectangle
						HEIGHT (destRect),	// height of destination rectangle
						mdc,				// handle to source device context
						srcRect.left,		// x-coordinate of upper-left corner of source rectangle
						srcRect.top,		// y-coordinate of upper-left corner of source rectangle
						WIDTH (srcRect),	// width of source rectangle
						HEIGHT (srcRect),	// height of source rectangle
						SRCCOPY);
	}

	::SelectObject (mdc, oldBitmap);
	::DeleteDC (mdc);
	::DeleteObject (bmp);
}


/***********************************************************************
 *
 * FUNCTION:	LCD_Init
 *
 * DESCRIPTION:	Called when the application starts up.  Loads in any
 *				resources, sets up color palettes, and creates the
 *				emulator LCD window.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_Init (void)
{
	// Initialize the flag used for emulating sounds based on our preference settings.
	// !!! TBD

//	::CPU_setsounds (gSettings.Sounds != 0);

	// Create the window class for the emulator window.

	WNDCLASS	wc;
	::ZeroMemory (&wc, sizeof (wc));

	wc.style			= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc		= &::DisplayWndProc;
	wc.hIcon			= ::LoadIcon (gInstance, MAKEINTRESOURCE (IDI_EMULATOR));
	wc.hCursor			= ::LoadCursor (NULL, IDC_ARROW);
	wc.lpszClassName	= kClassName;

	::RegisterClass (&wc);

	// Create the emulator window and show it.

	string	app (Platform::GetString (kStr_AppName));

	gEmulatorWindow = ::CreateWindow (	kClassName,
										app.c_str (),
										WS_CAPTION | WS_MINIMIZEBOX | WS_OVERLAPPED,
										gPrefs->GetWindowLocation().h,
										gPrefs->GetWindowLocation().v,
										CW_USEDEFAULT, CW_USEDEFAULT,
										NULL, NULL, 0, NULL);

	::LCD_ResetWindow ();

	::ShowWindow (gEmulatorWindow, SW_SHOWNORMAL);
	::SetForegroundWindow (gEmulatorWindow);
}


/***********************************************************************
 *
 * FUNCTION:	LCD_Quit
 *
 * DESCRIPTION:	Called when the application is about to end.  Releases
 *				our resources.  Doesn't need to delete the window, as
 *				its deletion was what got us here in the first place.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_Quit (void)
{
	// Release our resources before quitting.

	::DeleteObject (gBitmap);
	::DeleteObject (gPalette);
}


/***********************************************************************
 *
 * FUNCTION:	LCD_GetWindow
 *
 * DESCRIPTION:	Returns the HWND of the emulator window for any
 *				external routines that need it (such as those that need
 *				to create a new window with the LCD window as the
 *				parent).
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	HWND of the emulator LCD window.
 *
 ***********************************************************************/

HWND LCD_GetWindow (void)
{
	return gEmulatorWindow;
}


/***********************************************************************
 *
 * FUNCTION:	LCD_SetWindow
 *
 * DESCRIPTION:	Sets the HWND returned by LCD_GetWindow.  Called by the
 *				LCD's WndProc when the window is destroyed.
 *
 * PARAMETERS:	HWND of the emulator LCD window.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_SetWindow (HWND hwnd)
{
	gEmulatorWindow = hwnd;
}


/***********************************************************************
 *
 * FUNCTION:	LCD_ResetWindow
 *
 * DESCRIPTION:	Called in response to WM_CREATE messages, and when the
 *				user changes the appropriate settings in the Properties
 *				dialog box.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_ResetWindow (void)
{
	RECT wr, cr;

	// Set up some flags used when calling SetWindowPos.

	const UINT	swpFlags = SWP_FRAMECHANGED | SWP_NOACTIVATE | SWP_NOMOVE |
							SWP_NOSIZE | SWP_NOZORDER | SWP_NOCOPYBITS;

	// Delete the old bitmap and palette.

	::DeleteObject (gBitmap);
	::DeleteObject (gPalette);

	// Get the current (new) configuration.

	AppPreferences::Configuration	cfg = gPrefs->GetLastConfiguration ();

	// Based on the configuration, determine what bitmaps to use.

	UINT	resID = 0;
	if (gPrefs->GrayBackground ())
	{
		if (cfg.fSilkscreenType == AppPreferences::kSilkscreenJapanese)
		{
			if (gPrefs->GetScale () == 1)
				resID = IDB_JAPANESE_PP_SMALL;
			else
				resID = IDB_JAPANESE_PP_LARGE;
		}
		else
		{
			if (gPrefs->GetScale () == 1)
				resID = IDB_ENGLISH_PP_SMALL;
			else
				resID = IDB_ENGLISH_PP_LARGE;
		}
	}
	else
	{
		if (gPrefs->GetScale () == 1)
			resID = IDB_INTL_NONE_SMALL;
		else
			resID = IDB_INTL_NONE_LARGE;
	}

	// Get the determined/desired bitmap.  PrvLoadResourceBitmap is a private
	// function that also extracts and returns palette information.

	gBitmap = ::PrvLoadResourceBitmap (gInstance, MAKEINTRESOURCE (resID), &gPalette);

	// Change the window to accomodate the settings and bitmap.

	if (gPrefs->GrayBackground ())
	{
		// Change the window attributes, and then tell Windows about it.
		::SetWindowLong (gEmulatorWindow, GWL_STYLE, ::GetWindowLong (gEmulatorWindow, GWL_STYLE) & ~WS_CAPTION);
		::SetWindowPos (gEmulatorWindow, NULL, 0, 0, 0, 0, swpFlags);

		// Get the new window size/location.
		::GetWindowRect (gEmulatorWindow, &wr);
		::GetClientRect (gEmulatorWindow, &cr);

		// Resize the window.
		if (gPrefs->GetScale () == 1)
			::MoveWindow (gEmulatorWindow, wr.left, wr.top,
					kPlasticWidth1x + WIDTH(wr) - WIDTH(cr),
					kPlasticHeight1x + HEIGHT(wr) - HEIGHT(cr), TRUE);
		else
			::MoveWindow (gEmulatorWindow, wr.left, wr.top,
					kPlasticWidth2x + WIDTH(wr) - WIDTH(cr),
					kPlasticHeight2x + HEIGHT(wr) - HEIGHT(cr), TRUE);

		// Generate the framing region for the window.
		HRGN		region = NULL;
		HRGN		tempRgn = NULL;
		TriPoint*	p = NULL;

		// Get the right set of tripoints.
		if (gPrefs->GetScale () == 1)
			p = &gSmallPoints[0];
		else
			p = &gLargePoints[0];

		// Loop until we get to the y,-1,-1 at the end of the array.
		while (p[0].x1 != -1)
		{
			// If we haven't created out framing region yet, create it,
			// using the first rectangle indicated by the tripoint array.
			// This happens only the first time through the loop.
			if (region == NULL)
			{
				region = ::CreateRectRgn (p[0].x1, p[0].y, p[0].x2, p[1].y);
			}
			else
			{
				// Subsequent times through the loop, we get here, creating
				// a new region and merging it with the first one.  If we
				// haven't created that new region yet, do so.  Otherwise,
				// modify that region to contain the next piece of the
				// final region.
				if (tempRgn == NULL)
					tempRgn = ::CreateRectRgn (p[0].x1, p[0].y, p[0].x2, p[1].y);
				else
					SetRectRgn (tempRgn, p[0].x1, p[0].y, p[0].x2, p[1].y);

				::CombineRgn (region, region, tempRgn, RGN_OR);
			}

			++p;
		}

		if (tempRgn)
			::DeleteObject (tempRgn);

		// Set the framing region for the window.
		::SetWindowRgn (gEmulatorWindow, region, TRUE);
	}
	else
	{
		// We're using a plain white window.

		// Clear out any previous framing region.
		::SetWindowRgn (gEmulatorWindow, NULL, TRUE);

		// Change the window attributes, and then tell Windows about it.
		::SetWindowLong (gEmulatorWindow, GWL_STYLE, ::GetWindowLong (gEmulatorWindow, GWL_STYLE) | WS_CAPTION);
		::SetWindowPos (gEmulatorWindow, NULL, 0, 0, 0, 0, swpFlags);

		// Get the new window size/location.
		::GetWindowRect (gEmulatorWindow, &wr);
		::GetClientRect (gEmulatorWindow, &cr);

		// Resize the window.
		::MoveWindow (gEmulatorWindow, wr.left, wr.top,
					160 * gPrefs->GetScale () + WIDTH(wr) - WIDTH(cr),
					(160 + 60) * gPrefs->GetScale () + HEIGHT(wr) - HEIGHT(cr),
					TRUE);
	}

	// Set a timer so that we periodically refresh the window contents.
	// Also, invalidate the window contents now (necessary?).
	::SetTimer (gEmulatorWindow, 1, 1000 / 10 /*gSettings.FPS*/, NULL);
	::InvalidateRect (gEmulatorWindow, NULL, FALSE);
}


/***********************************************************************
 *
 * FUNCTION:	LCD_PenEvent
 *
 * DESCRIPTION:	Called in response to WM_LBUTTONDOWN, WM_LBUTTONUP, and
 *				WM_MOUSEMOVE messages.
 *
 * PARAMETERS:	hwnd - window received in the GetMessage MSG.
 *
 *				lparam - LPARAM received in the GetMessage MSG.
 *
 *				down - true if the mouse button is down.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_PenEvent (HWND hwnd, LPARAM lparam, Bool down)
{
	// If dragging the window, handle that.
	if (gMoveDrag)
	{
		POINT p;
		p.x = (short) LOWORD (lparam);
		p.y = (short) HIWORD (lparam);
		::ClientToScreen (hwnd, &p);

		p.x += gDragOrigRect.left - gDragBase.x;
		p.y += gDragOrigRect.top - gDragBase.y;
		::SetWindowPos (hwnd, NULL, p.x, p.y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);

		gMoveDrag = down;

		if (!gMoveDrag)
		{
			::ReleaseCapture ();
		}

		return;
	}

	// Determine the mouse's location in 1x scale and where 0,0
	// is 0,0 on the LCD.

	int	offsetX = gPrefs->GrayBackground () ? kPlasticBaseX1x : 0;
	int	offsetY = gPrefs->GrayBackground () ? kPlasticBaseY1x : 0;

	int x = LOWORD (lparam) / gPrefs->GetScale () - offsetX;
	int y = HIWORD (lparam) / gPrefs->GetScale () - offsetY;

	// If tracking pen movement across the screen handle that.
	if (gPenTracking)
	{
		if (x < 0)
			x = 0;
		if (x >= 160)
			x = 160 - 1;
		if (y < 0)
			y = 0;
		if (y >= 160 + 60)
			y = 160 + 60 - 1;

		::CPU_pen (down, x, y);
		gPenTracking = down;

		if (!gPenTracking)
		{
			::ReleaseCapture ();
		}
	}

	// If clicking on a hard button, handle that.
	else if (gButtonTracking)
	{
		gButtonTracking = down;
		if (!gButtonTracking)
		{
			::ReleaseCapture ();
			::CPU_button (0, gCurrentButton);
		}
	}

	// No previous mouse tracking operations are in progress.
	// Look to start something new.
	else if (down)
	{
		// If in the LCD area (including the silkscreen area)
		// start "pen tracking".
		if (x >= 0 && x < 160 && y >= 0 && y <= 160 + 60)
		{
			::SetCapture (hwnd);
			::CPU_pen (true, x, y);
			gPenTracking = TRUE;
		}
		else
		{
			// Test to see if the use clicked on any of the hard
			// buttons.  If so, start "button tracking".
			POINT	pt = {x + kPlasticBaseX1x, y + kPlasticBaseY1x};

			// See if we're in the bounding rectangle that surrounds all
			// of our hot buttons.  Testing for this has the effect of
			// not dropping into the "drag the window" code if we just
			// miss clicking on a button.

			if (::PtInRect (&kHardButtons[countof (kHardButtons) - 1], pt))
			{
				for (int ii = 0; ii < countof (kHardButtons) - 1; ++ii)
				{
					if (::PtInRect (&kHardButtons[ii], pt))
					{
						gCurrentButton = ii;	// Maps to kButton_Power, etc.
						::SetCapture (hwnd);
						::CPU_button (true, gCurrentButton);
						gButtonTracking = TRUE;
						break;
					}
				}
			}

			// Otherwise, we've clicked in the frame.
			// Start "move tracking".
			else
			{
				::SetCapture (hwnd);
				::GetWindowRect (hwnd, &gDragOrigRect);
				gDragBase.x = LOWORD (lparam);
				gDragBase.y = HIWORD (lparam);
				::ClientToScreen (hwnd, &gDragBase);
				gMoveDrag = TRUE;
			}
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	LCD_Refresh
 *
 * DESCRIPTION:	Called in response to WM_TIMER messages (to periodically
 *				update our screen.
 *
 * PARAMETERS:	hwnd - window received in the GetMessage MSG.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_Refresh (HWND hwnd)
{
	HDC			dc = ::GetDC (hwnd);
	HPALETTE	op = NULL;

	// If we're in realistic mode, install our palette preparatory
	// to drawing.

	if (gPrefs->GrayBackground ())
	{
		op = ::SelectPalette (dc, gPalette, FALSE);
		::RealizePalette (dc);
	}

	// Update the LCD screen.

	::PaintScreen (dc, FALSE);

	// Restore the old palette.

	if (op)
	{
		::SelectPalette (dc, op, TRUE);
	}

	::ReleaseDC (hwnd, dc);
}


/***********************************************************************
 *
 * FUNCTION:	LCD_Paint
 *
 * DESCRIPTION:	Redraws the window in response to WM_PAINT messages.
 *
 * PARAMETERS:	hwnd - window received in the GetMessage MSG.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_Paint (HWND hwnd)
{
	PAINTSTRUCT	ps;
	HDC			destDC = ::BeginPaint (hwnd, &ps);
	HPALETTE	op = NULL;

	if (gPrefs->GrayBackground ())
	{
		op = ::SelectPalette (destDC, gPalette, FALSE);
		::RealizePalette (destDC);
	}

	HDC		srcDC	= ::CreateCompatibleDC (NULL);
	HBITMAP	ob		= SelectBitmap (srcDC, gBitmap);

	int	top;
	int	left;
	int	width;
	int	height;

	if (gPrefs->GrayBackground ())
	{
		top = 0;
		left = 0;
		if (gPrefs->GetScale () == 1)
		{
			width = kPlasticWidth1x;
			height = kPlasticHeight1x;
		}
		else
		{
			width = kPlasticWidth2x;
			height = kPlasticHeight2x;
		}
	}
	else
	{
		top = 160 * gPrefs->GetScale ();
		left = 0;
		width = 160 * gPrefs->GetScale ();
		height = 60 * gPrefs->GetScale ();
	}

	::BitBlt (	destDC,		// dest DC
				left, top,	// dest x,y
				width,		// dest width
				height,		// dest height
				srcDC,		// src DC
				0, 0,		// src x,y
				SRCCOPY);	// raster op

	::SelectObject (srcDC, ob);
	::DeleteDC (srcDC);
	::PaintScreen (destDC, TRUE);

	if (op)
	{
		::SelectPalette (destDC, op, TRUE);
	}

	::EndPaint (hwnd, &ps);
}


/***********************************************************************
 *
 * FUNCTION:	LCD_SelectPalette
 *
 * DESCRIPTION:	Called in response to WM_PALETTECHANGED and
 *				WM_QUERYNEWPALETTE messages.  Installs the custom
 *				palette for our window (if we're using the "realistic"
 *				display, and not the white one).
 *
 * PARAMETERS:	hwnd - window received in the GetMessage MSG.
 *
 * RETURNED:	The result of RealizePalette (the number of entries in
 *				the logical palette mapped to the system palette if the
 *				function succeeds, GDI_ERROR if it fails) or -1 if we're
 *				not in "realistic" mode.
 *
 ***********************************************************************/

UINT LCD_SelectPalette (HWND hwnd)
{
	if (gPrefs->GrayBackground ())
	{
		HDC			dc = ::GetDC (hwnd);
		HPALETTE	op = ::SelectPalette (dc, gPalette, FALSE);
		UINT		n = ::RealizePalette (dc);

		::SelectPalette (dc, op, TRUE);
		::ReleaseDC (hwnd, dc);

		if (n > 0)
		{
			::InvalidateRect (hwnd, NULL, FALSE);
		}

		return n;
	}

	return -1;
}


/***********************************************************************
 *
 * FUNCTION:	LCD_ScreenShot
 *
 * DESCRIPTION:	Called to save the screen contents to a file when the
 *				user selects the "Save Screen..." menu item.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void LCD_ScreenShot (void)
{
	BOOL			wasRunning = ::CPU_stop ();
	char			fn[MAX_PATH] = {0};
	OPENFILENAME	reply;
	memset (&reply, 0, sizeof(reply));

	reply.lStructSize		= sizeof (OPENFILENAME);
	reply.hInstance			= gInstance;
	reply.lpstrFilter		= "Bitmap Files (*.bmp)\0*.bmp\0";
	reply.nFilterIndex		= 1;
	reply.lpstrFile			= fn;
	reply.nMaxFile			= sizeof (fn);
	reply.lpstrInitialDir	= gHostPrefs->GetLastPRCDir ().GetFilePath ();
	reply.Flags				= OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY | OFN_ENABLEHOOK | OFN_EXPLORER;
	reply.lpstrDefExt		= "bmp";
	reply.lpfnHook			= &::CenterDlgProc;

	if (::GetSaveFileName (&reply))
	{
		HANDLE	file = ::CreateFile (reply.lpstrFile, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
									FILE_ATTRIBUTE_NORMAL, NULL);

		if (file != INVALID_HANDLE_VALUE)
		{
			Screen::BufferInfo	info;
			
			if (Screen::GetBits (info, false /* !incremental */))
			{
				long	numColors = 1 << info.lcdDepth;

				DWORD	fileHeaderSize		= sizeof (BITMAPFILEHEADER);
				DWORD	bitmapHeaderSize	= sizeof (BITMAPINFOHEADER);
				DWORD	colorTableSize		= numColors * sizeof (RGBQUAD);
				DWORD	bitmapOffset		= fileHeaderSize + bitmapHeaderSize + colorTableSize;
				DWORD	bitmapSize			= info.myRowBytes * info.height;
				DWORD	fileSize			= bitmapOffset + bitmapSize;

				BITMAPFILEHEADER	fileHeader;

				fileHeader.bfType		= 'MB';
				fileHeader.bfSize		= fileSize;
				fileHeader.bfReserved1	= 0;
				fileHeader.bfReserved2	= 0;
				fileHeader.bfOffBits	= bitmapOffset;

				BITMAPINFOHEADER	bitmapHeader;

				bitmapHeader.biSize				= sizeof (BITMAPINFOHEADER);
				bitmapHeader.biWidth			= info.myRowBytes * 8 / info.myDepth;
				bitmapHeader.biHeight			= info.height;
				bitmapHeader.biPlanes			= 1;
				bitmapHeader.biBitCount			= info.myDepth;
				bitmapHeader.biCompression		= BI_RGB;
				bitmapHeader.biSizeImage		= 0;
				bitmapHeader.biXPelsPerMeter	= 0;
				bitmapHeader.biYPelsPerMeter	= 0;
				bitmapHeader.biClrUsed			= numColors;
				bitmapHeader.biClrImportant		= 0;

				RGBQUAD*	colors = (RGBQUAD*) _alloca (colorTableSize);
				::PrvSetColorRamp (colors, numColors, info.myCLUT);

				DWORD	written;

				::WriteFile (file, &fileHeader, fileHeaderSize, &written, NULL);
				::WriteFile (file, &bitmapHeader, bitmapHeaderSize, &written, NULL);
				::WriteFile (file, colors, colorTableSize, &written, NULL);

				// What the...?  When I tried writing the negative of height to
				// indicate a top-down DIB, neither Paintbrush nor VC++ could
				// open the file!  So let's do it the hard way...

	//			WriteFile (file, bitmap, bitmapSize, &written, NULL);

				for (int yy = info.height - 1; yy >= 0; --yy)
				{
					void*	src = info.myBuffer + yy * info.myRowBytes;
					::WriteFile (file, src, info.myRowBytes, &written, NULL);
				}
			}

			::CloseHandle (file);
		}
	}

	if (wasRunning)
		::CPU_run ();
}


/***********************************************************************
 *
 * FUNCTION:	PrvLoadResourceBitmap
 *
 * DESCRIPTION:	Load a bitmap resource and return it as a GDI object.
 *				Also create and return a palette for the bitmap.
 *
 *				From KnowledgeBase article Q124947.
 *
 * PARAMETERS:	hInstance - file containing the bitmap
 *
 *				lpString - name of the resource (could be an integer)
 *
 *				lphPalette - pointer to HPALETTE to receive the newly
 *					created palette.
 *
 * RETURNED:	Handle to the bitmap object.
 *
 ***********************************************************************/

HBITMAP PrvLoadResourceBitmap (HINSTANCE hInstance, LPSTR lpString, HPALETTE* lphPalette)
{
	HRSRC				hRsrc;
	HGLOBAL				hGlobal;
	HBITMAP				hBitmapFinal = NULL;
	LPBITMAPINFOHEADER	lpbi;
	HDC					hdc;
	int				iNumColors;

	hRsrc = FindResource(hInstance, lpString, RT_BITMAP);
	if (hRsrc != NULL)
	{
		hGlobal = LoadResource(hInstance, hRsrc);
		lpbi = (LPBITMAPINFOHEADER) LockResource (hGlobal);

		hdc = GetDC(NULL);
		*lphPalette = PrvCreateDIBPalette ((LPBITMAPINFO) lpbi, &iNumColors);
		if (*lphPalette)
		{
			SelectPalette (hdc, *lphPalette, FALSE);
			RealizePalette (hdc);
		}

		hBitmapFinal = CreateDIBitmap (hdc,
			(LPBITMAPINFOHEADER) lpbi,
			(LONG) CBM_INIT,
			(LPSTR) lpbi + lpbi->biSize + iNumColors * sizeof(RGBQUAD),
			(LPBITMAPINFO) lpbi,
			DIB_RGB_COLORS);

		ReleaseDC (NULL, hdc);
		UnlockResource (hGlobal);
		FreeResource (hGlobal);
	}

	return hBitmapFinal;
}


/***********************************************************************
 *
 * FUNCTION:	PrvCreateDIBPalette
 *
 * DESCRIPTION:	Create and return a palette for the givem bitmap.
 *
 *				From KnowledgeBase article Q124947.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

HPALETTE PrvCreateDIBPalette (LPBITMAPINFO lpbmi, LPINT lpiNumColors)
{
	LPBITMAPINFOHEADER	lpbi;
	LPLOGPALETTE		lpPal;
	HANDLE				hLogPal;
	HPALETTE			hPal = NULL;
	int					i;

	lpbi = (LPBITMAPINFOHEADER) lpbmi;
	if (lpbi->biBitCount <= 8)
		*lpiNumColors = (1 << lpbi->biBitCount);
	else
		*lpiNumColors = 0;  // No palette needed for 24 BPP DIB

	if (lpbi->biClrUsed > 0)
		*lpiNumColors = lpbi->biClrUsed;  // Use biClrUsed

	if (*lpiNumColors)
	{
		hLogPal = GlobalAlloc (GHND, sizeof (LOGPALETTE) + sizeof (PALETTEENTRY) * (*lpiNumColors));
		lpPal = (LPLOGPALETTE) GlobalLock (hLogPal);
		lpPal->palVersion		= 0x300;
		lpPal->palNumEntries	= *lpiNumColors;

		for (i = 0; i < *lpiNumColors; i++)
		{
			lpPal->palPalEntry[i].peRed		= lpbmi->bmiColors[i].rgbRed;
			lpPal->palPalEntry[i].peGreen	= lpbmi->bmiColors[i].rgbGreen;
			lpPal->palPalEntry[i].peBlue	= lpbmi->bmiColors[i].rgbBlue;
			lpPal->palPalEntry[i].peFlags	= 0;
		}

		hPal = CreatePalette (lpPal);
		GlobalUnlock (hLogPal);
		GlobalFree (hLogPal);
	}

	return hPal;
}
